static char rcsid[] = "$Header: /usr/users/terence/src/xtrek/netrek/RCS/socket.c,v 1.7 92/04/08 10:03:31 terence Exp $";

/* $Log:	socket.c,v $
 * Revision 1.7  92/04/08  10:03:31  terence
 * Added UDP v1.0 mods.
 * 
 * Revision 1.6  92/04/01  16:52:52  terence
 * removed #ifdefs corresponding to INCLUDE_SCAN and INCLUDE_VISTRACT
 * 
 * Revision 1.5  92/04/01  15:53:24  terence
 * Added a patch to readFromServer():
 *                 commSwitchTimeout = 0;
 *                 if (udpClientRecv != MODE_SIMPLE)                       <---
 *                     sendUdpReq(COMM_MODE + udpClientRecv);              <---
 *                 if (udpWin) {
 * 
 * Revision 1.4  92/04/01  15:51:57  terence
 * Added UDP v0.3 mods
 * 
 * Revision 1.3  92/03/19  21:29:25  terence
 * added UDP mods (version 0.2, 3/17/92)
 * 
 * Revision 1.2  92/03/12  00:13:24  terence
 * added UDP mods (huge)
 * 
 * Revision 1.1  92/03/11  23:47:05  terence
 * Initial revision
 *  */

/*
 * Socket.c
 *
 * Kevin P. Smith 1/29/89
 * UDP stuff v1.0 by Andy McFadden  Feb-Apr 1992
 *
 * UDP protocol v1.0
 *
 * Routines to allow connection to the xtrek server.
 */
#include "copyright2.h"

#ifndef GATEWAY
# define USE_PORTSWAP         /* instead of using recvfrom() */
#endif

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include "Wlib.h"
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"

/*#define INCLUDE_SCAN		/* include Amdahl scanning beams */
/*#define INCLUDE_VISTRACT	/* include visible tractor beams */

#ifdef GATEWAY
/*
 * (these values are now defined in "main.c":)
 * char *gw_mach	= "charon";	|client gateway; strcmp(serverName)
 * int   gw_serv_port	= 5000;		|what to tell the server to use
 * int   gw_port	= 5001;		|where we will contact gw
 * int   gw_local_port	= 5100;		|where we expect gw to contact us
 *
 * The client binds to "5100" and sends "5000" to the server (TCP).  The
 * server sees that and sends a UDP packet to gw on port udp5000, which passes
 * it through to port udp5100 on the client.  The client-gw gets the server's
 * host and port from recvfrom.  (The client can't use the same method since
 * these sockets are one-way only, so it connect()s to gw_port (udp5001)
 * on the gateway machine regardless of what the server sends.)
 *
 * So all we need in .gwrc is:
 *	udp 5000 5001 tde.uts 5100
 *
 * assuming the client is on tde.uts.  Note that a UDP declaration will
 * work for ANY server, but you need one per player, and the client has to
 * have the port numbers in advance.
 *
 * If we're using a standard server, we're set.  If we're running through a
 * gatewayed server, we have to do some unpleasant work on the server side...
 */
#endif

int handleMessage(), handlePhaser(), handlePlyrInfo();
int handlePlayer(), handleSelf(), handleStatus();
int handleWarning(), handleTorp(), handleTorpInfo();
int handlePlanet(), handleQueue(), handlePickok(), handleLogin();
int handlePlasmaInfo(), handlePlasma();
int handleKills(), handleFlags(), handlePStatus();
int handleMotd(), handleMask();
int handleBadVersion(), handlePlanetLoc();
int handleHostile(), handleStats(), handlePlyrLogin();
int handleReserved();
int handleUdpReply(), handleSequence();

struct packet_handler handlers[] = {
    { 0, NULL },	/* record 0 */
    { sizeof(struct mesg_spacket), handleMessage }, 	    /* SP_MESSAGE */
    { sizeof(struct plyr_info_spacket), handlePlyrInfo },   /* SP_PLAYER_INFO */
    { sizeof(struct kills_spacket), handleKills },	    /* SP_KILLS */
    { sizeof(struct player_spacket), handlePlayer },	    /* SP_PLAYER */
    { sizeof(struct torp_info_spacket), handleTorpInfo },   /* SP_TORP_INFO */
    { sizeof(struct torp_spacket), handleTorp }, 	    /* SP_TORP */
    { sizeof(struct phaser_spacket), handlePhaser },	    /* SP_PHASER */
    { sizeof(struct plasma_info_spacket), handlePlasmaInfo},/* SP_PLASMA_INFO */
    { sizeof(struct plasma_spacket), handlePlasma},	    /* SP_PLASMA */
    { sizeof(struct warning_spacket), handleWarning },	    /* SP_WARNING */
    { sizeof(struct motd_spacket), handleMotd },	    /* SP_MOTD */
    { sizeof(struct you_spacket), handleSelf },		    /* SP_YOU */
    { sizeof(struct queue_spacket), handleQueue },	    /* SP_QUEUE */
    { sizeof(struct status_spacket), handleStatus },	    /* SP_STATUS */
    { sizeof(struct planet_spacket), handlePlanet }, 	    /* SP_PLANET */
    { sizeof(struct pickok_spacket), handlePickok },	    /* SP_PICKOK */
    { sizeof(struct login_spacket), handleLogin }, 	    /* SP_LOGIN */
    { sizeof(struct flags_spacket), handleFlags },	    /* SP_FLAGS */
    { sizeof(struct mask_spacket), handleMask },	    /* SP_MASK */
    { sizeof(struct pstatus_spacket), handlePStatus },	    /* SP_PSTATUS */
    { sizeof(struct badversion_spacket), handleBadVersion },/* SP_BADVERSION */
    { sizeof(struct hostile_spacket), handleHostile },      /* SP_HOSTILE */
    { sizeof(struct stats_spacket), handleStats },	    /* SP_STATS */
    { sizeof(struct plyr_login_spacket), handlePlyrLogin }, /* SP_PL_LOGIN */
    { sizeof(struct reserved_spacket), handleReserved },    /* SP_RESERVED */
    { sizeof(struct planet_loc_spacket), handlePlanetLoc },  /* SP_PLANET_LOC */
    { sizeof(struct scan_spacket), NULL }, /* SP_SCAN (ATM) */
    { sizeof(struct udp_reply_spacket), handleUdpReply },   /* SP_UDP_STAT */
    { sizeof(struct sequence_spacket), handleSequence },    /* SP_SEQUENCE */
    { sizeof(struct sc_sequence_spacket), handleSequence }  /* SP_SC_SEQUENCE */
};

int sizes[] = {
    0,	/* record 0 */
    sizeof(struct mesg_cpacket), 		/* CP_MESSAGE */
    sizeof(struct speed_cpacket),		/* CP_SPEED */
    sizeof(struct dir_cpacket),			/* CP_DIRECTION */
    sizeof(struct phaser_cpacket),		/* CP_PHASER */
    sizeof(struct plasma_cpacket),		/* CP_PLASMA */
    sizeof(struct torp_cpacket),		/* CP_TORP */
    sizeof(struct quit_cpacket), 		/* CP_QUIT */
    sizeof(struct login_cpacket),		/* CP_LOGIN */
    sizeof(struct outfit_cpacket),		/* CP_OUTFIT */
    sizeof(struct war_cpacket),			/* CP_WAR */
    sizeof(struct practr_cpacket),		/* CP_PRACTR */
    sizeof(struct shield_cpacket),		/* CP_SHIELD */
    sizeof(struct repair_cpacket),		/* CP_REPAIR */
    sizeof(struct orbit_cpacket),		/* CP_ORBIT */
    sizeof(struct planlock_cpacket),		/* CP_PLANLOCK */
    sizeof(struct playlock_cpacket),		/* CP_PLAYLOCK */
    sizeof(struct bomb_cpacket),		/* CP_BOMB */
    sizeof(struct beam_cpacket),		/* CP_BEAM */
    sizeof(struct cloak_cpacket),		/* CP_CLOAK */
    sizeof(struct det_torps_cpacket),		/* CP_DET_TORPS */
    sizeof(struct det_mytorp_cpacket),		/* CP_DET_MYTORP */
    sizeof(struct copilot_cpacket),		/* CP_COPILOT */
    sizeof(struct refit_cpacket),		/* CP_REFIT */
    sizeof(struct tractor_cpacket),		/* CP_TRACTOR */
    sizeof(struct repress_cpacket),		/* CP_REPRESS */
    sizeof(struct coup_cpacket),		/* CP_COUP */
    sizeof(struct socket_cpacket),		/* CP_SOCKET */
    sizeof(struct options_cpacket),		/* CP_OPTIONS */
    sizeof(struct bye_cpacket),			/* CP_BYE */
    sizeof(struct dockperm_cpacket),		/* CP_DOCKPERM */
    sizeof(struct updates_cpacket),		/* CP_UPDATES */
    sizeof(struct resetstats_cpacket),		/* CP_RESETSTATS */
    sizeof(struct reserved_cpacket),		/* CP_RESERVED */
    sizeof(struct scan_cpacket),		/* CP_SCAN (ATM) */
    sizeof(struct udp_req_cpacket),		/* CP_UDP_REQ */
    sizeof(struct sequence_cpacket)		/* CP_SEQUENCE */
};

#define NUM_PACKETS (sizeof(handlers) / sizeof(handlers[0]) - 1)
#define NUM_SIZES (sizeof(sizes) / sizeof(sizes[0]) - 1)

int serverDead=0;

static int udpLocalPort = 0;
static int udpServerPort = 0;
static long serveraddr = 0;
static long sequence;
static int drop_flag = 0;
static int chan = -1;		/* tells sequence checker where packet is from*/
static short fSpeed, fDirection, fShield, fOrbit, fRepair, fBeamup, fBeamdown,
	fCloak, fBomb, fDockperm, fPhaser, fPlasma, fPlayLock, fPlanLock,
	fTractor, fRepress;

/* reset all the "force command" variables */
resetForce()
{
    fSpeed = fDirection = fShield = fOrbit = fRepair = fBeamup = fBeamdown =
    fCloak = fBomb = fDockperm = fPhaser = fPlasma = fPlayLock = fPlanLock =
    fTractor = fRepress = -1;
}

/*
 * If something we want to happen hasn't yet, send it again.
 *
 * The low byte is the request, the high byte is a max count.  When the max
 * count reaches zero, the client stops trying.  Checking is done with a
 * macro for speed & clarity.
 */
#define FCHECK_FLAGS(flag, force, const) {			\
	if (force > 0) {					\
	    if (((me->p_flags & flag) && 1) ^ ((force & 0xff) && 1)) {	\
		speedReq.type = const;				\
		speedReq.speed = (force & 0xff);		\
		sendServerPacket(&speedReq);			\
		V_UDPDIAG(("Forced %d:%d\n", const, force & 0xff));	\
		force -= 0x100;					\
		if (force < 0x100) force = -1;	/* give up */	\
	    } else						\
		force = -1;					\
	}							\
}
#define FCHECK_VAL(value, force, const) {			\
	if (force > 0) {					\
	    if ((value) != (force & 0xff)) {			\
		speedReq.type = const;				\
		speedReq.speed = (force & 0xff);		\
		sendServerPacket(&speedReq);			\
		V_UDPDIAG(("Forced %d:%d\n", const, force & 0xff));	\
		force -= 0x100;					\
		if (force < 0x100) force = -1;	/* give up */	\
	    } else						\
		force = -1;					\
	}							\
}
#define FCHECK_TRACT(flag, force, const) {			\
	if (force > 0) {					\
	    if (((me->p_flags & flag) && 1) ^ ((force & 0xff) && 1)) {	\
		tractorReq.type = const;			\
		tractorReq.state = ((force & 0xff) >= 0x40);	\
		tractorReq.pnum = (force & 0xff) & (~0x40);	\
		sendServerPacket(&tractorReq);			\
		V_UDPDIAG(("Forced %d:%d/%d\n", const,		\
			tractorReq.state, tractorReq.pnum));	\
		force -= 0x100;					\
		if (force < 0x100) force = -1;	/* give up */	\
	    } else						\
		force = -1;					\
	}							\
}

checkForce()
{
    struct speed_cpacket speedReq;
    struct tractor_cpacket tractorReq;

    FCHECK_VAL(me->p_speed, fSpeed, CP_SPEED);	/* almost always repeats */
    FCHECK_VAL(me->p_dir, fDirection, CP_DIRECTION);	/* (ditto) */
    FCHECK_FLAGS(PFSHIELD, fShield, CP_SHIELD);
    FCHECK_FLAGS(PFORBIT, fOrbit, CP_ORBIT);
    FCHECK_FLAGS(PFREPAIR, fRepair, CP_REPAIR);
    FCHECK_FLAGS(PFBEAMUP, fBeamup, CP_BEAM);
    FCHECK_FLAGS(PFBEAMDOWN, fBeamdown, CP_BEAM);
    FCHECK_FLAGS(PFCLOAK, fCloak, CP_CLOAK);
    FCHECK_FLAGS(PFBOMB, fBomb, CP_BOMB);
    FCHECK_FLAGS(PFDOCKOK, fDockperm, CP_DOCKPERM);
    FCHECK_VAL(phasers[me->p_no].ph_status, fPhaser, CP_PHASER);/* bug: dir 0 */
    FCHECK_VAL(plasmatorps[me->p_no].pt_status, fPlasma, CP_PLASMA); /*(ditto)*/
    FCHECK_FLAGS(PFPLOCK, fPlayLock, CP_PLAYLOCK);
    FCHECK_FLAGS(PFPLLOCK, fPlanLock, CP_PLANLOCK);

    FCHECK_TRACT(PFTRACT, fTractor, CP_TRACTOR);
    FCHECK_TRACT(PFPRESS, fRepress, CP_REPRESS);
}


connectToServer(port)
int port;
{
    int s;
    struct sockaddr_in addr;
    struct sockaddr_in naddr;
    int len;
/*    int readfds;*/
    fd_set readfds;
    struct timeval timeout;
    struct hostent *hp;

    serverDead=0;
    if (sock!=-1) {
	shutdown(sock, 2);
	sock= -1;
    }
    sleep(3);		/* I think this is necessary for some unknown reason */

    printf("Waiting for connection (port %d).\n", port);

    if ((s=socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	printf("I can't create a socket\n");
	exit(2);
    }

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_port = htons(port);

    if (bind(s, &addr, sizeof(addr)) < 0) {
	sleep(10);
	if (bind(s, &addr, sizeof(addr)) < 0) {
	    sleep(10);
	    if (bind(s, &addr, sizeof(addr)) < 0) {
		printf("I can't bind to port!\n");
		exit(3);
	    }
	}
    }

    listen(s,1);

    len=sizeof(naddr);

tryagain:
    timeout.tv_sec=240;	/* four minutes */
    timeout.tv_usec=0;
/*    readfds=(1<<s);*/
    FD_ZERO(&readfds);
    FD_SET(s, &readfds);
    if (select(32, &readfds, NULL, NULL, &timeout) == 0) {
	printf("Well, I think the server died!\n");
	exit(0);
    }

    sock=accept(s,&naddr,&len);

    if (sock==-1) {
	extern int errno;
	printf("Augh!  Bad accept: %d\n", errno);
	goto tryagain;
    }

/*    printf("Got connection.\n");*/

    close(s);
    pickSocket(port);	/* new socket != port */

    /*
     * This isn't strictly necessary; it tries to determine who the
     * caller is, and set "serverName" and "serveraddr" appropriately.
     */
    len = sizeof(struct sockaddr_in);
    if (getpeername(sock, (struct sockaddr *) &addr, &len) < 0) {
	perror("unable to get peername");
	serverName = "nowhere";
    } else {
	serveraddr = addr.sin_addr.s_addr;
	hp = gethostbyaddr((char *)&addr.sin_addr.s_addr, sizeof(long),AF_INET);
	if (hp != NULL) {
	    serverName = (char *) malloc(strlen(hp->h_name)+1);
	    strcpy(serverName, hp->h_name);
	} else {
	    serverName = (char *) malloc(strlen(inet_ntoa(addr.sin_addr))+1);
	    strcpy(serverName, inet_ntoa(addr.sin_addr));
	}
    }
    printf("Got connection from server %s (0x%x)\n", serverName, serveraddr);
}

callServer(port, server)
int port;
char *server;
{
    int s;
    struct sockaddr_in addr;
    struct hostent *hp;

    serverDead=0;

    printf("Calling %s (%d).\n", server, port);

    if ((s=socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	printf("I can't create a socket\n");
	exit(0);
    }

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    
    if ((addr.sin_addr.s_addr = inet_addr(server)) == -1) {
	if ((hp = gethostbyname(server)) == NULL) {
	    printf("Who is %s?\n", server);
	    exit(0);
	} else {
	    addr.sin_addr.s_addr = *(long *) hp->h_addr;
	}
    }
    serveraddr = addr.sin_addr.s_addr;

    if (connect(s, &addr, sizeof(addr)) < 0) {
	extern int errno;	/* silly compilers 1/21/92 TC */
	printf("Server not listening! Errno: %d\n", errno);
	exit(0);
    }

    printf("Got connection.\n");

    sock=s;
    pickSocket(port);	/* new socket != port */
}

isServerDead()
{
    return(serverDead);
}

socketPause()
{
    struct timeval timeout;
/*    int readfds;*/
    fd_set readfds;

    timeout.tv_sec=1;
    timeout.tv_usec=0;
/*    readfds = 1<<sock;*/
    FD_ZERO(&readfds);
    FD_SET(sock, &readfds);
    select(32,&readfds,0,0,&timeout);
}

readFromServer()
{
    struct timeval timeout;
    fd_set readfds;
    int retval = 0;

    if (serverDead) return(0);
    if (commMode == COMM_TCP) drop_flag=0;	/* just in case */
    timeout.tv_sec=0;
    timeout.tv_usec=0;
/*    readfds = 1<<sock;*/
    FD_ZERO(&readfds);
    FD_SET(sock, &readfds);
    if (udpSock >= 0)
	FD_SET(udpSock, &readfds);
    if (select(32,&readfds,0,0,&timeout) != 0) {
	/* Read info from the xtrek server */
	if (FD_ISSET(sock, &readfds)) {
	    chan = sock;
	    retval += doRead(sock);
	}
	if (udpSock >= 0 && FD_ISSET(udpSock, &readfds)) {
	    V_UDPDIAG(("Activity on UDP socket\n"));
	    chan = udpSock;
	    if (commStatus == STAT_VERIFY_UDP) {
		warning("UDP connection established");
		sequence = 0;		/* reset sequence #s */
		resetForce();
		commMode = COMM_UDP;
		commStatus = STAT_CONNECTED;
                if (udpClientRecv != MODE_SIMPLE)
                    sendUdpReq(COMM_MODE + udpClientRecv);
		commSwitchTimeout = 0;
		if (udpClientRecv != MODE_SIMPLE)
		    sendUdpReq(COMM_MODE + udpClientRecv);
		if (udpWin) {
		    udprefresh(UDP_CURRENT);
		    udprefresh(UDP_STATUS);
		}
	    }
	    retval += doRead(udpSock);
	}
    }

    /* if switching comm mode, decrement timeout counter */
    if (commSwitchTimeout > 0) {
/*	printf("commSwitchTimeout = %d\n", commSwitchTimeout);*/
	if (!(--commSwitchTimeout)) {
	    /*
	     * timed out; could be initial request to non-UDP server (which
	     * won't be answered), or the verify packet got lost en route
	     * to the server.  Could also be a request for TCP which timed
	     * out (weird), in which case we just reset anyway.
	     */
	    commModeReq = commMode = COMM_TCP;
	    commStatus = STAT_CONNECTED;
	    if (udpSock >= 0)
		closeUdpConn();
	    if (udpWin) {
		udprefresh(UDP_CURRENT);
		udprefresh(UDP_STATUS);
	    }
	    warning("Timed out waiting for UDP response from server");
	    UDPDIAG(("Timed out waiting for UDP response from server\n"));
	}
    }

    /* if we're in a UDP "force" mode, check to see if we need to do something*/
    if (commMode == COMM_UDP && udpClientSend > 1)
	checkForce();

    return (retval != 0);		/* convert to 1/0 */
}

/* this used to be part of the routine above */
char buf[BUFSIZ*2];
doRead(asock)
int asock;
{
    struct timeval timeout;
    fd_set readfds;
    char *bufptr;
    int size;
    int count;
    int temp;

    timeout.tv_sec=0;
    timeout.tv_usec=0;

	count=read(asock,buf,2*BUFSIZ);
	if (count<=0) {
/*	    printf("asock=%d, sock=%d, udpSock=%d, errno=%d\n",
		asock, sock, udpSock, errno);*/
	    if (asock == udpSock) {
		if (errno == ECONNREFUSED) {
		    struct sockaddr_in addr;

		    UDPDIAG(("Hiccup(%d)!  Reconnecting\n", errno));
		    addr.sin_addr.s_addr = serveraddr;
		    addr.sin_port = htons(udpServerPort);
		    addr.sin_family = AF_INET;
		    if (connect(udpSock, &addr, sizeof(addr)) < 0) {
			perror("connect");
			UDPDIAG(("Unable to reconnect\n"));
			/* and fall through to disconnect */
		    } else {
			UDPDIAG(("Reconnect successful\n"));
			return (0);
		    }
		}

		UDPDIAG(("*** UDP disconnected (res=%d, err=%d)\n",
			count, errno));
		warning("UDP link severed");
		printUdpInfo();
		closeUdpConn();
		commMode = commModeReq = COMM_TCP;
		commStatus = STAT_CONNECTED;
		if (udpWin) {
		    udprefresh(UDP_STATUS);
		    udprefresh(UDP_CURRENT);
		}
		return (0);
	    }
	    printf("1) Got read() of %d (err=%d). Server dead\n", count, errno);
	    serverDead=1;
	    return(0);
	}

	bufptr=buf;
	while (bufptr < buf+count) {
	    if (*bufptr < 1 || *bufptr > NUM_PACKETS || handlers[*bufptr].size == 0) {
		printf("Unknown packet type: %d, aborting...\n", *bufptr);
		return(0);
	    }
	    size=handlers[*bufptr].size;
	    while (size>count+(buf-bufptr)) {
		/* We wait for up to ten seconds for rest of packet.
		 * If we don't get it, we assume the server died. 
		 */
		timeout.tv_sec=20;
		timeout.tv_usec=0;
		FD_ZERO(&readfds);
		FD_SET(asock, &readfds);
		/*readfds=1<<asock;*/
		if (select(32,&readfds,0,0,&timeout) == 0) {
		    printf("Packet fragment.  Server must be dead\n");
		    serverDead=1;
		    return(0);
		}
		temp=read(asock,buf+count,size-(count+(buf-bufptr)));
		count+=temp;
		if (temp<=0) {
		    printf("2) Got read() of %d.  Server is dead\n", temp);
		    serverDead=1;
		    return(0);
		}
	    }
	    if (handlers[*bufptr].handler != NULL) {
		if (asock != udpSock ||
		    (!drop_flag || *bufptr == SP_SEQUENCE || *bufptr == SP_SC_SEQUENCE))
		    (*(handlers[*bufptr].handler))(bufptr);
		else
		    UDPDIAG(("Ignored type %d\n", *bufptr));
	    } else {
		printf("Handler for packet %d not installed...\n", *bufptr);
	    }
	    bufptr+=size;
	    if (bufptr>buf+BUFSIZ) {
		bcopy(buf+BUFSIZ, buf, BUFSIZ);
		if (count==BUFSIZ*2) {
		    FD_ZERO(&readfds);
		    FD_SET(asock, &readfds);
		    /*readfds = 1<<asock;*/
		    if (select(32,&readfds,0,0,&timeout) != 0) {
			temp=read(asock,buf+BUFSIZ,BUFSIZ);
			count=BUFSIZ+temp;
			if (temp<=0) {
			    printf("3) Got read() of %d.  Server is dead\n", temp);
			    serverDead=1;
			    return(0);
			}
		    } else {
			count=BUFSIZ;
		    }
		} else {
		    count-=BUFSIZ;
		}
		bufptr-=BUFSIZ;
	    }
	}
	return(1);
}

handleTorp(packet)
struct torp_spacket *packet;
{
    struct torp *thetorp;

    thetorp= &torps[ntohs(packet->tnum)];
    thetorp->t_x=ntohl(packet->x);
    thetorp->t_y=ntohl(packet->y);
    thetorp->t_dir=packet->dir;
}

handleTorpInfo(packet)
struct torp_info_spacket *packet;
{
    struct torp *thetorp;

    thetorp= &torps[ntohs(packet->tnum)];
    if (packet->status == TEXPLODE && thetorp->t_status == TFREE) {
	/* FAT: redundant explosion; don't update p_ntorp */
	/*printf("texplode ignored\n");*/
	return;
    }
    if (thetorp->t_status==TFREE && packet->status) {
	players[thetorp->t_owner].p_ntorp++;
    }
    if (thetorp->t_status && packet->status==TFREE) {
	players[thetorp->t_owner].p_ntorp--;
    }
    thetorp->t_war=packet->war;
    if (packet->status != thetorp->t_status) {
	/* FAT: prevent explosion reset */
	thetorp->t_status=packet->status;
	if (thetorp->t_status == TEXPLODE) {
	    thetorp->t_fuse=NUMDETFRAMES;
	}
    }
}

handleStatus(packet)
struct status_spacket *packet;
{
    status->tourn=packet->tourn;
    status->armsbomb=ntohl(packet->armsbomb);
    status->planets=ntohl(packet->planets);
    status->kills=ntohl(packet->kills);
    status->losses=ntohl(packet->losses);
    status->time=ntohl(packet->time);
    status->timeprod=ntohl(packet->timeprod);
}

handleSelf(packet)
struct you_spacket *packet;
{
    me= &players[packet->pnum];
    myship = &(me->p_ship);
    mystats= &(me->p_stats);
    me->p_hostile = packet->hostile;
    me->p_swar = packet->swar;
    me->p_armies = packet->armies;
    me->p_flags = ntohl(packet->flags);
    me->p_damage = ntohl(packet->damage);
    me->p_shield = ntohl(packet->shield);
    me->p_fuel = ntohl(packet->fuel);
    me->p_etemp = ntohs(packet->etemp);
    me->p_wtemp = ntohs(packet->wtemp);
    me->p_whydead = ntohs(packet->whydead);
    me->p_whodead = ntohs(packet->whodead);
}


handlePlayer(packet)
struct player_spacket *packet;
{
    register struct player *pl;

    pl= &players[packet->pnum];
    pl->p_dir=packet->dir;
    pl->p_speed=packet->speed;
    pl->p_x=ntohl(packet->x);
    pl->p_y=ntohl(packet->y);
    redrawPlayer[packet->pnum]=1;
}

handleWarning(packet)
struct warning_spacket *packet;
{
    warning(packet->mesg);
}

sendShortPacket(type, state)
char type, state;
{
    struct speed_cpacket speedReq;

    speedReq.type=type;
    speedReq.speed=state;
    sendServerPacket(&speedReq);

    /* if we're sending in UDP mode, be prepared to force it */
    if (commMode == COMM_UDP && udpClientSend >= 2) {
	switch (type) {
	case CP_SPEED:	fSpeed	= state | 0x100; break;
	case CP_DIRECTION:	fDirection = state | 0x100; break;
	case CP_SHIELD:	fShield	= state | 0xa00; break;
	case CP_ORBIT:	fOrbit	= state | 0xa00; break;
	case CP_REPAIR:	fRepair	= state | 0xa00; break;
	case CP_CLOAK:	fCloak	= state | 0xa00; break;
	case CP_BOMB:	fBomb	= state | 0xa00; break;
	case CP_DOCKPERM:	fDockperm = state | 0xa00; break;
	case CP_PLAYLOCK:	fPlayLock = state | 0xa00; break;
	case CP_PLANLOCK:	fPlanLock = state | 0xa00; break;
	case CP_BEAM:
	    if (state == 1) fBeamup = 1 | 0x500;
	    else          fBeamdown = 2 | 0x500;
	    break;
	}

	/* force weapons too? */
	if (udpClientSend >= 3) {
	    switch (type) {
	    case CP_PHASER: fPhaser	= state | 0x100; break;
	    case CP_PLASMA: fPlasma	= state | 0x100; break;
	    }
	}
    }
}

sendServerPacket(packet)
/* Pick a random type for the packet */
struct player_spacket *packet;
{
    int size;
	
    if (serverDead) return;
    if (packet->type<1 || packet->type>NUM_SIZES || sizes[packet->type]==0) {
	printf("Attempt to send strange packet %d!\n", packet->type);
        return;
    }
    size=sizes[packet->type];
    if (commMode == COMM_UDP) {
	/* for now, just sent everything via TCP */
    }
    if (commMode == COMM_TCP || !udpClientSend) {
	/* special case for verify packet */
	if (packet->type == CP_UDP_REQ) {
	    if (((struct udp_req_cpacket *) packet)->request == COMM_VERIFY)
		goto send_udp;
	}

	/*
	 * business as usual
	 * (or player has turned off UDP transmission)
	 */
	if (gwrite(sock, packet, size) != size) {
	    printf("gwrite failed.  Server must be dead\n");
	    serverDead=1;
	}

    } else {
	/*
	 * UDP stuff
	 */
	switch (packet->type) {
	case CP_SPEED:
	case CP_DIRECTION:
	case CP_PHASER:
	case CP_PLASMA:
	case CP_TORP:
	case CP_QUIT:
	case CP_PRACTR:
	case CP_SHIELD:
	case CP_REPAIR:
	case CP_ORBIT:
	case CP_PLANLOCK:
	case CP_PLAYLOCK:
	case CP_BOMB:
	case CP_BEAM:
	case CP_CLOAK:
	case CP_DET_TORPS:
	case CP_DET_MYTORP:
	case CP_REFIT:
	case CP_TRACTOR:
	case CP_REPRESS:
	case CP_COUP:
	case CP_DOCKPERM:
	case CP_SCAN:
	    /* non-critical or semi-critical stuff, use UDP */
send_udp:
	    V_UDPDIAG(("Sent %d on UDP port\n", packet->type));
	    if (gwrite(udpSock, packet, size) != size) {
		UDPDIAG(("gwrite on UDP failed.  Closing UDP connection\n"));
		warning("UDP link severed");
		/*serverDead=1;*/
		commModeReq = commMode = COMM_TCP;
		commStatus = STAT_CONNECTED;
		commSwitchTimeout = 0;
		if (udpWin) {
		    udprefresh(UDP_STATUS);
		    udprefresh(UDP_CURRENT);
		}
		if (udpSock >= 0)
		    closeUdpConn();
	    }
	    break;

	default:
	    /* critical stuff, use TCP */
	    if (gwrite(sock, packet, size) != size) {
		printf("gwrite failed.  Server must be dead\n");
		serverDead=1;
	    }
	}
    }
}

handlePlanet(packet)
struct planet_spacket *packet;
{
    struct planet *plan;
    int redraw = 0;	/* FAT: prevent excessive redraws */

    plan= &planets[packet->pnum];
    if (plan->pl_owner != packet->owner) redraw++;
    plan->pl_owner=packet->owner;
    if (plan->pl_info != packet->info) redraw++;
    plan->pl_info=packet->info;
    /* Redraw the planet because it was updated by server */
    if (plan->pl_flags != ntohs(packet->flags)) redraw++;
    plan->pl_flags=(int) ntohs(packet->flags);
    if (plan->pl_armies != ntohl(packet->armies)) redraw++;
    plan->pl_armies=ntohl(packet->armies);
    if (plan->pl_info==0) {
	plan->pl_owner=NOBODY;
    }
    if (redraw) plan->pl_flags |= PLREDRAW;
}

handlePhaser(packet)
struct phaser_spacket *packet;
{
    struct phaser *phas;

    phas= &phasers[packet->pnum];
    phas->ph_status=packet->status;
    phas->ph_dir=packet->dir;
    phas->ph_x=ntohl(packet->x);
    phas->ph_y=ntohl(packet->y);
    phas->ph_target=ntohl(packet->target);
}

handleMessage(packet)
struct mesg_spacket *packet;
{
    if (packet->m_from >= MAXPLAYER) packet->m_from=255;
    dmessage(packet->mesg, packet->m_flags, packet->m_from, packet->m_recpt);
}

handleQueue(packet)
struct queue_spacket *packet;
{
    queuePos = ntohs(packet->pos);
}

sendTeamReq(team,ship)
int team, ship;
{
    struct outfit_cpacket outfitReq;

    outfitReq.type=CP_OUTFIT;
    outfitReq.team=team;
    outfitReq.ship=ship;
    sendServerPacket(&outfitReq);
}

handlePickok(packet)
struct pickok_spacket *packet;
{
    pickOk=packet->state;
}

sendLoginReq(name,pass,login,query)
char *name, *pass;
char *login;
char query;
{
    struct login_cpacket packet;

    strcpy(packet.name, name);
    strcpy(packet.password, pass);
    if (strlen(login)>15) login[15]=0;
    strcpy(packet.login, login);
    packet.type=CP_LOGIN;
    packet.query=query;
    sendServerPacket(&packet);
}

handleLogin(packet)
struct login_spacket *packet;
{
    loginAccept=packet->accept;
    if (packet->accept) {
	bcopy(packet->keymap, mystats->st_keymap, 96);
	mystats->st_flags = ntohl(packet->flags);
	showShields = (me->p_stats.st_flags / ST_SHOWSHIELDS) & 1;
	mapmode = (me->p_stats.st_flags / ST_MAPMODE) & 1;
	namemode = (me->p_stats.st_flags / ST_NAMEMODE) & 1;
	keeppeace = (me->p_stats.st_flags / ST_KEEPPEACE) & 1;
	showlocal = (me->p_stats.st_flags / ST_SHOWLOCAL) & 3;
	showgalactic = (me->p_stats.st_flags / ST_SHOWGLOBAL) & 3;
    }
}

sendTractorReq(state, pnum)
char state;
char pnum;
{
    struct tractor_cpacket tractorReq;

    tractorReq.type=CP_TRACTOR;
    tractorReq.state=state;
    tractorReq.pnum=pnum;
    sendServerPacket(&tractorReq);

    if (state)
	fTractor = pnum | 0x40;
    else
	fTractor = 0;
}

sendRepressReq(state, pnum)
char state;
char pnum;
{
    struct repress_cpacket repressReq;

    repressReq.type=CP_REPRESS;
    repressReq.state=state;
    repressReq.pnum=pnum;
    sendServerPacket(&repressReq);

    if (state)
	fRepress = pnum | 0x40;
    else
	fRepress = 0;
}

sendDetMineReq(torp)
short torp;
{
    struct det_mytorp_cpacket detReq;

    detReq.type=CP_DET_MYTORP;
    detReq.tnum=htons(torp);
    sendServerPacket(&detReq);
}

handlePlasmaInfo(packet)
struct plasma_info_spacket *packet;
{
    struct plasmatorp *thetorp;

    thetorp= &plasmatorps[ntohs(packet->pnum)];
    if (packet->status == PTEXPLODE && thetorp->pt_status == PTFREE) {
	/* FAT: redundant explosion; don't update p_nplasmatorp */
	return;
    }
    if (!thetorp->pt_status && packet->status) {
	players[thetorp->pt_owner].p_nplasmatorp++;
    }
    if (thetorp->pt_status && !packet->status) {
	players[thetorp->pt_owner].p_nplasmatorp--;
    }
    thetorp->pt_war=packet->war;
    if (thetorp->pt_status != packet->status) {
	/* FAT: prevent explosion timer from being reset */
	thetorp->pt_status=packet->status;
	if (thetorp->pt_status == PTEXPLODE) {
	    thetorp->pt_fuse=NUMDETFRAMES;
	}
    }
}

handlePlasma(packet)
struct plasma_spacket *packet;
{
    struct plasmatorp *thetorp;

    thetorp= &plasmatorps[ntohs(packet->pnum)];
    thetorp->pt_x=ntohl(packet->x);
    thetorp->pt_y=ntohl(packet->y);
}

handleFlags(packet)
struct flags_spacket *packet;
{
    if (players[packet->pnum].p_flags != ntohl(packet->flags) ||
	players[packet->pnum].p_tractor != ((short)packet->tractor & (~0x40))){
	/* FAT: prevent redundant player update */

	redrawPlayer[packet->pnum]=1;
    }

    players[packet->pnum].p_flags=ntohl(packet->flags);
   /*redrawPlayer[packet->pnum]=1;*/
}

handleKills(packet)
struct kills_spacket *packet;
{
    if (players[packet->pnum].p_kills != ntohl(packet->kills)/100.0) {
        /* FAT: prevent redundant player update */
        players[packet->pnum].p_kills=ntohl(packet->kills)/100.0;
        updatePlayer[packet->pnum]=1;
    }
}

handlePStatus(packet)
struct pstatus_spacket *packet;
{
    if (players[packet->pnum].p_status == packet->status) return; /* ignore redundant packet - TC */
    if (packet->status==PEXPLODE) {
	players[packet->pnum].p_explode=0;
    }
    /* Ignore DEAD status.
     * Instead, we treat it as PEXPLODE.
     * This gives us time to animate all the frames necessary for 
     *  the explosions at our own pace.
     */
    if (packet->status==PDEAD) {
	packet->status=PEXPLODE;
    }
    players[packet->pnum].p_status=packet->status;
    redrawPlayer[packet->pnum]=1;
    updatePlayer[packet->pnum]=1;
}

handleMotd(packet)
struct motd_spacket *packet;
{
    newMotdLine(packet->line);
}

sendMessage(mes, group, indiv)
char *mes;
int group, indiv;
{
    struct mesg_cpacket mesPacket;

    mesPacket.type=CP_MESSAGE;
    mesPacket.group=group;
    mesPacket.indiv=indiv;
    strcpy(mesPacket.mesg, mes);
    sendServerPacket(&mesPacket);
}

handleMask(packet)
struct mask_spacket *packet;
{
    tournMask=packet->mask;
}

sendOptionsPacket()
{
    struct options_cpacket optPacket;

    optPacket.type=CP_OPTIONS;
    optPacket.flags = 
	htonl(ST_MAPMODE * (mapmode!=0) +
	ST_NAMEMODE * namemode +
	ST_SHOWSHIELDS * showShields +
	ST_KEEPPEACE * keeppeace + 
	ST_SHOWLOCAL * showlocal +
	ST_SHOWGLOBAL * showgalactic);
    bcopy(mystats->st_keymap, optPacket.keymap, 96);
    sendServerPacket(&optPacket);
}

pickSocket(old)
int old;
{
    int newsocket;
    struct socket_cpacket sockPack;

    newsocket=(getpid() & 32767);
    while (newsocket < 2048 || newsocket==old) {
	newsocket=(newsocket + 10687) & 32767;
    }
    sockPack.type=CP_SOCKET;
    sockPack.socket=htonl(newsocket);
    sockPack.version=(char) SOCKVERSION;
    sockPack.udp_version=(char) UDPVERSION;
    sendServerPacket(&sockPack);
    /* Did we get new socket # sent? */
    if (serverDead) return;
    nextSocket=newsocket;
}

handleBadVersion(packet)
struct badversion_spacket *packet;
{
    switch(packet->why) {
    case 0:
	printf("Sorry, this is an invalid client version.\n");
	printf("You need a new version of the client code.\n");
	break;
    default:
	printf("Sorry, but you cannot play xtrek now.\n");
	printf("Try again later.\n");
	break;
    }
    exit(1);
}

gwrite(fd, buf, bytes)
int fd;
char *buf;
register int bytes;
{
    long orig = bytes;
    register long n;
    while (bytes) {
        n = write(fd, buf, bytes);
        if (n < 0) {
	    if (fd == udpSock) {
		fprintf(stderr, "Tried to write %d, 0x%x, %d (error %d)\n",
		    fd, buf, bytes, errno);
		printUdpInfo();
	    }
            return(-1);
	}
        bytes -= n;
        buf += n;
    }
    return(orig);
}

handleHostile(packet)
struct hostile_spacket *packet;
{
    register struct player *pl;

    pl= &players[packet->pnum];
    if (pl->p_swar != packet->war ||
	pl->p_hostile != packet->hostile) {
	/* FAT: prevent redundant player update & redraw */
	updatePlayer[packet->pnum]=1;
	pl->p_swar=packet->war;
	pl->p_hostile=packet->hostile;
	redrawPlayer[packet->pnum]=1;
    }
}

handlePlyrLogin(packet)
struct plyr_login_spacket *packet;
{
    register struct player *pl;

    updatePlayer[packet->pnum]=1;
    pl= &players[packet->pnum];

    strcpy(pl->p_name, packet->name);
    strcpy(pl->p_monitor, packet->monitor);
    strcpy(pl->p_login, packet->login);
    pl->p_stats.st_rank=packet->rank;
    if (packet->pnum == me->p_no) {
	/* This is me.  Set some stats */
	if (lastRank== -1) {
	    if (loggedIn) {
		lastRank=packet->rank;
	    }
	} else {
	    if (lastRank != packet->rank) {
		lastRank=packet->rank;
		promoted=1;
	    }
	}
    }
}

handleStats(packet)
struct stats_spacket *packet;
{
    register struct player *pl;

    updatePlayer[packet->pnum]=1;
    pl= &players[packet->pnum];
    pl->p_stats.st_tkills=ntohl(packet->tkills);
    pl->p_stats.st_tlosses=ntohl(packet->tlosses);
    pl->p_stats.st_kills=ntohl(packet->kills);
    pl->p_stats.st_losses=ntohl(packet->losses);
    pl->p_stats.st_tticks=ntohl(packet->tticks);
    pl->p_stats.st_tplanets=ntohl(packet->tplanets);
    pl->p_stats.st_tarmsbomb=ntohl(packet->tarmies);
    pl->p_stats.st_sbkills=ntohl(packet->sbkills);
    pl->p_stats.st_sblosses=ntohl(packet->sblosses);
    pl->p_stats.st_armsbomb=ntohl(packet->armies);
    pl->p_stats.st_planets=ntohl(packet->planets);
    pl->p_stats.st_maxkills=ntohl(packet->maxkills) / 100.0;
    pl->p_stats.st_sbmaxkills=ntohl(packet->sbmaxkills) / 100.0;
}

handlePlyrInfo(packet)
struct plyr_info_spacket *packet;
{
    register struct player *pl;
    static int lastship= -1;

    updatePlayer[packet->pnum]=1;
    pl= &players[packet->pnum];
    getship(&pl->p_ship, packet->shiptype);
    pl->p_team=packet->team;
    pl->p_mapchars[0]=teamlet[pl->p_team];
    pl->p_mapchars[1]=shipnos[pl->p_no];
    if (me==pl && lastship!=me->p_ship.s_type) {
	calibrate_stats();
	redrawTstats();
    }
    redrawPlayer[packet->pnum]=1;
}

sendUpdatePacket(speed)
long speed;
{
    struct updates_cpacket packet;

    packet.type=CP_UPDATES;
    timerDelay=speed;
    packet.usecs=htonl(speed);
    sendServerPacket(&packet);
}

handlePlanetLoc(packet)
struct planet_loc_spacket *packet;
{
    struct planet *pl;

    pl= &planets[packet->pnum];
    pl->pl_x=ntohl(packet->x);
    pl->pl_y=ntohl(packet->y);
    strcpy(pl->pl_name, packet->name);
    pl->pl_namelen=strlen(packet->name);
    pl->pl_flags|=PLREDRAW;
    reinitPlanets=1;
}

handleReserved(packet)
struct reserved_spacket *packet;
{
    struct reserved_cpacket response;

    encryptReservedPacket(packet, &response, serverName, me->p_no);
    sendServerPacket(&response);
}

/*
 * ---------------------------------------------------------------------------
 *	UDP stuff from here on
 * ---------------------------------------------------------------------------
 */

sendUdpReq(req)
int req;
{
    struct udp_req_cpacket packet;

    packet.type = CP_UDP_REQ;
    packet.request = req;

    if (req >= COMM_MODE) {
	packet.request = COMM_MODE;
	packet.connmode = req - COMM_MODE;
	sendServerPacket(&packet);
	return;
    }

    if (req == COMM_UPDATE) {
	sendServerPacket(&packet);
	warning("Sent request for full update");
	return;
    }

    /*
     * Past this point, req had better be COMM_TCP or COMM_UDP.
     */

    if (req == commModeReq) {
	warning("Request is in progress, do not disturb");
	return;
    }

    if (req == COMM_UDP) {
	/* open UDP port */
	if (openUdpConn() >= 0) {
	    UDPDIAG(("Bound to local port %d on fd %d\n",udpLocalPort,udpSock));
	} else {
	    UDPDIAG(("Bind to local port %d failed\n", udpLocalPort));
	    commModeReq = COMM_TCP;
	    commStatus = STAT_CONNECTED;
	    commSwitchTimeout = 0;
	    if (udpWin)
		udprefresh(UDP_STATUS);
	    warning("Unable to establish UDP connection");

	    return;
	}
    }

    /* send the request */
    packet.port = htonl(udpLocalPort);
#ifdef GATEWAY
    if (!strcmp(serverName, gw_mach)) {
	packet.port = htons(gw_serv_port); /* gw port that server should call */
	UDPDIAG(("+ Telling server to contact us on %d\n", gw_serv_port));
    }
#endif
#ifdef USE_PORTSWAP
    packet.connmode = CONNMODE_PORT;		/* have him send his port */
#else
    packet.connmode = CONNMODE_PACKET;		/* we get addr from packet */
#endif
    sendServerPacket(&packet);

    /* update internal state stuff */
    commModeReq = req;
    if (req == COMM_TCP)
	commStatus = STAT_SWITCH_TCP;
    else
	commStatus = STAT_SWITCH_UDP;
    commSwitchTimeout = 25;	/* wait 25 updates (about five seconds) */
    UDPDIAG(("Sent request for %s mode\n", (req == COMM_TCP) ?
	"TCP" : "UDP"));

#ifndef USE_PORTSWAP
    if ((req == COMM_UDP) && recvUdpConn() < 0) {
	UDPDIAG(("Sending TCP reset message\n"));
	packet.request = COMM_TCP;
	packet.port = 0;
	commModeReq = COMM_TCP;
	sendServerPacket(&packet);
	/* we will likely get a SWITCH_UDP_OK later; better ignore it */
	commModeReq = COMM_TCP;
	commStatus = STAT_CONNECTED;
	commSwitchTimeout = 0;
    }
#endif

    if (udpWin)
	udprefresh(UDP_STATUS);
}

handleUdpReply(packet)
struct udp_reply_spacket *packet;
{
    struct udp_req_cpacket response;

    UDPDIAG(("Received UDP reply %d\n", packet->reply));
    commSwitchTimeout = 0;

    response.type = CP_UDP_REQ;

    switch (packet->reply) {
    case SWITCH_TCP_OK:
	if (commMode == COMM_TCP) {
	    UDPDIAG(("Got SWITCH_TCP_OK while in TCP mode; ignoring\n"));
	} else {
	    commMode = COMM_TCP;
	    commStatus = STAT_CONNECTED;
	    warning("Switched to TCP-only connection");
	    closeUdpConn();
	    UDPDIAG(("UDP port closed\n"));
	    if (udpWin) {
		udprefresh(UDP_STATUS);
		udprefresh(UDP_CURRENT);
	    }
	}
	break;
    case SWITCH_UDP_OK:
	if (commMode == COMM_UDP) {
	    UDPDIAG(("Got SWITCH_UDP_OK while in UDP mode; ignoring\n"));
	} else {
	    /* the server is forcing UDP down our throat? */
	    if (commModeReq != COMM_UDP) {
		UDPDIAG(("Got unsolicited SWITCH_UDP_OK; ignoring\n"));
	    } else {
#ifdef USE_PORTSWAP
		udpServerPort = ntohl(packet->port);
		if (connUdpConn() < 0) {
		    UDPDIAG(("Unable to connect, resetting\n"));
		    warning("Connection attempt failed");
		    commModeReq = COMM_TCP;
		    commStatus = STAT_CONNECTED;
		    if (udpSock >= 0)
			closeUdpConn(udpSock);
		    if (udpWin) {
			udprefresh(UDP_STATUS);
			udprefresh(UDP_CURRENT);
		    }
		    response.request = COMM_TCP;
		    response.port = 0;
		    goto send;
		}
#else
		/* this came down UDP, so we MUST be connected */
		/* (do the verify thing anyway just for kicks) */
#endif
		UDPDIAG(("Connected to server's UDP port\n"));
		commStatus = STAT_VERIFY_UDP;
		if (udpWin)
		    udprefresh(UDP_STATUS);
		response.request = COMM_VERIFY;	/* send verify request on UDP */
		response.port = 0;
		commSwitchTimeout = 25;	/* wait 25 updates */
send:
		sendServerPacket(&response);
	    }
	}
	break;
    case SWITCH_DENIED:
	if (ntohs(packet->port)) {
	    UDPDIAG(("Switch to UDP failed (different version)\n"));
	    warning("UDP protocol request failed (bad version)");
	} else {
	    UDPDIAG(("Switch to UDP denied\n"));
	    warning("UDP protocol request denied");
	}
	commModeReq = commMode;
	commStatus = STAT_CONNECTED;
	commSwitchTimeout = 0;
	if (udpWin)
	    udprefresh(UDP_STATUS);
	if (udpSock >= 0)
	    closeUdpConn();
	break;
    case SWITCH_VERIFY:
	UDPDIAG(("Received UDP verification\n"));
	break;
    default:
	fprintf(stderr, "netrek: Got funny reply (%d) in UDP_REPLY packet\n",
		packet->reply);
	break;
    }
}

#define MAX_PORT_RETRY	10
openUdpConn()
{
    struct sockaddr_in addr;
    struct hostent *hp;
    int attempts;

    if (udpSock >= 0) {
	fprintf(stderr, "netrek: tried to open udpSock twice\n");
	return (0);	/* pretend we succeeded (this could be bad) */
    }

    if ((udpSock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	perror("netrek: unable to create DGRAM socket");
	return (-1);
    }

    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_family = AF_INET;

    errno = 0;
    udpLocalPort = (getpid() & 32767) + (random() % 256);
    for (attempts = 0; attempts < MAX_PORT_RETRY; attempts++) {
	while (udpLocalPort < 2048) {
	    udpLocalPort = (udpLocalPort + 10687) & 32767;
	}
#ifdef GATEWAY
	/* we need the gateway to know where to find us */
	if (!strcmp(serverName, gw_mach)) {
	    UDPDIAG(("+ gateway test: binding to %d\n", gw_local_port));
	    udpLocalPort = gw_local_port;
	}
#endif
	addr.sin_port = htons(udpLocalPort);
	if (bind(udpSock, &addr, sizeof(addr)) >= 0)
	    break;
    }
    if (attempts == MAX_PORT_RETRY) {
	perror("netrek: bind");
	UDPDIAG(("Unable to find a local port to bind to\n"));
	close(udpSock);
	udpSock = -1;
	return (-1);
    }

    UDPDIAG(("Local port is %d\n", udpLocalPort));

    /* determine the address of the server */
    if (!serveraddr) {
	if ((addr.sin_addr.s_addr = inet_addr(serverName)) == -1) {
	    if ((hp = gethostbyname(serverName)) == NULL) {
		printf("Who is %s?\n", serverName);
		exit(0);
	    } else {
		addr.sin_addr.s_addr = *(long *) hp->h_addr;
	    }
	}
	serveraddr = addr.sin_addr.s_addr;
	UDPDIAG(("Found serveraddr == 0x%x\n", serveraddr));
    }
    return (0);
}

#ifdef USE_PORTSWAP
connUdpConn()
{
    struct sockaddr_in addr;

    addr.sin_addr.s_addr = serveraddr;
    addr.sin_family = AF_INET;
    addr.sin_port = htons(udpServerPort);

    UDPDIAG(("Connecting to host 0x%x on port %d\n", serveraddr,udpServerPort));
    if (connect(udpSock, &addr, sizeof(addr)) < 0) {
	fprintf(stderr, "Error %d: ");
	perror("netrek: unable to connect UDP socket");
	printUdpInfo();		/* debug */
	return (-1);
    }

    return (0);
}
#endif

#ifndef USE_PORTSWAP
recvUdpConn()
{
    fd_set readfds;
    struct timeval to;
    struct sockaddr_in from;
    struct sockaddr_in addr;
    int fromlen, res;

    bzero(&from, sizeof(from));		/* don't get garbage if really broken */

    /* we patiently wait until the server sends a packet to us */
    /* (note that we silently eat the first one) */
    UDPDIAG(("Issuing recvfrom() call\n"));
    printUdpInfo();
    fromlen = sizeof(from);
    FD_ZERO(&readfds);
    FD_SET(udpSock, &readfds);
    to.tv_sec = 3;		/* wait 3 seconds, then abort */
    to.tv_usec = 0;
    if ((res = select(32, &readfds, 0, 0, &to)) <= 0) {
	if (!res) {
	    UDPDIAG(("timed out waiting for response\n"));
	    warning("UDP connection request timed out");
	    return (-1);
	} else {
	    perror("select() before recvfrom()");
	    return (-1);
	}
    }
    if (recvfrom(udpSock, buf, BUFSIZ, 0, &from, &fromlen) < 0) {
        perror("recvfrom");
	UDPDIAG(("recvfrom failed, aborting UDP attempt"));
	return (-1);
    }

    if (from.sin_addr.s_addr != serveraddr) {
	UDPDIAG(("Warning: from 0x%x, but server is 0x%x\n",
		from.sin_addr.s_addr, serveraddr));
    }
    if (from.sin_family != AF_INET) {
	UDPDIAG(("Warning: not AF_INET (%d)\n", from.sin_family));
    }
    udpServerPort = ntohs(from.sin_port);
    UDPDIAG(("recvfrom() succeeded; will use server port %d\n", udpServerPort));
#ifdef GATEWAY
    if (!strcmp(serverName, gw_mach)) {
	UDPDIAG(("+ actually, I'm going to use %d\n", gw_port));
	udpServerPort = gw_port;
	from.sin_port = htons(udpServerPort);
    }
#endif

    if (connect(udpSock, &from, sizeof(from)) < 0) {
	perror("netrek: unable to connect UDP socket after recvfrom()");
	close(udpSock);
	udpSock = -1;
	return (-1);
    }

    return (0);
}
#endif

closeUdpConn()
{
    V_UDPDIAG(("Closing UDP socket\n"));
    if (udpSock < 0) {
	fprintf(stderr, "netrek: tried to close a closed UDP socket\n");
	return (-1);
    }
    shutdown(udpSock, 2);
    close(udpSock);
    udpSock = -1;
}

/* debugging info */
printUdpInfo()
{
    struct sockaddr_in addr;
    int len;

    len = sizeof(addr);
    if (getsockname(udpSock, &addr, &len) < 0) {
/*	perror("printUdpInfo: getsockname");*/
	return;
    }
    UDPDIAG(("LOCAL: addr=0x%x, family=%d, port=%d\n", addr.sin_addr.s_addr,
	addr.sin_family, ntohs(addr.sin_port)));

    if (getpeername(udpSock, &addr, &len) < 0) {
/*	perror("printUdpInfo: getpeername");*/
	return;
    }
    UDPDIAG(("PEER : addr=0x%x, family=%d, port=%d\n", addr.sin_addr.s_addr,
	addr.sin_family, ntohs(addr.sin_port)));
}

/* handles both SP_SEQUENCE and SP_SC_SEQUENCE */
handleSequence(packet)
struct sequence_spacket *packet;
{
    static int recent_count=0, recent_dropped=0;
    long newseq;

    drop_flag = 0;
    if (chan != udpSock)
	return;		/* don't pay attention to TCP sequence #s */

    udpTotal++;
    recent_count++;

    /* update percent display every 256 updates (~50 seconds usually) */
    if (!(udpTotal & 0xff))
	if (udpWin) udprefresh(UDP_DROPPED);

    newseq = (long) ntohs(packet->sequence);
/*    printf("read %d - ", newseq);*/

    if (((unsigned short) sequence) > 65000 &&
	((unsigned short) newseq) < 1000) {
	/* we rolled, set newseq = 65536+sequence and accept it */
	sequence = ((sequence + 65536) & 0xffff0000) | newseq;
    } else {
	/* adjust newseq and do compare */
	newseq |= (sequence & 0xffff0000);

	if (!udpSequenceChk) {	/* put this here so that turning seq check */
	    sequence = newseq;	/* on and off doesn't make us think we lost */
	    return;		/* a whole bunch of packets. */
	}

	if (newseq > sequence) {
	    /* accept */
	    if (newseq != sequence+1) {
		udpDropped += (newseq-sequence)-1;
		udpTotal += (newseq-sequence)-1;	/* want TOTAL packets */
		recent_dropped += (newseq-sequence)-1;
		recent_count += (newseq-sequence)-1;
		if (udpWin)
		    udprefresh(UDP_DROPPED);
		UDPDIAG(("sequence=%d, newseq=%d, we lost some\n",
		    sequence, newseq));
	    }
	    sequence = newseq;
	} else {
	    /* reject */
	    if (packet->type == SP_SC_SEQUENCE) {
		V_UDPDIAG(("(ignoring repeat %d)\n", newseq));
	    } else {
		UDPDIAG(("sequence=%d, newseq=%d, ignoring transmission\n",
		    sequence, newseq));
	    }
	    drop_flag = 1;
	}
    }
/*    printf("newseq %d, sequence %d\n", newseq, sequence);*/

    if (recent_count > UDP_RECENT_INTR) {
	/* once a minute (at 5 upd/sec), report on how many were dropped */
	/* during the last UDP_RECENT_INTR updates                       */
	udpRecentDropped = recent_dropped;
	recent_count = recent_dropped = 0;
	if (udpWin)
	    udprefresh(UDP_DROPPED);
    }
}

