/* x10window.c
 *
 * Kevin P. Smith  6/11/89
 */
#include "copyright2.h"
#include <stdio.h>
#include <X/Xlib.h>
#include <assert.h>
#include "Wlib.h"

Display *W_Display;
W_Font W_BigFont, W_RegularFont, W_HighlightFont, W_UnderlineFont;
W_Color W_White=0, W_Black=1, W_Red=2, W_Green=3, W_Yellow=4, W_Cyan=5;
W_Color W_Grey=6;
int W_Textwidth, W_Textheight;
char *getdefault();
char *strdup();

static W_Event W_myevent;
static int W_isEvent=0;

struct colors {
    char *name;
    int pixelValue;
    Pixmap pixmap;
};

struct icon {
    Window window;
    Bitmap bitmap;
    int width, height;
    Pixmap pixmap;
};

#define WIN_GRAPH	1
#define WIN_TEXT	2
#define WIN_MENU	3
#define WIN_SCROLL	4

struct window {
    Window window;
    int type;
    char *data;
    int mapped;
    int width,height;
    char *name;
};

struct stringList {
    char *string;
    W_Color color;
    struct stringList *next;
};

struct menuItem {
    char *string;
    int color;
};

struct colors colortable[] = {
    { "white",      0, 0},
    { "black",      0, 0},	
    { "red",        0, 0},
    { "green",      0, 0},
    { "yellow",     0, 0},
    { "cyan",  	    0, 0},
    { "light grey", 0, 0}
};

struct windowlist {
    struct window *window;
    struct windowlist *next;
};

#define HASHSIZE 29
#define hash(x) (((int) (x)) % HASHSIZE)

struct windowlist *hashtable[HASHSIZE];

struct window *newWindow();
struct window *findWindow();
char *malloc();
short *x11tox10bits();

struct window myroot;

#define NCOLORS (sizeof(colortable)/sizeof(colortable[0]))
#define W_Void2Window(win) ((win) ? (struct window *) (win) : &myroot)
#define W_Window2Void(window) ((W_Window) (window))
#define W_Void2Icon(bit) ((struct icon *) (bit))
#define W_Icon2Void(bit) ((W_Icon) (bit))
#define W_Void2Font(font) ((FontInfo *) (font))
#define W_Font2Void(font) ((W_Font) (font))
#define TILESIDE 16

#define WIN_EDGE 5	/* border on l/r edges of text windows */
#define MENU_PAD 4	/* border on t/b edges of text windows */
#define MENU_BAR 4	/* width of menu bar */

static short    gray[TILESIDE] = {
        0xaaaa, 0x5555, 0xaaaa, 0x5555,
        0xaaaa, 0x5555, 0xaaaa, 0x5555,
        0xaaaa, 0x5555, 0xaaaa, 0x5555,
        0xaaaa, 0x5555, 0xaaaa, 0x5555,
};

static short    striped[TILESIDE] = {
        0xff00, 0xff00, 0xff00, 0xff00,
        0x0ff0, 0x0ff0, 0x0ff0, 0x0ff0,
        0x00ff, 0x00ff, 0x00ff, 0x00ff,
        0xf00f, 0xf00f, 0xf00f, 0xf00f,
};

void W_Initialize(str)
char *str;
{
    int i;

#ifdef DEBUG
    printf("Initializing...\n");
#endif
    for (i=0; i<HASHSIZE; i++) {
	hashtable[i]=NULL;
    }
    if ((W_Display = XOpenDisplay(str)) == NULL) {
	fprintf(stderr, "I cannot open your display!\n");
	exit(1);
    }
    myroot.window=RootWindow;
    myroot.type=WIN_GRAPH;
    GetFonts();
    GetColors();
}

GetFonts()
{
    char *font;

    if ((font = getdefault("font")) == NULL) 
	font = "6x10";
    if ((W_RegularFont = W_Font2Void(XOpenFont(font))) == 0) {
	perror(font);
	exit(1);
    }
    W_Textwidth=6;		/* 6x10 font, right? */
    W_Textheight=10;

    if ((font = getdefault("boldfont")) == NULL) 
	font = "6x10b";
    if ((W_HighlightFont = W_Font2Void(XOpenFont(font))) == 0) 
	W_HighlightFont = W_RegularFont;

    if ((font = getdefault("italicFont")) == NULL) 
	font = "6x10i";
    if ((W_UnderlineFont = W_Font2Void(XOpenFont(font))) == 0) 
	W_UnderlineFont = W_RegularFont;
    
    if ((font = getdefault("bigFont")) == NULL) 
	font = "sbdr40sx";
    if ((W_BigFont = W_Font2Void(XOpenFont(font))) == 0) 
	W_BigFont = W_RegularFont;
}

GetColors()
{
    int i;
    Color def;
    char defaultstring[100];
    char *defaults;

    if (DisplayCells() <= 2) {
	for (i=0; i<NCOLORS; i++) {
	    colortable[i].pixelValue=WhitePixel;
	    colortable[i].pixmap=WhitePixmap;
	}
	colortable[W_Black].pixelValue=BlackPixel;
	colortable[W_Black].pixmap=BlackPixmap;
	colortable[W_Red].pixmap = XMakePixmap(
    	    XStoreBitmap(TILESIDE, TILESIDE, striped), WhitePixel, BlackPixel);
	colortable[W_Yellow].pixmap = XMakePixmap(
	    XStoreBitmap(TILESIDE, TILESIDE, gray), WhitePixel, BlackPixel);
    } else {
	for (i=0; i<NCOLORS; i++) {
	    sprintf(defaultstring, "color.%s", colortable[i].name);
	    defaults=getdefault(defaultstring);
	    if (!defaults) defaults=colortable[i].name;
	    XParseColor(defaults, &def);
	    XGetHardwareColor(&def);
	    colortable[i].pixelValue = def.pixel;
	    colortable[i].pixmap = XMakeTile(def.pixel);
	}
    }
}

W_Window W_MakeWindow(name,x,y,width,height,parent,border,color)
char *name;
int x,y,width,height;
W_Window parent;
int border;
W_Color color;
{
    struct window *newwin;
    Window wparent;
	
#ifdef DEBUG
    printf("New window...\n");
#endif
    checkGeometry(name, &x, &y, &width, &height);
    checkParent(name, &parent);
    wparent=W_Void2Window(parent)->window;
    newwin=newWindow(
	XCreateWindow(wparent, x, y, width, height, border, 
	colortable[color].pixmap, colortable[W_Black].pixmap),
	WIN_GRAPH);
    newwin->name=strdup(name);
    if (name!=NULL) XStoreName(newwin->window, name);
    newwin->width=width;
    newwin->height=height;
    XSelectInput(newwin->window, 
	KeyPressed|ButtonPressed|ExposeRegion|ExposeWindow);
    if (checkMapped(name)) W_MapWindow(W_Window2Void(newwin));
#ifdef DEBUG
    printf("New graphics window %d, child of %d\n", newwin, parent);
#endif
    return(W_Window2Void(newwin));
}

void W_ChangeBorder(window, color)
W_Window window;
int color;
{
#ifdef DEBUG
    printf("Changing border of %d\n", window);
#endif
    XChangeBorder(W_Void2Window(window)->window, colortable[color].pixmap);
}

void W_MapWindow(window)
W_Window window;
{
    struct window *win;

#ifdef DEBUG
    printf("Mapping %d\n", window);
#endif
    win=W_Void2Window(window);
    if (win->mapped) return;
    win->mapped=1;
    XMapWindow(win->window);
}

void W_UnmapWindow(window)
W_Window window;
{
    struct window *win;

#ifdef DEBUG
    printf("UnMapping %d\n", window);
#endif
    win=W_Void2Window(window);
    if (win->mapped==0) return;
    win->mapped=0;
    XUnmapWindow(win->window);
}

int W_IsMapped(window)
W_Window window;
{
    struct window *win;

    win=W_Void2Window(window);
    if (win==NULL) return(0);
    return(win->mapped);
}

void W_ClearArea(window, x, y, width, height, color)
W_Window window;
int x, y, width, height;
W_Color color;
{
    struct window *win;

#ifdef DEBUG
    printf("Clearing (%d %d) x (%d %d) with %d on %d\n", x,y,width,height,
	color,window);
#endif
    win=W_Void2Window(window);
    switch(win->type) {
    case WIN_GRAPH:
	XPixSet(win->window,x,y,width,height,colortable[color].pixelValue);
	break;
    default:
	XPixSet(win->window,WIN_EDGE+x*W_Textwidth,
	    MENU_PAD+y*W_Textheight,width*W_Textwidth,
	    height*W_Textheight,colortable[color].pixelValue);
	break;
    }
}

void W_ClearWindow(window)
W_Window window;
{
#ifdef DEBUG
    printf("Clearing %d\n", window);
#endif
    XClear(W_Void2Window(window)->window);
}

int W_EventsPending()
{
    if (W_isEvent) return(1);
    while (XPending()) {
	if (W_SpNextEvent(&W_myevent)) {
	    W_isEvent=1;
	    return(1);
	}
    }
    return(0);
}

void W_NextEvent(wevent)
W_Event *wevent;
{
    if (W_isEvent) {
	*wevent=W_myevent;
	W_isEvent=0;
	return;
    }
    while (W_SpNextEvent(wevent)==0);
}

int W_SpNextEvent(wevent)
W_Event *wevent;
{
    XEvent event;
    XKeyOrButtonEvent *tmp;
    XExposeWindowEvent *expose;
    char *str;
    int nchars;
    struct window *win;

#ifdef DEBUG
    printf("Getting an event...\n");
#endif
    tmp= (XKeyOrButtonEvent *) &event;
    expose= (XExposeWindowEvent *) &event;
    for (;;) {
	XNextEvent(&event);
	win=findWindow(event.window);
        if (win==NULL) return(0);
	if ((event.type==KeyPressed || event.type==ButtonPressed) &&
	    win->type == WIN_MENU) {
	    if (tmp->y % (W_Textheight + MENU_PAD*2+MENU_BAR) >= 
		W_Textheight+MENU_PAD*2) return(0);
	    tmp->y=tmp->y/(W_Textheight+MENU_PAD*2+MENU_BAR);
	}
	switch((int) event.type) {
	case KeyPressed:
	    str=XLookupMapping(&event, &nchars);
	    if (nchars>0) {
		wevent->type=W_EV_KEY;
		wevent->Window=W_Window2Void(win);
		wevent->x=tmp->x;
		wevent->y=tmp->y;
		wevent->key= *str;
		return(1);
	    }
	    return(0);
	    break;
	case ButtonPressed:
	    wevent->type=W_EV_BUTTON;
	    wevent->Window=W_Window2Void(win);
	    wevent->x=tmp->x;
	    wevent->y=tmp->y;
	    switch(tmp->detail & 0xf) {
	    case RightButton:
		wevent->key=W_RBUTTON;
		break;
	    case LeftButton:
		wevent->key=W_LBUTTON;
		break;
	    case MiddleButton:
		wevent->key=W_MBUTTON;
		break;
	    }
	    return(1);
	case ExposeWindow:
	    if (win->type == WIN_SCROLL) {
		resizeScrolling(win, expose->width, expose->height);
	    }
	case ExposeRegion:
	    if (win->type == WIN_SCROLL) {
		redrawScrolling(win);
		return(0);
	    }
	    if (win->type == WIN_MENU) {
		redrawMenu(win);
		return(0);
	    }
	    wevent->type=W_EV_EXPOSE;
	    wevent->Window=W_Window2Void(win);
	    return(1);
	default:
	    return(0);
	    break;
	}
    }
}

void W_MakeLine(window, x0, y0, x1, y1, color)
W_Window window;
int x0, y0, x1, y1;
W_Color color;
{
    Window win;

#ifdef DEBUG
    printf("Line on %d\n", window);
#endif
    win=W_Void2Window(window)->window;
    XLine(win, x0, y0, x1, y1, 1, 1, colortable[color].pixelValue, 
	GXcopy, AllPlanes);
}

void W_WriteText(window, x, y, color, str, len, font)
W_Window window;
int x, y,len;
W_Color color;
W_Font font;
char *str;
{
    struct window *win;

#ifdef DEBUG
    printf("Text for %d @ (%d, %d) in %d: [%s]\n", window, x,y,color,str);
#endif
    win=W_Void2Window(window);
    switch(win->type) {
    case WIN_GRAPH:
	XText(win->window, x, y, str, len, (W_Void2Font(font))->id, 
	    colortable[color].pixelValue, colortable[W_Black].pixelValue);
	break;
    case WIN_SCROLL:
	/* x and y and font ignored */
	XMoveArea(win->window, WIN_EDGE, MENU_PAD+W_Textheight, 
	    WIN_EDGE, MENU_PAD, win->width*W_Textwidth,
	    (win->height-1)*W_Textheight);
	XPixSet(win->window, WIN_EDGE, MENU_PAD+W_Textheight*(win->height-1), 
	    W_Textwidth*win->width, W_Textheight, 
	    colortable[W_Black].pixelValue);
	XText(win->window, WIN_EDGE, MENU_PAD+W_Textheight*(win->height-1), 
	    str, len,
	    (W_Void2Font(W_RegularFont))->id, colortable[color].pixelValue, 
	    colortable[W_Black].pixelValue);
	AddToScrolling(win, color, str, len);
	break;
    case WIN_MENU:
	/* x and font ignored */
	changeMenuItem(win, y, str, len, color);
	break;
    default:
	XText(win->window, x*W_Textwidth+WIN_EDGE, MENU_PAD+y*W_Textheight, 
	    str, len, 
	    (W_Void2Font(font))->id, colortable[color].pixelValue,
	    colortable[W_Black].pixelValue);
	break;
    }
}

void W_MaskText(window, x, y, color, str, len, font)
W_Window window;
int x, y,len;
W_Color color;
W_Font font;
char *str;
{
    struct window *win;

#ifdef DEBUG
    printf("TextMask for %d @ (%d, %d) in %d: [%s]\n", window, x,y,color,str);
#endif
    win=W_Void2Window(window);
    /* Handle only graphics windows.
     */
    XTextMask(win->window, x, y, str, len, (W_Void2Font(font))->id, 
	colortable[color].pixelValue);
}

W_Icon W_StoreBitmap(width, height, data, window)
int width, height;
W_Window window;
unsigned char *data;
{
    Bitmap newbit;
    struct icon *newicon;
    short *x10data;

#ifdef DEBUG
    printf("Storing bitmap for %d (%d x %d):  ", window,width,height);
    fflush(stdout);
#endif
    newicon=(struct icon *) malloc(sizeof(struct icon));
    newicon->width=width;
    newicon->height=height;
    x10data=x11tox10bits(width,height,data);
    newbit=XStoreBitmap(width, height, x10data);
#ifdef DEBUG
    printf("Stored\n");
#endif
    newicon->bitmap=newbit;
    newicon->window=W_Void2Window(window)->window;
    newicon->pixmap=0;
    free((char *) x10data);
    return(W_Icon2Void(newicon));
}

void W_WriteBitmap(x, y, bit, color)
int x,y;
W_Icon bit;
W_Color color;
{
    struct icon *icon;

    icon=W_Void2Icon(bit);
#ifdef DEBUG
    printf("Writing bitmap to %d\n", icon->window);
#endif
    XPixFill(icon->window, x, y, icon->width, icon->height, 
	colortable[color].pixelValue, icon->bitmap, GXcopy, AllPlanes);
}

void W_TileWindow(window, bit)
W_Window window;
W_Icon bit;
{
    Window win;
    struct icon *icon;

#ifdef DEBUG
    printf("Tiling window %d\n", window);
#endif
    icon=W_Void2Icon(bit);
    win=W_Void2Window(window)->window;

    if (icon->pixmap==0) {
	icon->pixmap=XMakePixmap(icon->bitmap, colortable[W_White].pixelValue,
	    colortable[W_Black].pixelValue);
    }
    XChangeBackground(win, icon->pixmap);
    XClear(win);
}

void W_UnTileWindow(window)
W_Window window;
{
    Window win;

#ifdef DEBUG
    printf("Untiling window %d\n", window);
#endif
    win=W_Void2Window(window)->window;
    XChangeBackground(win, colortable[W_Black].pixmap);
    XClear(win);
}

W_Window W_MakeTextWindow(name,x,y,width,height,parent,border)
char *name;
int x,y,width,height;
W_Window parent;
int border;
{
    struct window *newwin;
    Window wparent;
	
#ifdef DEBUG
    printf("New window...\n");
#endif
    checkGeometry(name, &x, &y, &width, &height);
    checkParent(name, &parent);
    wparent=W_Void2Window(parent)->window;
    newwin=newWindow(
	XCreateWindow(wparent, x, y, 
	width*W_Textwidth+WIN_EDGE*2, MENU_PAD*2+height*W_Textheight, border, 
	colortable[W_White].pixmap, colortable[W_Black].pixmap),
	WIN_TEXT);
    newwin->name=strdup(name);
    if (name!=NULL) XStoreName(newwin->window, name);
    newwin->width=width;
    newwin->height=height;
    XSelectInput(newwin->window, ExposeRegion|ExposeWindow);
    if (checkMapped(name)) W_MapWindow(W_Window2Void(newwin));
#ifdef DEBUG
    printf("New text window %d, child of %d\n", newwin, parent);
#endif
    return(W_Window2Void(newwin));
}

struct window *newWindow(window, type)
Window window;
int type;
{
    struct window *newwin;

    newwin=(struct window *) malloc(sizeof(struct window));
    newwin->window=window;
    newwin->type=type;
    newwin->mapped=0;
    addToHash(newwin);
    return(newwin);
}

struct window *findWindow(window)
Window window;
{
    struct windowlist *entry;

    entry=hashtable[hash(window)];
    while (entry!=NULL) {
	if (entry->window->window == window) return(entry->window);
	entry=entry->next;
    }
    return(NULL);
}

addToHash(win)
struct window *win;
{
    struct windowlist **new;

#ifdef DEBUG
    printf("Adding to %d\n", hash(win->window));
#endif
    new= &hashtable[hash(win->window)];
    while (*new != NULL) {
	new= &((*new)->next);
    }
    *new=(struct windowlist *) malloc(sizeof(struct windowlist));
    (*new)->next=NULL;
    (*new)->window=win;
}

W_Window W_MakeScrollingWindow(name,x,y,width,height,parent,border)
char *name;
int x,y,width,height;
W_Window parent;
int border;
{
    struct window *newwin;
    Window wparent;

#ifdef DEBUG
    printf("New window...\n");
#endif
    checkGeometry(name, &x, &y, &width, &height);
    checkParent(name, &parent);
    wparent=W_Void2Window(parent)->window;
    newwin=newWindow(
	XCreateWindow(wparent, x, y,
        width*W_Textwidth+WIN_EDGE*2, MENU_PAD*2+height*W_Textheight, border,
        colortable[W_White].pixmap, colortable[W_Black].pixmap),
        WIN_SCROLL);
    newwin->name=strdup(name);
    if (name!=NULL) XStoreName(newwin->window, name);
    XSelectInput(newwin->window, ExposeRegion|ExposeWindow);
    if (checkMapped(name)) W_MapWindow(W_Window2Void(newwin));
    newwin->data=NULL;
    newwin->width=width;
    newwin->height=height;
#ifdef DEBUG
    printf("New scroll window %d, child of %d\n", newwin, parent);
#endif
    return(W_Window2Void(newwin));
}

/* Add a string to the string list of the scrolling window. 
 */
AddToScrolling(win, color, str, len)
struct window *win;
W_Color color;
char *str;
int len;
{
    struct stringList **strings;
    char *newstring;
    int count;

    strings= (struct stringList **) &(win->data);
    count=0;
    while ((*strings)!=NULL) {
	count++;
	strings= &((*strings)->next);
    }
    (*strings)=(struct stringList *) malloc(sizeof(struct stringList));
    (*strings)->next=NULL;
    (*strings)->color=color;
    newstring=(char *) malloc(len+1);
    strncpy(newstring, str, len);
    newstring[len]=0;
    (*strings)->string=newstring;
    if (count >= 100) {		/* Arbitrary maximum size. */
	struct stringList *temp;

	temp=(struct stringList *) win->data;
	free(temp->string);
	temp=temp->next;
	free((char *) win->data);
	win->data=(char *) temp;
    }
}

redrawScrolling(win)
struct window *win;
{
    int count;
    struct stringList *list;
    int y;

    XClear(win->window);
    count=0;
    list = (struct stringList *) win->data;
    while (list != NULL) {
	list=list->next;
	count++;
    }
    list = (struct stringList *) win->data;
    while (count > win->height) {
	list=list->next;
	count--;
    }
    y=(win->height-count)*W_Textheight;
    if (count==0) return;
    while (list != NULL) {
	XText(win->window, WIN_EDGE, MENU_PAD+y, list->string, 
	    strlen(list->string),
	    (W_Void2Font(W_RegularFont))->id, 
	    colortable[list->color].pixelValue, colortable[W_Black].pixelValue);
	list=list->next;
	y=y+W_Textheight;
    }
}

resizeScrolling(win, width, height)
struct window *win;
int width, height;
{
    win->height=(height - MENU_PAD*2)/W_Textheight;
    win->width=(width - WIN_EDGE*2)/W_Textwidth;
}

W_Window W_MakeMenu(name,x,y,width,height,parent,border)
char *name;
int x,y,width,height;
W_Window parent;
int border;
{
    struct window *newwin;
    struct menuItem *items;
    Window wparent;
    int i;

#ifdef DEBUG
    printf("New window...\n");
#endif
    checkGeometry(name, &x, &y, &width, &height);
    checkParent(name, &parent);
    wparent=W_Void2Window(parent)->window;
    newwin=newWindow(
        XCreateWindow(wparent, x, y,
        width*W_Textwidth+WIN_EDGE*2, 
	height*(W_Textheight+MENU_PAD*2)+(height-1)*MENU_BAR, border,
        colortable[W_White].pixmap, colortable[W_Black].pixmap),
        WIN_MENU);
    newwin->name=strdup(name);
    if (name!=NULL) XStoreName(newwin->window, name);
    XSelectInput(newwin->window, 
	KeyPressed|ButtonPressed|ExposeRegion|ExposeWindow);
    if (checkMapped(name)) W_MapWindow(W_Window2Void(newwin));
    items=(struct menuItem *) malloc(height*sizeof(struct menuItem));
    for (i=0; i<height; i++) {
	items[i].string=NULL;
	items[i].color=W_White;
    }
    newwin->data=(char *) items;
    newwin->width=width;
    newwin->height=height;
#ifdef DEBUG
    printf("New menu window %d, child of %d\n", newwin, parent);
#endif
    return(W_Window2Void(newwin));
}

redrawMenu(win)
struct window *win;
{
    int count;

    for (count=1; count<win->height; count++) {
	XPixSet(win->window, 0, 
	    count*(W_Textheight+MENU_PAD*2)+(count-1)*MENU_BAR, 
	    win->width*W_Textwidth+WIN_EDGE*2, 
		MENU_BAR, colortable[W_White].pixelValue);
    }
    for (count=0; count<win->height; count++) {
	redrawMenuItem(win,count);
    }
}

redrawMenuItem(win, n)
struct window *win;
int n;
{
    struct menuItem *items;

    items=(struct menuItem *) win->data;
    XPixSet(win->window, WIN_EDGE, 
	n*(W_Textheight+MENU_PAD*2+MENU_BAR)+MENU_PAD,
	win->width*W_Textwidth,
	W_Textheight, colortable[W_Black].pixelValue);
    if (items[n].string) {
	XText(win->window, WIN_EDGE, 
	    n*(W_Textheight+MENU_PAD*2+MENU_BAR)+MENU_PAD,
	    items[n].string, strlen(items[n].string),
	    (W_Void2Font(W_RegularFont)->id),
	    colortable[items[n].color].pixelValue, 
	    colortable[W_Black].pixelValue);
    }
}

changeMenuItem(win, n, str, len, color)
struct window *win;
int n;
char *str;
int len;
W_Color color;
{
    struct menuItem *items;
    char *news;

    items=(struct menuItem *) win->data;
    if (items[n].string) {
	free(items[n].string);
    }
    news=malloc(len+1);
    strncpy(news,str,len);
    news[len]=0;
    items[n].string=news;
    items[n].color=color;
    redrawMenuItem(win, n);
}

void W_DefineCursor(window,width,height,bits,mask,xhot,yhot)
W_Window window;
int width, height, xhot, yhot;
unsigned char *bits, *mask;
{
    static unsigned char *oldbits=NULL;
    static Cursor curs;
    short *curdata;
    short *maskdata;
    struct window *win;

#ifdef DEBUG
    printf("Defining cursor for %d\n", window);
#endif
    win=W_Void2Window(window);
    if (!oldbits || oldbits!=bits) {
	oldbits=bits;
	curdata=x11tox10bits(width,height,bits);
	maskdata=x11tox10bits(width,height,mask);
	curs=XCreateCursor(width, height, curdata, maskdata,
	    xhot, yhot, colortable[W_White].pixelValue, 
	    colortable[W_Black].pixelValue, GXcopy);
	free((char *) curdata);
	free((char *) maskdata);
    }
    XDefineCursor(win->window, curs);
}

short *x11tox10bits(width, height, bits)
int width, height;
unsigned char *bits;
{
    short *x10data;
    int bytewidth;
    short *d;
    int i,j;

    x10data=(short *) malloc(sizeof(short) * height * ((width+15)/16));
    bytewidth=(width+7)/8;
    d=x10data;
    for (i=0; i<height; i++) {
	for (j=0; j<bytewidth/2; j++) {
	    *d = (*bits) | (*(bits+1) << 8);
	    bits+=2;
	    d++;
	}
	if (bytewidth & 1) {
	    *d = (*bits);
	    d++;
	    bits++;
	}
    }
    return(x10data);
}

void W_Beep()
{
    XFeep(0);
}

int W_WindowWidth(window)
W_Window window;
{
    return(W_Void2Window(window)->width);
}

int W_WindowHeight(window)
W_Window window;
{
    return(W_Void2Window(window)->height);
}

int W_Socket()
{
    return(dpyno());
}

void W_DestroyWindow(window)
W_Window window;
{
    struct window *win;
#ifdef DEBUG
    printf("Destroying %d\n", window);
#endif
    
    win=W_Void2Window(window);
    deleteWindow(win);
    XDestroyWindow(win->window);
    /* BUG: Free other stuff for scrolling and menu windows. 
     */
    free((char *) win);
}

deleteWindow(window)
struct window *window;
{
    struct windowlist **rm;
    struct windowlist *temp;

    rm= &hashtable[hash(window->window)];
    while (*rm != NULL && (*rm)->window!=window) {
	rm= &((*rm)->next);
    }
    if (*rm==NULL) {
	printf("Attempt to delete non-existent window!\n");
	return;
    }
    temp= *rm;
    *rm=temp->next;
    free((char *) temp);
}

void W_SetIconWindow(main, icon)
W_Window main;
W_Window icon;
{
    XSetIconWindow(W_Void2Window(main)->window, W_Void2Window(icon)->window);
}

checkGeometry(name, x, y, width, height)
char *name;
int *x, *y, *width, *height;
{
    char *adefault;
    char buf[100];
    char *s;

    sprintf(buf, "%s.geometry", name);
    adefault=getdefault(buf);
    if (adefault==NULL) return;
    /* geometry should be of the form 502x885+1+1, 502x885, or +1+1 */
    s=adefault;
    if (*s!='+') {
	while (*s!='x' && *s!=0) s++;
	*width = atoi(adefault);
	if (*s==0) return;
	s++;
	adefault=s;
	while (*s!='+' && *s!=0) s++;
	*height = atoi(adefault);
	if (*s==0) return;
    }
    s++;
    adefault=s;
    while (*s!='+' && *s!=0) s++;
    *x = atoi(adefault);
    if (*s==0) return;
    s++;
    adefault=s;
    *y = atoi(adefault);
    return;
}

checkParent(name, parent)
char *name;
W_Window *parent;
{
    char *adefault;
    char buf[100];
    int i;
    struct windowlist *windows;

    sprintf(buf, "%s.parent", name);
    adefault=getdefault(buf);
    if (adefault==NULL) return;
    /* parent must be name of other window or "root" */
    if (strcmpi(adefault, "root")==0) {
	*parent=W_Window2Void(&myroot);
	return;
    }
    for (i=0; i<HASHSIZE; i++) {
	windows=hashtable[i];
	while (windows != NULL) {
	    if (strcmpi(adefault, windows->window->name)==0) {
		*parent=W_Window2Void(windows->window);
		return;
	    }
	    windows=windows->next;
	}
    }
}

checkMapped(name)
char *name;
{
    char *adefault;
    char buf[100];

    sprintf(buf, "%s.mapped", name);
    return(booleanDefault(buf, 0));
}
