static char rcsid[] = "$Header: /usr/users/terence/src/xtrek/ntserv/RCS/orbit.c,v 1.2 92/03/19 21:34:37 terence Exp $";

/* $Log:	orbit.c,v $
 * Revision 1.2  92/03/19  21:34:37  terence
 * Added UDP mods (version 0.2, 3/17/92)
 * 
 * Revision 1.1  92/03/11  15:54:51  terence
 * Initial revision
 *  */

/*
 * orbit.c
 */
#include "copyright.h"

#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include "defs.h"
#include "struct.h"
#include "data.h"

/* orbit a planet */

orbit()
{
    register int i;
    register struct planet *l;
    unsigned char dir;
    int dx, dy;

    if (me->p_flags & PFORBIT) return; /* bug fix: no reverse/rapid orbiting
                                          6/23/92 TMC */
    if (me->p_speed > ORBSPEED ) {
	warning("Helmsman: Captain, the maximum safe speed for docking or orbiting is warp 2!");
	return;
    }

    me->p_flags &= ~(PFORBIT | PFPLOCK | PFPLLOCK | PFTRACT | PFPRESS);

    if (me->p_flags & PFDOCK) {
	players[me->p_docked].p_docked--;
	players[me->p_docked].p_port[me->p_port[0]] = VACANT;
	me->p_flags &= ~PFDOCK;
    }

    if (me->p_ship.s_type != STARBASE) {
	for (i=0; i<MAXPLAYER; i++) {
	    if (me->p_no == i) 
		continue;
	    if (players[i].p_ship.s_type != STARBASE)
		continue;
	    if (!isAlive(&players[i]))
		continue;
	    if (!friendlyPlayer(&players[i]))
		continue;
	    dx = ABS(players[i].p_x - me->p_x);
	    dy = ABS(players[i].p_y - me->p_y);
	    if (dx > DOCKDIST || dy > DOCKDIST)
		continue;
		
	    dock(i);
	    return;
	}
    }

    for (i = 0, l = &planets[i]; i < MAXPLANETS; i++, l++) {
	dx = ABS(l->pl_x - me->p_x);
	dy = ABS(l->pl_y - me->p_y);
	if (dx > ENTORBDIST || dy > ENTORBDIST)		/*XXX*/
	    continue;
	if (dx * dx + dy * dy > ENTORBDIST * ENTORBDIST)
	    continue;
	if (!(me->p_team & planets[i].pl_owner) && 
	        me->p_ship.s_type==STARBASE) {
	    warning("Central Command regulations prohibits you from orbiting foreign planets");
	    return;
	}

	dir = (unsigned char) (atan2((double) (me->p_x - l->pl_x),
				     (double) (l->pl_y - me->p_y))
			       / 3.14159 * 128.);
	l->pl_info |= me->p_team;
	me->p_dir = dir + 64;
	me->p_flags |= PFORBIT;
	me->p_x = l->pl_x + ORBDIST * Cos[dir];
	me->p_y = l->pl_y + ORBDIST * Sin[dir];
	me->p_speed = me->p_desspeed = 0;
	me->p_planet = l->pl_no;
	return;
    }

    warning("Helmsman:  Sensors read no valid targets in range to dock or orbit sir!");
}


dock(base_id)
     int base_id;
{
    int i, port_id=NUMPORTS;
    char buf[80];		/* bug fix 3/11/92 TC*/

    if (!(players[base_id].p_flags & PFDOCKOK)) {
	sprintf(buf, "Starbase %s refusing us docking permission captain.", 
	    players[base_id].p_name);
	warning(buf);
	return;
    }

    /* Try to dock on closest port, else pick first vacant port. */
    if ((me->p_x < players[base_id].p_x && me->p_y < players[base_id].p_y) && (players[base_id].p_port[3] == VACANT)) { 
	port_id = 3;
    } else if ((me->p_x < players[base_id].p_x && me->p_y > players[base_id].p_y) && (players[base_id].p_port[2] == VACANT)) {
	port_id = 2;
    } else if ((me->p_x > players[base_id].p_x && me->p_y > players[base_id].p_y) && (players[base_id].p_port[1] == VACANT)) {
	port_id = 1;
    } else if ((me->p_x > players[base_id].p_x && me->p_y < players[base_id].p_y) && (players[base_id].p_port[0] == VACANT)) {
	port_id = 0;
    } else if (players[base_id].p_port[0] == VACANT) {
	port_id = 0;
    } else if (players[base_id].p_port[1] == VACANT) {
	port_id = 1;
    } else if (players[base_id].p_port[2] == VACANT) {
	port_id = 2;
    } else  if (players[base_id].p_port[3] == VACANT) {
	port_id = 3;
    }

    if (port_id < NUMPORTS) {
	me->p_x = players[base_id].p_x + DOCKDIST * Cos[(port_id*90+45)*255/360];
	me->p_y = players[base_id].p_y + DOCKDIST * Sin[(port_id*90+45)*255/360];
	me->p_speed = 0;
	me->p_desspeed = 0;
	me->p_dir = (90 + port_id*90 + 45) * 255 / 360;
	me->p_desdir = me->p_dir;

	me->p_flags |= PFDOCK;
	players[base_id].p_docked++;
	players[base_id].p_port[port_id] = me->p_no;
	me->p_docked = base_id;
	me->p_port[0] = port_id;

	sprintf(buf, "Helmsman:  Docking manuever completed Captain.  All moorings secured at port %d.", port_id);
	warning(buf);
    } else {
	sprintf(buf, "Starbase %s: Permission to dock denied, all ports currently occupied.", players[base_id].p_name);
	warning(buf);
    }
}  /* dock() */
