/* robotII.c
 */
#include "copyright.h"

#include <X11/Xlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <setjmp.h>
#include "defs.h"
#include "struct.h"
#include "data.h"

Window mapw;				/* XXX */

struct itimerval udt;
extern int redrawall;		/* maint: missing "extern" 6/22/92 TC */
extern int lastm;		/* maint: missing "extern" 6/22/92 TC */

/* lots of neat flags */
int hostile;
int debug;
int level;
int fleet;
int sticky;
int berserk;
int practice;
int nofuel;
int polymorphic;		/* match opponent's ship class 8/15/91 TC */

int target;			/* Terminator's target 7/27/91 TC */
int phrange;			/* phaser range 7/31/91 TC */
int trrange;			/* tractor range 8/2/91 TC */

/* velocities 8/9/91 TC */

int dogslow;			/* dodge speed (hard turn) */
int dogfast;			/* dodge speed (soft turn) */
int runslow;			/* run speed (flat etemp) */
int runfast;			/* run speed (gain etemp) */
int closeslow;			/* approach speed (hard turn) */
int closefast;			/* approach speed (soft turn) */

/* char *rnames[4] = { "M5", "Colossus", "Guardian", "HAL"};*/
char *rnames[6] = { "M5", "Colossus", "Guardian", "HAL", "HunterKiller",
		    "TERMINATOR"};

main(argc, argv)
int argc;
char **argv;
{
    register int i;
    int rmove();
    int team = -1;
    int bteam;
    int pno;
    char str[80];
    int class;			/* ship class 8/9/91 TC */
    
    class = BATTLESHIP;
    target = -1;		/* no target 7/27/91 TC */
    for( ; argc>1 && argv[1][0]=='-'; argc--,argv++) {
	switch(argv[1][1]) {
	  case 'F':
	    nofuel++;
	    break;
	  case 'P':
	    polymorphic++;
	    break;
	  case 'f':
	    fleet++;
	    break;
	  case 's':
	    sticky++;
	    break;
	  case 'd':
	    debug++;
	    break;
	  case 'h':
	    hostile++;
	    break;
	  case 'p':
	    practice++;
	    break;
	  case 'b':
	    berserk++;
	    break;
	  case 'l':
	    if (argv[1][2] != NULL)
		level = atoi(&argv[1][2]);
	    else
		level = 100;
	    break;
	  case 'c':		/* ship class */
	    switch (argv[1][2]) {
	      case 'a': class = ASSAULT; break;
	      case 'b': class = BATTLESHIP; break;
	      case 'c': class = CRUISER; break;
	      case 'd': class = DESTROYER; break;
	      case 's': class = SCOUT; break;
	      case 'o': class = STARBASE; break; /* not fully implemented */
	      default:
		fprintf(stderr, "Unknown ship class, must be one of [abcdsoA].\n");
		exit(1);
		break;
	    }
	    break;
	  case 'T':		/* team */
	    switch (argv[1][2]) {
	      case 'f':
		team = 0;
		bteam = FED;
		break;
	      case 'r':
		team = 1;
		bteam = ROM;
		break;
	      case 'k':
		team = 2;
		bteam = KLI;
		break;
	      case 'o':
		team = 3;
		bteam = ORI;
		break;
	      case 'n':	/* change 5/10/91 TC neutral */
	      case 'i':
		team = 4;
		bteam = FED|ROM|KLI|ORI; /* don't like anybody */
		break;
	      case 't': {
		  char c;
		  c = argv[1][3];
		  
		  team = 5;
		  bteam = FED|ROM|KLI|ORI; /* don't like anybody */
		  target = -1;
		  if (c == '\0') {
		      fprintf(stderr, "Must specify target.  e.g. -Tt3.\n");
		      exit(1);
		  }
		  if ((c >= '0') && (c <= '9'))
		      target = c - '0';
		  else if ((c >= 'a') && (c <= 'j'))
		      target = c - 'a' + 10;
		  else {
		      fprintf(stderr, "Must specify target.  e.g. -Tt3.\n");
		      exit(1);
		  }
	      }			/* end case 't' */
		break;
	      default:
		fprintf(stderr, "Unknown team type.  Usage -Tx where x is [frko]\n");
		exit(1);
	    }			/* end switch argv */
	    break;
	  default:
	    fprintf(stderr, "Unknown option '%c' (must be one of: bcdfpsFPT)\n", argv[1][1]);
	    exit(1);
	}			/* end switch argv[1][1] */
	
	
    }				/* end for */
    srandom(getpid() * time((long *) 0));
    readsysdefaults();
    if (team < 0 || team >= 6) { /* change 7/27/91 TC was 4 now 6 */
	if (debug)
	    fprintf(stderr, "Choosing random team.\n");
	team = random() % 4;
    }
    openmem();
    pno = findrslot();  /* Robots are entitled to tester's slots. */
    me = &players[pno];
    me->p_no=pno;
    myship = &me->p_ship;
    mystats = &me->p_stats;
    lastm = mctl->mc_current;
    /* At this point we have memory set up.  If we aren't a fleet, we don't
       want to replace any other robots on this team, so we'll check the
       other players and get out if there are any on our team.
    */

    if (practice) {
	strcpy(pseudo, "Hoser");
    } else {
	strcpy(pseudo, rnames[team]);
    }
    strcpy(login, "Robot");

    (void) strncpy(me->p_login, login, sizeof (me->p_login));
    me->p_login[sizeof(me->p_login) - 1] = '\0';
    (void) strncpy(me->p_monitor, "Nowhere", sizeof(me->p_monitor));
    me->p_monitor[sizeof(me->p_monitor) - 1] = '\0';

#ifdef FULL_HOSTNAMES
    /* repeat "Nowhere" for completeness 4/13/92 TC */
    (void) strncpy(me->p_full_hostname, "Nowhere", sizeof(me->p_full_hostname));
    me->p_full_hostname[sizeof(me->p_full_hostname) - 1] = '\0';
#endif

    if (target >= 0)		/* hack 7/27/91 TC */
	enter(team, target, pno, class); /* was BATTLESHIP 8/9/91 TC */
    else
	enter(team, 0, pno, class); /* was BATTLESHIP 8/9/91 TC */

    me->p_pos = -1;			/* So robot stats don't get saved */
    me->p_flags |= PFROBOT;			/* Mark as a robot */
    if ((berserk) || (team == 4)) /* indeps are hostile */
	me->p_hostile = (FED|ROM|KLI|ORI);	/* unless they are berserk */
    else if (team == 5)
	me->p_hostile = 0;	/* Termies declare war later */
    else if (practice)
	me->p_hostile = bteam;			/* or practice */
    else
	me->p_hostile = 0;			/* robots are peaceful */

    if (practice)
	me->p_flags |= PFPRACTR;		/* Mark as a practice robot */
    signal(SIGALRM, rmove);

    config();
    if (practice) {
	udt.it_interval.tv_sec = 1;		/* Robots get to move 1/sec */
	udt.it_interval.tv_usec = 000000;
    }
    else {
	udt.it_interval.tv_sec = 0;		/* Robots get to move 2/sec */
	udt.it_interval.tv_usec = 500000;
    }
    udt.it_value.tv_sec = 1;
    udt.it_value.tv_usec = 0;
    if (setitimer(ITIMER_REAL, &udt, 0) < 0) {
	perror("setitimer");
	me->p_status = PFREE;		/* Put robot in game */
	exit(1);
    }
    /* allows robots to be forked by the daemon -- Evil ultrix bullshit */
    sigsetmask(0);

    if (team == 4) { 
	int count = 0;
	for (i = 0; i < MAXPLAYER; i++) {
	    if (players[i].p_status == PALIVE)
		count++;
	}
	if (((count <= 1) && (!fleet)) ||
	    ((count < 1) && (fleet))) {	/* if one or less players, don't show */
	    if (debug)
		fprintf(stderr, "No one to hose.\n");
	    players[pno].p_status = PFREE;
	    exit(1);
	}
    }
    
    
    if (!fleet) { 
	for (i = 0; i < MAXPLAYER; i++) {
	    if ((players[i].p_status == PALIVE) && (players[i].p_team == bteam)) {
		if (debug)
		    fprintf(stderr, "Galaxy already defended\n");
		players[pno].p_status = PFREE;
		exit(1);
	    }
	}
    }

    me->p_status = PALIVE;		/* Put robot in game */
    while (1) {
	pause();
    }
}

findrslot()
{
    register int i;

/*    for (i = 0; i < MAXPLAYER; i++) { */
    for (i = MAXPLAYER-1; i >= 0; i--) {
	/* reverse entry..let's be trendy 4/6/92 TC */
        if (players[i].p_status == PFREE) {     /* We have a free slot */
            players[i].p_status = POUTFIT;      /* possible race code */
            break;
        }
    }
    if ((i == MAXPLAYER) || (i == -1)) {
	if (debug) {
	    fprintf(stderr, "No more room in game\n");
	}
        exit (0);
    }
    bzero(&players[i].p_stats, sizeof(struct stats));  
    players[i].p_stats.st_tticks=1;
    return(i);
}

warning(mess)
int mess;
{
    if (debug)
	fprintf(stderr, "warning: %s\n", mess);
}

config()
{
    /* calc class-specific stuff */

    phrange = PHASEDIST * me->p_ship.s_phaserdamage / 100;
    trrange = TRACTDIST * me->p_ship.s_tractrng;

    switch (myship->s_type) {
      case SCOUT:
	dogslow = 5;
	dogfast = 7;
	runslow = 8;
	runfast = 9;
	closeslow = 5;
	closefast = 7;
	break;
      case DESTROYER:
	dogslow = 4;
	dogfast = 6;
	runslow = 6;
	runfast = 8;
	closeslow = 4;
	closefast = 6;
	break;
      case CRUISER:
	dogslow = 4;
	dogfast = 6;
	runslow = 6;
	runfast = 7;
	closeslow = 4;
	closefast = 6;
	break;
      case BATTLESHIP:
	dogslow = 3;
	dogfast = 5;
	runslow = 5;
	runfast = 6;
	closeslow = 3;
	closefast = 4;
	break;
      case ASSAULT:
	dogslow = 3;
	dogfast = 5;
	runslow = 6;
	runfast = 7;
	closeslow = 3;
	closefast = 4;
	break;
      case STARBASE:
	dogslow = 2;
	dogfast = 2;
	runslow = 2;
	runfast = 2;
	closeslow = 2;
	closefast = 2;
	break;
    }

    if (debug) printf("My phaser range: %d.\n", phrange);
    if (debug) printf("My tractor range: %d.\n", trrange);

    if (!nofuel) {
	myship->s_phasercost = 0;
	myship->s_torpcost = 0;
	myship->s_cloakcost = 0;
    }
    if (target >= 0) {		/* 7/27/91 TC */
	myship->s_maxspeed = 10; /* so you can't run */
	myship->s_warpcost = 1;
	myship->s_egncoolrate = 100;
    }
}
