static char rcsid[] = "$Header: /usr/users/terence/src/xtrek/ntserv/RCS/socket.c,v 1.7 92/04/08 10:08:34 terence Exp $";

/* $Log:	socket.c,v $
 * Revision 1.7  92/04/08  10:08:34  terence
 * Added UDP v1.0 mods.
 * 
 * Revision 1.6  92/04/02  15:44:17  terence
 * fixed 't-mode' stuck on bug
 * 
 * Revision 1.5  92/04/01  16:53:53  terence
 * removed #ifdefs corresponding to INCLUDE_SCAN and INCLUDE_VISTRACT
 * 
 * Revision 1.4  92/04/01  15:35:44  terence
 * UDP v0.3 mods added
 * 
 * Revision 1.3  92/03/19  21:34:01  terence
 * Added UDP mods (version 0.2, 3/17/92)
 * 
 * Revision 1.2  92/03/11  16:32:49  terence
 * added UDP mods (huge)
 *  */

/*
 * Socket.c 
 *
 * Kevin P. Smith  1/26/89
 * UDP stuff v1.0 by Andy McFadden  Feb-Apr 1992
 *
 * UDP protocol v1.0
 *
 * Provides all of the support for sending packets to the client.
 */
#include "copyright2.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <math.h>
#include <errno.h>
#include "defs.h"
#include "struct.h"
#include "data.h"
#include "packets.h"

#ifdef OOPS
#ifdef MIPSEL
#ifndef htonl

#define htonl(x) ((u_long) (((u_long) x << 24) | (((u_long) x & 0xff00) << 8) | \
		  (((u_long) x & 0xff0000) >> 8) | ((u_long) x >> 24)))
#define ntohl(x) htonl(x)
#define htons(x) ((u_short) (((u_short) x << 8) | ((u_short) x >> 8)))
#define ntohs(x) htons(x)

#endif htonl
#endif MIPSEL
#endif OOPS

/*#define DOUBLE_UDP		/* comment this out to disable it */

/*#define BROKEN	/* for local stress testing; drops 20% of packets */
/*#define HOSED		/* combine with BROKEN; drops 90% of packets */
/*#define FATDIAG	/* define to get LOTS of fat UDP debugging messages */

int handleTorpReq(), handlePhasReq(), handleSpeedReq();
int handleDirReq(), handleShieldReq(), handleRepairReq(), handleOrbitReq();
int handlePractrReq(), handleBombReq(), handleBeamReq(), handleCloakReq();
int handleDetTReq(), handleCopilotReq();
int handleOutfit(), handleLoginReq();
int handlePlasmaReq(), handleWarReq(), handlePlanlockReq();
int handlePlaylockReq(), handleDetMReq();
int handleTractorReq(), handleRepressReq();
int handleCoupReq(), handleRefitReq(), handleMessageReq();
int handleQuitReq(), handleOptionsPacket();
int handleSocketReq(), handleByeReq();
int handleDockingReq(), handleReset();
int handleUpdatesReq(), handleReserved();
int handleScan();
int handleUdpReq(), handleSequence();
int exit();

static int remoteaddr= -1;		/* inet address in net format */

struct packet_handler handlers[] = {
    { 0, NULL},		/* record 0 */
    { sizeof(struct mesg_cpacket), handleMessageReq },	   /* CP_MESSAGE */
    { sizeof(struct speed_cpacket), handleSpeedReq },	   /* CP_SPEED */
    { sizeof(struct dir_cpacket), handleDirReq },	   /* CP_DIRECTION */
    { sizeof(struct phaser_cpacket), handlePhasReq },	   /* CP_PHASER */
    { sizeof(struct plasma_cpacket), handlePlasmaReq },	   /* CP_PLASMA */
    { sizeof(struct torp_cpacket), handleTorpReq },	   /* CP_TORP */
    { sizeof(struct quit_cpacket), handleQuitReq },	   /* CP_QUIT */
    { sizeof(struct login_cpacket), handleLoginReq },	   /* CP_LOGIN */
    { sizeof(struct outfit_cpacket), handleOutfit },	   /* CP_OUTFIT */
    { sizeof(struct war_cpacket), handleWarReq },	   /* CP_WAR */
    { sizeof(struct practr_cpacket), handlePractrReq },	   /* CP_PRACTR */
    { sizeof(struct shield_cpacket), handleShieldReq },	   /* CP_SHIELD */
    { sizeof(struct repair_cpacket), handleRepairReq },	   /* CP_REPAIR */
    { sizeof(struct orbit_cpacket), handleOrbitReq },	   /* CP_ORBIT */
    { sizeof(struct planlock_cpacket), handlePlanlockReq },/* CP_PLANLOCK */
    { sizeof(struct playlock_cpacket), handlePlaylockReq },/* CP_PLAYLOCK */
    { sizeof(struct bomb_cpacket), handleBombReq },	   /* CP_BOMB */
    { sizeof(struct beam_cpacket), handleBeamReq },	   /* CP_BEAM */
    { sizeof(struct cloak_cpacket), handleCloakReq },	   /* CP_CLOAK */
    { sizeof(struct det_torps_cpacket), handleDetTReq },   /* CP_DET_TORPS */
    { sizeof(struct det_mytorp_cpacket), handleDetMReq },  /* CP_DET_MYTORP */
    { sizeof(struct copilot_cpacket), handleCopilotReq },  /* CP_COPLIOT */
    { sizeof(struct refit_cpacket), handleRefitReq }, 	   /* CP_REFIT */
    { sizeof(struct tractor_cpacket), handleTractorReq },  /* CP_TRACTOR */
    { sizeof(struct repress_cpacket), handleRepressReq },  /* CP_REPRESS */
    { sizeof(struct coup_cpacket), handleCoupReq },	   /* CP_COUP */
    { sizeof(struct socket_cpacket), handleSocketReq },	   /* CP_SOCKET */
    { sizeof(struct options_cpacket), handleOptionsPacket },/* CP_OPTIONS */
    { sizeof(struct bye_cpacket), handleByeReq },	   /* CP_BYE */
    { sizeof(struct dockperm_cpacket), handleDockingReq }, /* CP_DOCKPERM */
    { sizeof(struct updates_cpacket), handleUpdatesReq 	}, /* CP_UPDATES */
    { sizeof(struct resetstats_cpacket), handleReset },    /* CP_RESETSTATS */
    { sizeof(struct reserved_cpacket), handleReserved },   /* CP_RESERVED */
    { sizeof(struct scan_cpacket), handleScan },	   /* CP_SCAN (ATM) */
    { sizeof(struct udp_req_cpacket), handleUdpReq },	   /* CP_UDP_REQ */
    { sizeof(struct sequence_cpacket), handleSequence }	   /* CP_SEQUENCE */
};

int sizes[] = {
    0,	/* record 0 */
    sizeof(struct mesg_spacket),		/* SP_MESSAGE */
    sizeof(struct plyr_info_spacket), 		/* SP_PLAYER_INFO */
    sizeof(struct kills_spacket),		/* SP_KILLS */
    sizeof(struct player_spacket),		/* SP_PLAYER */
    sizeof(struct torp_info_spacket),		/* SP_TORP_INFO */
    sizeof(struct torp_spacket), 		/* SP_TORP */
    sizeof(struct phaser_spacket),		/* SP_PHASER */
    sizeof(struct plasma_info_spacket),		/* SP_PLASMA_INFO */
    sizeof(struct plasma_spacket),		/* SP_PLASMA */
    sizeof(struct warning_spacket),		/* SP_WARNING */
    sizeof(struct motd_spacket),		/* SP_MOTD */
    sizeof(struct you_spacket),			/* SP_YOU */
    sizeof(struct queue_spacket),		/* SP_QUEUE */
    sizeof(struct status_spacket),		/* SP_STATUS */
    sizeof(struct planet_spacket),		/* SP_PLANET */
    sizeof(struct pickok_spacket),		/* SP_PICKOK */
    sizeof(struct login_spacket),		/* SP_LOGIN */
    sizeof(struct flags_spacket),		/* SP_FLAGS */
    sizeof(struct mask_spacket),		/* SP_MASK */
    sizeof(struct pstatus_spacket),		/* SP_PSTATUS */
    sizeof(struct badversion_spacket),		/* SP_BADVERSION */
    sizeof(struct hostile_spacket),		/* SP_HOSTILE */
    sizeof(struct stats_spacket),		/* SP_STATS */
    sizeof(struct plyr_login_spacket),		/* SP_PL_LOGIN */
    sizeof(struct reserved_spacket), 		/* SP_RESERVED */
    sizeof(struct planet_loc_spacket),		/* SP_PLANET_LOC */
    sizeof(struct scan_spacket),		/* SP_SCAN (ATM) */
    sizeof(struct udp_reply_spacket),		/* SP_UDP_REPLY */
    sizeof(struct sequence_spacket),		/* SP_SEQUENCE */
    sizeof(struct sc_sequence_spacket)		/* SP_SC_SEQUENCE */
};

#define NUM_PACKETS (sizeof(handlers) / sizeof(handlers[0]) - 1)
#define NUM_SIZES (sizeof(sizes) / sizeof(sizes[0]) - 1)

int packetsReceived[NUM_PACKETS+1] = { 0 };
int packetsSent[NUM_SIZES+1] = { 0 };

#define BUFSIZE 16738
#define UDPBUFSIZE 960			/* (tweakable; should be under 1300) */
static char buf[BUFSIZE];		/* Socket buffer */
static char udpbuf[UDPBUFSIZE];		/* UDP socket buffer */
static char *bufptr=buf;
static char *udpbufptr=udpbuf;
#ifdef DOUBLE_UDP
static char scbuf[UDPBUFSIZE];		/* semi-critical UDP socket buffer */
static char *scbufptr=scbuf;		/* (only used for double UDP) */
#endif
int clientDead=0;
static long sequence;			/* the holy sequence number */

static int udpLocalPort = 0;
static int udpClientPort = 0;
static int udpMode = MODE_SIMPLE;	/* what kind of UDP trans we want */

/* this stuff is used for Fat UDP */
typedef void * PTR;			/* adjust this if you lack (void *) */
typedef struct fat_node_t {
    PTR packet;
    int pkt_size;
    struct fat_node_t *prev;
    struct fat_node_t *next;
} FAT_NODE;
/* needed for fast lookup of semi-critical fat nodes */
FAT_NODE fat_kills[MAXPLAYER];
FAT_NODE fat_torp_info[MAXPLAYER * MAXTORP];
FAT_NODE fat_phaser[MAXPLAYER];
FAT_NODE fat_plasma_info[MAXPLAYER * MAXPLASMA];
FAT_NODE fat_you;
FAT_NODE fat_status;
FAT_NODE fat_planet[MAXPLANETS];
FAT_NODE fat_flags[MAXPLAYER];
FAT_NODE fat_hostile[MAXPLAYER];
/* define the lists */
#define MAX_FAT_LIST	5		/* tweakable; should be > 1 */
typedef struct {
    FAT_NODE *head;
    FAT_NODE *tail;
} FAT_LIST;
FAT_LIST fatlist[MAX_FAT_LIST], tmplist[MAX_FAT_LIST];
/* tweakable parameters; compare with UDPBUFSIZE */
/* NOTE: FAT_THRESH + MAX_FAT_DATA must be < UDPBUFSIZE                */
/*       MAX_FAT_DATA must be larger than biggest semi-critical packet */
#define FAT_THRESH	500		/* if more than this, don't add fat */
#define MAX_FAT_DATA	100		/* add at most this many bytes */
#define MAX_NONFAT	10		/* if we have this much left, stop */


struct plyr_info_spacket clientPlayersInfo[MAXPLAYER];
struct plyr_login_spacket clientLogin[MAXPLAYER];
struct hostile_spacket clientHostile[MAXPLAYER];
struct stats_spacket clientStats[MAXPLAYER];
struct player_spacket clientPlayers[MAXPLAYER];
struct kills_spacket clientKills[MAXPLAYER];
struct flags_spacket clientFlags[MAXPLAYER];
struct pstatus_spacket clientPStatus[MAXPLAYER];
int msgCurrent;
struct torp_info_spacket clientTorpsInfo[MAXPLAYER*MAXTORP];
struct torp_spacket clientTorps[MAXPLAYER*MAXTORP];
struct phaser_spacket clientPhasers[MAXPLAYER];
struct you_spacket clientSelf;
struct status_spacket clientStatus;
struct planet_spacket clientPlanets[MAXPLANETS];
struct planet_loc_spacket clientPlanetLocs[MAXPLANETS];
struct plasma_info_spacket clientPlasmasInfo[MAXPLAYER*MAXPLASMA];
struct plasma_spacket clientPlasmas[MAXPLAYER*MAXPLASMA];
int mustUpdate[MAXPLAYER];

connectToClient(machine, port)
char *machine;
int port;
{
    int ns;
    struct sockaddr_in addr;
    struct hostent *hp;
    /*int len,cc;*/
    char buf[BUFSIZ];
    /*int i;*/

    if (sock!=-1) {
	shutdown(sock,2);
	sock= -1;
    }
    /* printf("Connecting to %s through %d\n", machine, port); */

    if ((ns=socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	printf("I cannot create a socket\n");
	exit(2);
    }

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);

    if (remoteaddr!=-1) {
	addr.sin_addr.s_addr=remoteaddr;
    } else if ((addr.sin_addr.s_addr = inet_addr(machine)) == -1) {
	if ((hp = gethostbyname(machine)) == NULL) {
	    printf("I cannot get host name\n");
	    close(ns);
	    exit(1);
	} else {
	    addr.sin_addr.s_addr = *(long *) hp->h_addr;
	}
    }
    remoteaddr=addr.sin_addr.s_addr;

    if (connect(ns, &addr, sizeof(addr)) < 0) {
	/* printf("I cannot connect through port %d\n", port); */
	close(ns);
	return(0);
    }
    sock=ns;
    initClientData();
    testtime = -1;
    return(1);
}

/* Check the socket to read it's inet addr for possible future use.
 */
checkSocket()
{
    struct sockaddr_in sin;
    int length;

    length=sizeof(sin);
    if (getpeername(sock, (struct sockaddr *) &sin, &length) < 0) {
	/* A bad thing. */
	return;
    }

    remoteaddr=sin.sin_addr.s_addr;
}

initClientData()
/* invalidates all data, so it is all sent to the client */
{
    int i;

    clientDead=0;
    for (i=0; i<MAXPLAYER; i++) {
	clientHostile[i].hostile= -1;
	clientStats[i].losses= -1;
	clientLogin[i].rank= -1;
	clientPlayersInfo[i].shiptype= -1;
	clientPStatus[i].status= -1;
	clientPlayers[i].x= htonl(-1);
	clientPhasers[i].status= -1;
	clientKills[i].kills= htonl(-1);
	clientFlags[i].flags= htonl(-1);
	mustUpdate[i]=0;

	fat_hostile[i].packet = (PTR) &clientHostile[i];
	fat_hostile[i].pkt_size = sizeof(struct hostile_spacket);
	fat_hostile[i].prev = fat_hostile[i].next = (FAT_NODE *) NULL;
	fat_phaser[i].packet = (PTR) &clientPhasers[i];
	fat_phaser[i].pkt_size = sizeof(struct phaser_spacket);
	fat_phaser[i].prev = fat_phaser[i].next = (FAT_NODE *) NULL;
	fat_kills[i].packet = (PTR) &clientKills[i];
	fat_kills[i].pkt_size = sizeof(struct kills_spacket);
	fat_kills[i].prev = fat_kills[i].next = (FAT_NODE *) NULL;
	fat_flags[i].packet = (PTR) &clientFlags[i];
	fat_flags[i].pkt_size = sizeof(struct flags_spacket);
	fat_flags[i].prev = fat_flags[i].next = (FAT_NODE *) NULL;
    }
    for (i=0; i<MAXPLAYER*MAXTORP; i++) {
	clientTorpsInfo[i].status= -1;
	clientTorps[i].x= -1;

	fat_torp_info[i].packet = (PTR) &clientTorpsInfo[i];
	fat_torp_info[i].pkt_size = sizeof(struct torp_info_spacket);
	fat_torp_info[i].prev = fat_torp_info[i].next = (FAT_NODE *) NULL;
    }
    for (i=0; i<MAXPLAYER*MAXPLASMA; i++) {
	clientPlasmasInfo[i].status= -1;
	clientPlasmas[i].x= -1;

	fat_plasma_info[i].packet = (PTR) &clientPlasmasInfo[i];
	fat_plasma_info[i].pkt_size = sizeof(struct plasma_info_spacket);
	fat_plasma_info[i].prev = fat_torp_info[i].next = (FAT_NODE *) NULL;
    }
    for (i=0; i<MAXPLANETS; i++) {
	clientPlanets[i].armies= htonl(-1);
	clientPlanetLocs[i].x= htonl(-1);

	fat_planet[i].packet = (PTR) &clientPlanets[i];
	fat_planet[i].pkt_size = sizeof(struct planet_spacket);
	fat_planet[i].prev = fat_planet[i].next = (FAT_NODE *) NULL;
    }
    msgCurrent=(mctl->mc_current+1) % MAXMESSAGE;
    clientSelf.pnum= -1;

    fat_you.packet = (PTR) &clientSelf;
    fat_you.pkt_size = sizeof(struct you_spacket);
    fat_you.prev = fat_you.next = (FAT_NODE *) NULL;
    fat_status.packet = (PTR) &clientStatus;
    fat_status.pkt_size = sizeof(struct status_spacket);
    fat_status.prev = fat_status.next = (FAT_NODE *) NULL;

    for (i = 0; i < MAX_FAT_LIST; i++)
	fatlist[i].head = fatlist[i].tail = (FAT_NODE *) NULL;
}

isClientDead()
{
    return(clientDead);
}

updateClient()
{
    int i;

    for (i=0; i<MAXPLAYER; i++) {
	mustUpdate[i]=0;
    }

    updateTorps();
    updatePlasmas();
    updateStatus();
    updateSelf();
    updatePhasers();
    updateShips();
    updatePlanets();
    updateMessages();
    if (bufptr==buf && (commMode!=COMM_UDP || udpbufptr==buf)) {
	/* We sent nothing!  We better send something to wake him */
	sendClientPacket(&clientSelf);
    }
    flushSockBuf();
    repCount++;
}

updateStatus()
{
    /* Update status every 10 seconds? */
    if (repCount % efticks(50) == 0 &&
	((ntohl(clientStatus.timeprod) != status->timeprod) ||
	 (clientStatus.tourn != status->tourn))) { /* bug fix 4/2/92 TC */
	clientStatus.type=SP_STATUS;
	clientStatus.tourn=status->tourn;
	clientStatus.armsbomb=htonl(status->armsbomb);
	clientStatus.planets=htonl(status->planets);
	clientStatus.kills=htonl(status->kills);
	clientStatus.losses=htonl(status->losses);
	clientStatus.time=htonl(status->time);
	clientStatus.timeprod=htonl(status->timeprod);
	sendClientPacket(&clientStatus);
    }
}

updateSelf()
{
    if (ntohl(clientSelf.fuel) != me->p_fuel ||
	    ntohl(clientSelf.shield) != me->p_shield ||
	    ntohl(clientSelf.damage) != me->p_damage ||
	    ntohs(clientSelf.etemp) != me->p_etemp ||
	    ntohs(clientSelf.wtemp) != me->p_wtemp ||
	    ntohl(clientSelf.flags) != me->p_flags || 
	    clientSelf.armies != me->p_armies ||
	    clientSelf.swar != me->p_swar ||
	    ntohs(clientSelf.whydead) != me->p_whydead ||
	    ntohs(clientSelf.whodead) != me->p_whodead ||
	    clientSelf.pnum != me->p_no) {

	/* we want to send it, but how? */
	clientSelf.type = SP_YOU;

	if (commMode == COMM_UDP) {
	    if (    ntohl(clientSelf.flags) != me->p_flags ||
		    clientSelf.armies != me->p_armies ||
		    clientSelf.swar != me->p_swar) {
		clientSelf.type=SP_YOU | 0x40;	/* mark as semi-critical */
	    }
	    if (    ntohs(clientSelf.whydead) != me->p_whydead ||
		    ntohs(clientSelf.whodead) != me->p_whodead ||
		    clientSelf.pnum != me->p_no) {
		clientSelf.type=SP_YOU | 0x80;	/* mark as critical */
	    }
	}

	clientSelf.pnum=me->p_no;
	clientSelf.flags=htonl(me->p_flags);
	clientSelf.swar=me->p_swar;
	clientSelf.hostile=me->p_hostile;
	clientSelf.armies=me->p_armies;
	clientSelf.shield=htonl(me->p_shield);
	clientSelf.fuel=htonl(me->p_fuel);
	clientSelf.etemp=htons(me->p_etemp);
	clientSelf.wtemp=htons(me->p_wtemp);
	clientSelf.whydead=htons(me->p_whydead);
	clientSelf.whodead=htons(me->p_whodead);
	clientSelf.damage=htonl(me->p_damage);
	/* clientSelf.tractor=(char)me->p_tractor |0x40;		 ATM - visible tractor */
	sendClientPacket(&clientSelf);
    }
}

updateShips()
{
    register int i;
    register struct player *pl;
    register struct plyr_info_spacket *cpli;
    register struct player_spacket *cpl;
    register struct kills_spacket *kills;
    register struct flags_spacket *flags;
    register struct pstatus_spacket *pstatus;
    register struct plyr_login_spacket *login;
    register struct hostile_spacket *hostile;
    register struct stats_spacket *stats;
    int update;
    int x,y;
/*#define FLAGMASK (PFSHIELD|PFBOMB|PFORBIT|PFCLOAK|PFROBOT|PFBEAMUP|PFBEAMDOWN|PFPRACTR|PFDOCK|PFTRACT|PFPRESS|PFDOCKOK) atm mask */

/* #define FLAGMASK (PFSHIELD|PFBOMB|PFORBIT|PFCLOAK|PFROBOT|PFBEAMUP|PFBEAMDOWN|PFPRACTR|PFDOCK|PFTRACT|PFPRESS|PFDOCKOK) aieee, too much.  7/27/91 TC */ 

#define FLAGMASK (PFSHIELD|PFBOMB|PFORBIT|PFCLOAK|PFROBOT|PFPRACTR|PFDOCK|PFTRACT|PFPRESS|PFDOCKOK)
#define INVISOMASK (PFCLOAK|PFROBOT|PFPRACTR|PFDOCKOK)

    /* Please excuse the ugliness of this loop declaration */
    for (i=0, pl=players, cpli=clientPlayersInfo, cpl=clientPlayers, kills=clientKills, flags=clientFlags, pstatus=clientPStatus, login=clientLogin, hostile=clientHostile, stats=clientStats;
    	    i<MAXPLAYER; 
	    i++, pl++, cpli++, cpl++, kills++, flags++, pstatus++, login++, hostile++, stats++) {
	update=0;
	if (strcmp(pl->p_name, login->name)!=0 ||
	    pl->p_stats.st_rank != login->rank ||
	    strcmp(pl->p_monitor, login->monitor)!=0) {
	    strncpy(login->name,pl->p_name,15);
	    strncpy(login->monitor,pl->p_monitor,15);
	    strncpy(login->login,pl->p_login,15);
	    login->name[15]=0;
	    login->monitor[15]=0;
	    login->login[15]=0;
	    login->type=SP_PL_LOGIN;
	    login->pnum=i;
	    login->rank=pl->p_stats.st_rank;
	    sendClientPacket(login);
	}
	if ((pl!=me && (pl->p_swar & me->p_team)!=hostile->war) ||
	    (pl!=me && (pl->p_hostile & me->p_team)!=hostile->hostile) ||
	    (pl==me && pl->p_swar != hostile->war) ||
	    (pl==me && pl->p_hostile != hostile->hostile)) {
	    hostile->type=SP_HOSTILE;
	    if (pl==me) {
		hostile->war=pl->p_swar;
		hostile->hostile=pl->p_hostile;
	    } else {
		hostile->war=(pl->p_swar & me->p_team);
		hostile->hostile=(pl->p_hostile & me->p_team);
	    }
	    hostile->pnum=i;
	    sendClientPacket(hostile);
	}
	/* Send stat packets once per five updates.
	 * But, only send one.  We will cycle through them all eventually.
	 */
	if (repCount % (MAXPLAYER * efticks(5)) == i*efticks(5) &&
	    (stats->tkills != htonl(pl->p_stats.st_tkills) ||
	     stats->tlosses != htonl(pl->p_stats.st_tlosses) ||
	     stats->kills != htonl(pl->p_stats.st_kills) ||
	     stats->losses != htonl(pl->p_stats.st_losses) ||
	     stats->tticks != htonl(pl->p_stats.st_tticks) ||
	     stats->tplanets != htonl(pl->p_stats.st_tplanets) ||
	     stats->tarmies != htonl(pl->p_stats.st_tarmsbomb) ||
	     stats->sbkills != htonl(pl->p_stats.st_sbkills) ||
	     stats->sblosses != htonl(pl->p_stats.st_sblosses) ||
	     stats->armies != htonl(pl->p_stats.st_armsbomb) ||
	     stats->planets != htonl(pl->p_stats.st_planets))) {
	    stats->tkills = htonl(pl->p_stats.st_tkills);
	    stats->tlosses = htonl(pl->p_stats.st_tlosses);
	    stats->kills = htonl(pl->p_stats.st_kills);
	    stats->losses = htonl(pl->p_stats.st_losses);
	    stats->tticks = htonl(pl->p_stats.st_tticks);
	    stats->tplanets = htonl(pl->p_stats.st_tplanets);
	    stats->tarmies = htonl(pl->p_stats.st_tarmsbomb);
	    stats->sbkills = htonl(pl->p_stats.st_sbkills);
	    stats->sblosses = htonl(pl->p_stats.st_sblosses);
	    stats->armies = htonl(pl->p_stats.st_armsbomb);
	    stats->planets = htonl(pl->p_stats.st_planets);
	    stats->maxkills=htonl((int) (pl->p_stats.st_maxkills * 100));
	    stats->sbmaxkills=htonl((int) (pl->p_stats.st_sbmaxkills * 100));
	    stats->type=SP_STATS;
	    stats->pnum=i;
	    sendClientPacket(stats);
	}
	if (pl->p_ship.s_type != cpli->shiptype ||
	    pl->p_team != cpli->team) {
	    cpli->type=SP_PLAYER_INFO;
	    cpli->pnum=i;
	    cpli->shiptype=pl->p_ship.s_type;
	    cpli->team=pl->p_team;
	    sendClientPacket(cpli);
	}
	if (kills->kills != htonl((int) (pl->p_kills*100))) {
	    kills->type=SP_KILLS;
	    kills->pnum=i;
	    kills->kills=htonl((int) (pl->p_kills*100));
	    sendClientPacket(kills);
	}
	if (pstatus->status != pl->p_status) {
	    /* We update the location of people whose status has changed.
	     * (like if they just re-entered...)
	     */
	    update=1;
	    pstatus->type=SP_PSTATUS;
	    pstatus->pnum=i;
	    pstatus->status=pl->p_status;
	    if (pl->p_status == PFREE) { /* I think this will turn off ignores
					  for players that leave. 7/24/91 TC */
		extern int ignored[];
		ignored[i] = 0;
	    }
	    sendClientPacket(pstatus);
	}

	/* Used to send flags here, see below 8/7/91 TC */

	if (!hiddenenemy || pl->p_team == me->p_team || 
	    (pl->p_flags & PFSEEN) || (status->tourn==0/*&& hiddenenemy<2*/)) {
	    x=pl->p_x;
	    y=pl->p_y;
	    if (me!=pl && flags->flags != htonl(FLAGMASK & pl->p_flags)) {
		flags->type=SP_FLAGS;
		flags->pnum=i;
		flags->flags=htonl(FLAGMASK & pl->p_flags);
		/* flags->tractor = (char) pl->p_tractor |0x40;	/* ATM - visible tractor */
	    sendClientPacket(flags);
	    }
	} else {
	    /* A hack to make him inviso */
	    x= -100000;
	    y= -100000;

	    /* reduce flag info if he's inviso 8/7/91 TC */

	    if (me!=pl && flags->flags != htonl(INVISOMASK & pl->p_flags)) {
		flags->type=SP_FLAGS;
		flags->pnum=i;
		flags->flags=htonl(INVISOMASK & pl->p_flags);
		sendClientPacket(flags);
	    }
	}
	if (x != ntohl(cpl->x) || y != ntohl(cpl->y) || 
		pl->p_dir != cpl->dir || pl->p_speed != cpl->speed) {
	    /* We update the player if:
	     * 1) haven't updated him for 9 intervals.
	     * 2) he is on the screen
	     * 3) he was on the screen recently.
	     */
	    if (!update && repCount % efticks(9) != 0 && 
		(ntohl(cpl->x) < me->p_x - SCALE*WINSIDE/2 ||
		 ntohl(cpl->x) > me->p_x + SCALE*WINSIDE/2 ||
		 ntohl(cpl->y) > me->p_y + SCALE*WINSIDE/2 ||
		 ntohl(cpl->y) < me->p_y - SCALE*WINSIDE/2) &&
		(y > me->p_y + SCALE*WINSIDE/2 ||
		 x > me->p_x + SCALE*WINSIDE/2 ||
		 x < me->p_x - SCALE*WINSIDE/2 ||
		 y < me->p_y - SCALE*WINSIDE/2)) continue;
	    /* If the guy is cloaked, give information only occasionally,
	     *  and make it slightly inaccurate.
	     * Also, we don't give a direction.  The client has no reason
	     *  to know.
	     */
	    if ((pl->p_flags & PFCLOAK) && 
		  (pl->p_cloakphase == (CLOAK_PHASES - 1)) && 
		  (me != pl) && !mustUpdate[i]) {
		if (repCount % efticks(9) != 0) continue;
		cpl->type=SP_PLAYER;
		cpl->pnum=i;
		cpl->x=htonl(x+(random() % 2000)-1000);
		cpl->y=htonl(y+(random() % 2000)-1000);
		sendClientPacket(cpl);
		continue;
	    }
	    cpl->type=SP_PLAYER;
	    cpl->pnum=i;
	    cpl->x=htonl(x);
	    cpl->y=htonl(y);
	    cpl->speed=pl->p_speed;
	    cpl->dir=pl->p_dir;
	    sendClientPacket(cpl);
	}
    }
}

updateTorps()
{
    register struct torp *torp;
    register int i;
    register struct torp_info_spacket *tpi;
    register struct torp_spacket *tp;

    for (i=0, torp=torps, tpi=clientTorpsInfo, tp=clientTorps; 
	    i<MAXPLAYER*MAXTORP; 
	    i++, torp++, tpi++, tp++) {
	if (torp->t_owner==me->p_no) {
	    if (torp->t_war!=tpi->war ||
		torp->t_status!=tpi->status) {
		tpi->type=SP_TORP_INFO;
		tpi->war=torp->t_war;
		tpi->status=torp->t_status;
		tpi->tnum=htons(i);
		sendClientPacket(tpi);
	    }
	    if (tp->x!=htonl(torp->t_x) ||
		tp->y!=htonl(torp->t_y)) {
		tp->type=SP_TORP;
		tp->x=htonl(torp->t_x);
		tp->y=htonl(torp->t_y);
		tp->dir=torp->t_dir;
		tp->tnum=htons(i);
		sendClientPacket(tp);
	    }
	} else {	/* Someone else's torp... */
	    if (torp->t_y > me->p_y + SCALE*WINSIDE/2 ||
		torp->t_x > me->p_x + SCALE*WINSIDE/2 ||
		torp->t_x < me->p_x - SCALE*WINSIDE/2 ||
		torp->t_y < me->p_y - SCALE*WINSIDE/2 ||
		torp->t_status==TFREE) {
		if (torp->t_status==TFREE && tpi->status==TEXPLODE) {
		    tpi->status=TFREE;
		    continue;
		}
		if (tpi->status!=TFREE) {
		    tpi->status=TFREE;
		    tpi->tnum=htons(i);
		    tpi->type=SP_TORP_INFO;
		    sendClientPacket(tpi);
		}
	    } else {	/* in view */
		if (tp->x!=htonl(torp->t_x) ||
		    tp->y!=htonl(torp->t_y)) {
		    tp->x=htonl(torp->t_x);
		    tp->y=htonl(torp->t_y);
		    tp->dir=torp->t_dir;
		    tp->tnum=htons(i);
		    tp->type=SP_TORP;
		    sendClientPacket(tp);
		}
		if (torp->t_status!=tpi->status ||
		    (torp->t_war & me->p_team)!=tpi->war) {
		    /* Let the client fade away the explosion on its own */
		    tpi->war=torp->t_war & me->p_team;
		    tpi->type=SP_TORP_INFO;
		    tpi->tnum=htons(i);
		    tpi->status=torp->t_status;
		    sendClientPacket(tpi);
		}
	    }
	}
    }
}

updatePlasmas()
{
    register struct plasmatorp *torp;
    register int i;
    register struct plasma_info_spacket *tpi;
    register struct plasma_spacket *tp;

    for (i=0, torp=plasmatorps, tpi=clientPlasmasInfo, tp=clientPlasmas; 
	    i<MAXPLAYER*MAXPLASMA;
	    i++, torp++, tpi++, tp++) {
	if (torp->pt_owner==me->p_no) {
	    if (torp->pt_war!=tpi->war ||
		torp->pt_status!=tpi->status) {
		tpi->type=SP_PLASMA_INFO;
		tpi->war=torp->pt_war;
		tpi->status=torp->pt_status;
		tpi->pnum=htons(i);
		sendClientPacket(tpi);
	    }
	    if (tp->x!=htonl(torp->pt_x) ||
		tp->y!=htonl(torp->pt_y)) {
		tp->type=SP_PLASMA;
		tp->x=htonl(torp->pt_x);
		tp->y=htonl(torp->pt_y);
		tp->pnum=htons(i);
		sendClientPacket(tp);
	    }
	} else {	/* Someone else's torp... */
	    if (torp->pt_y > me->p_y + SCALE*WINSIDE/2 ||
		torp->pt_x > me->p_x + SCALE*WINSIDE/2 ||
		torp->pt_x < me->p_x - SCALE*WINSIDE/2 ||
		torp->pt_y < me->p_y - SCALE*WINSIDE/2 ||
		torp->pt_status==PTFREE) {
		if (torp->pt_status==PTFREE && tpi->status==PTEXPLODE) {
		    tpi->status=PTFREE;
		    continue;
		}
		if (tpi->status!=PTFREE) {
		    tpi->status=PTFREE;
		    tpi->pnum=htons(i);
		    tpi->type=SP_PLASMA_INFO;
		    sendClientPacket(tpi);
		}
	    } else {	/* in view */
		/* Send torp (we assume it moved) */
		tp->x=htonl(torp->pt_x);
		tp->y=htonl(torp->pt_y);
		tp->pnum=htons(i);
		tp->type=SP_PLASMA;
		sendClientPacket(tp);
		if (torp->pt_status!=tpi->status ||
		    (torp->pt_war & me->p_team)!=tpi->war) {
		    tpi->war=torp->pt_war & me->p_team;
		    tpi->type=SP_PLASMA_INFO;
		    tpi->pnum=htons(i);
		    tpi->status=torp->pt_status;
		    sendClientPacket(tpi);
		}
	    }
	}
    }
}

updatePhasers()
{
    register int i;
    register struct phaser_spacket *ph;
    register struct phaser *phase;
    register struct player *pl;

    for (i=0, ph=clientPhasers, phase=phasers, pl=players; 
	 i<MAXPLAYER; i++, ph++, phase++, pl++) {
	if (pl->p_y > me->p_y + SCALE*WINSIDE/2 ||
	    pl->p_x > me->p_x + SCALE*WINSIDE/2 ||
	    pl->p_x < me->p_x - SCALE*WINSIDE/2 ||
	    pl->p_y < me->p_y - SCALE*WINSIDE/2) {
	    if (ph->status!=PHFREE) {
		ph->pnum=i;
		ph->type=SP_PHASER;
		ph->status=PHFREE;
		sendClientPacket(ph);
	    }
	} else {
	    if (phase->ph_status==PHHIT) {
		mustUpdate[phase->ph_target]=1;
	    }
	    if (ph->status!=phase->ph_status ||
		  ph->dir!=phase->ph_dir ||
		  ph->target!=htonl(phase->ph_target)) { 
		ph->pnum=i;
		ph->type=SP_PHASER;
		ph->status=phase->ph_status;
		ph->dir=phase->ph_dir;
		ph->x=htonl(phase->ph_x);
		ph->y=htonl(phase->ph_y);
		ph->target=htonl(phase->ph_target);
		sendClientPacket(ph);
	    }
	}
    }
}

updatePlanets()
{
    register int i;
    register struct planet_spacket *pl;
    register struct planet *plan;
    register struct planet_loc_spacket *pll;

#define PLFLAGMASK (PLREPAIR|PLFUEL|PLAGRI|PLHOME|PLCOUP)

    for (i=0, pl=clientPlanets, plan=planets, pll=clientPlanetLocs; 
	 i<MAXPLANETS;
	 i++, pl++, plan++, pll++) {
	/* Send him info about him not having info if he doesn't but thinks
	 *  he does.
	 * Also send him info on every fifth cycle if the planet needs to
	 *  be redrawn.
	 */
	if (((plan->pl_info & me->p_team)==0) && (pl->info & me->p_team)) {
	    pl->type=SP_PLANET;
	    pl->pnum=i;
	    pl->info=0;
	    sendClientPacket(pl);
	} else 
	if (plan->pl_info & me->p_team &&
	    (pl->info != plan->pl_info ||
	     pl->armies != htonl(plan->pl_armies) ||
	     pl->owner != plan->pl_owner ||
	     pl->flags != htons((short) (plan->pl_flags & PLFLAGMASK)))) {
	    pl->type=SP_PLANET;
	    pl->pnum=i;
	    pl->info=plan->pl_info;
	    pl->flags=htons((short) (plan->pl_flags & PLFLAGMASK));
	    pl->armies=htonl(plan->pl_armies);
	    pl->owner=plan->pl_owner;
	    sendClientPacket(pl);
	} 
	/* Assume that the planet only needs to be updated once... */

	/* Odd, changes in pl_y not supported.  5/31/92 TC */

	if ((pll->x != htonl(plan->pl_x)) || (pll->y != htonl(plan->pl_y))) {
	    pll->x=htonl(plan->pl_x);
	    pll->y=htonl(plan->pl_y);
	    pll->pnum=i;
	    strcpy(pll->name, plan->pl_name);
	    pll->type=SP_PLANET_LOC;
	    sendClientPacket(pll);
	}
    }
}

updateMessages()
{
    int i;
    struct message *cur;
    struct mesg_spacket msg;
    int updated=0;

    for (i=msgCurrent; i!=(mctl->mc_current+1) % MAXMESSAGE; i=(i+1) % MAXMESSAGE) {
	if (i==MAXMESSAGE) i=0;
	cur= &messages[i];

	if (cur->m_flags & MVALID &&
	    (cur->m_flags & MALL ||
	     (cur->m_flags & MTEAM && cur->m_recpt & me->p_team) ||
	     (cur->m_flags & MINDIV && cur->m_recpt == me->p_no) /*||
	     (cur->m_flags & MGOD && cur->m_recpt == me->p_no)*/ )) {
	    if (!updated) {
		update_sys_defaults();
		updated=1;
	    }
	    msg.type=SP_MESSAGE;
	    strncpy(msg.mesg, cur->m_data, 80);
	    msg.mesg[79]='\0';
	    msg.m_flags=cur->m_flags;
	    msg.m_recpt=cur->m_recpt;
	    msg.m_from=cur->m_from;

	    if ((cur->m_from < 0) || (cur->m_from > MAXPLAYER))
		sendClientPacket(&msg);
	    else if (cur->m_flags & MALL && !(ignored[cur->m_from] & MALL))
		sendClientPacket(&msg);
	    else if (cur->m_flags & MTEAM && !(ignored[cur->m_from] & MTEAM))
		sendClientPacket(&msg);
	    else if (cur->m_flags & MINDIV) {
		
		/* session stats now parsed here.  parseQuery == true */
		/* means eat message 4/17/92 TC */
		
		if (!parseQuery(&msg))
		    if (ignored[cur->m_from] & MINDIV)
			bounce("That player is currently ignoring you.",
			       cur->m_from);
		    else
			sendClientPacket(&msg);
	    }
	}
	msgCurrent=(msgCurrent+1) % MAXMESSAGE;
    }
}

sendQueuePacket(pos)
short pos;
{
    struct queue_spacket qPacket;

    qPacket.type=SP_QUEUE;
    qPacket.pos=htons(pos);
    sendClientPacket(&qPacket);
    flushSockBuf();
}

sendClientPacket(packet)
/* Pick a random type for the packet */
struct player_spacket *packet;
{
    int orig_type;
    int size, issc;
    int cc;
    static int oldStatus = POUTFIT;

    orig_type = packet->type;
/*    packet->type &= ~(0x40 | 0x80);	/* clear special flags */
    packet->type &= (char) 0x3f; /* above doesn't work? 4/18/92 TC */
#ifdef MAYBE
    /*
     * If we're dead, dying, or just born, we definitely want the transmission
     * to get through (otherwise we can get stuck).  I don't think this will
     * be a problem for anybody, though it might hang for a bit if the TCP
     * connection is bad.
     */
    /* Okay, now I'm not so sure.  Whatever. */
    if (oldStatus != PALIVE || (me != NULL && me->p_status != PALIVE))
	orig_type = packet->type | 0x80;	/* pretend it's critical */
#endif
    if (packet->type<1 || packet->type>NUM_SIZES || sizes[packet->type]==0) {
	printf("Attempt to send strange packet %d\n", packet->type);
	return;
    }
    packetsSent[packet->type]++;
    if (commMode == COMM_TCP || (commMode == COMM_UDP && udpMode == MODE_TCP)) {
	/*
	 * business as usual
	 */
	size=sizes[packet->type];
	if (bufptr-buf+size >= BUFSIZE) {
	    if ((cc = gwrite(sock, buf, bufptr-buf)) != bufptr-buf) {
		fprintf(stderr, "std gwrite failed (%d, error %d)\n",cc,errno);
		clientDead=1;
	    }
	    bufptr=buf;
	}
	bcopy(packet, bufptr, size);
	bufptr+=size;

    } else {
	/*
	 * do UDP stuff unless it's a "critical" packet
	 * (note that both kinds get a sequence number appended) (FIX)
	 */
	issc = 0;
	switch (orig_type) {
	case SP_KILLS:
	case SP_TORP_INFO:
	case SP_PHASER:
	case SP_PLASMA_INFO:
	case SP_YOU|0x40:	/* ??? what is this? */
	case SP_STATUS:
	case SP_PLANET:
	case SP_FLAGS:
	case SP_HOSTILE:
	    /* these are semi-critical; flag as semi-critical and fall through*/
	    issc = 1;

	case SP_PLAYER:
	case SP_TORP:
	case SP_YOU:
	case SP_PLASMA:
	case SP_STATS:
/*	case SP_SCAN:*/
	case SP_UDP_REPLY:	/* only reply when COMM_UDP is SWITCH_VERIFY */
	    /* these are non-critical updates; send them via UDP */
	    V_UDPDIAG(("Sending type %d\n", packet->type));
	    size=sizes[packet->type];
	    if (udpbufptr-udpbuf+size >= UDPBUFSIZE) {
		if ((cc = gwrite(udpSock, udpbuf, udpbufptr-udpbuf)) !=
							    udpbufptr-udpbuf) {
		    fprintf(stderr, "UDP gwrite failed (%d, error %d)\n",
			cc, errno);
/*		    clientDead=1;*/
		    UDPDIAG(("*** UDP disconnected for %s\n", me->p_name));
                    printUdpInfo();
                    closeUdpConn();
                    commMode = COMM_TCP;
		}
#ifdef DOUBLE_UDP
		sendSC();	/* send semi-critical info, if needed */
#endif
		udpbufptr=udpbuf + addSequence(udpbuf);
	    }
	    bcopy(packet, udpbufptr, size);
	    udpbufptr+=size;
#ifdef DOUBLE_UDP
	    if (issc && udpMode == MODE_DOUBLE) {
		bcopy(packet, scbufptr, size);
		scbufptr+=size;
		V_UDPDIAG((" adding SC\n"));
	    }
#endif
	    if (issc && udpMode == MODE_FAT) {
		updateFat(packet);
	    }
	    break;

	default:
	    /* these are critical packets; send them via TCP */
	    size=sizes[packet->type];
	    if (bufptr-buf+size >= BUFSIZE) {
		if (gwrite(sock, buf, bufptr-buf) != bufptr-buf) {
		    fprintf(stderr, "TCP gwrite failed (%d, error %d)\n",
			cc, errno);
		    clientDead=1;
		}
		bufptr=buf /*+ addSequence(buf)*/;
	    }
	    bcopy(packet, bufptr, size);
	    bufptr+=size;
	    break;
	}
    }

    if (me != NULL) oldStatus = me->p_status;
}

/* Flush the socket buffer */
flushSockBuf()
{
    int cc;

    if (clientDead)
	return;
    if (bufptr != buf) {
	if ((cc = gwrite(sock, buf, bufptr-buf)) != bufptr-buf) {
	    fprintf(stderr, "std flush gwrite failed (%d, error %d)\n",
		cc, errno);
	    clientDead=1;
	}
	bufptr=buf /*+ addSequence(buf)*/;
    }
    /*
     * This is where we try to add fat.  There's no point in checking at
     * the other places which call gwrite(), because they only call it when
     * the buffer is already full.
     */
    if (udpSock >= 0 && udpMode == MODE_FAT && (udpbufptr-udpbuf)<FAT_THRESH)
	fatten();
    if (udpSock >= 0 && udpbufptr != udpbuf) {
#ifdef BROKEN
	/* debugging only!! */
# ifdef HOSED
	if (sequence % 10) {
# else
	if (sequence % 5 == 0) {
# endif
	    /* act as if we did the gwrite(), but don't */
	    udpbufptr=udpbuf + addSequence(udpbuf);
	    goto foo;
	}
#endif
	if ((cc=gwrite(udpSock, udpbuf, udpbufptr-udpbuf)) != udpbufptr-udpbuf){
	    fprintf(stderr, "UDP flush gwrite failed (%d, error %d)\n",
		cc, errno);
/*	    clientDead=1;*/
	    UDPDIAG(("*** UDP disconnected for %s\n", me->p_name));
	    printUdpInfo();
	    closeUdpConn();
	    commMode = COMM_TCP;
	}
#ifdef DOUBLE_UDP
	sendSC();
#endif
	udpbufptr=udpbuf + addSequence(udpbuf);
    }
#ifdef BROKEN
foo:
#endif
    if (udpMode == MODE_FAT)
	fatMerge();
}

socketPause()
{
    struct timeval timeout;
    int readfds;

    timeout.tv_sec=1;
    timeout.tv_usec=0;
    readfds = 1<<sock;
    select(32,&readfds,0,0,&timeout); 
}

/* Find out if client has any requests */
readFromClient()
{
    struct timeval timeout;
    fd_set readfds;
    int retval = 0;

    if (clientDead) return(0);
    timeout.tv_sec=0;
    timeout.tv_usec=0;
    FD_ZERO(&readfds);
    FD_SET(sock, &readfds);
    if (udpSock >= 0)
	FD_SET(udpSock, &readfds);
    if (select(32,&readfds,0,0,&timeout) != 0) {
	/* Read info from the xtrek client */
	if (FD_ISSET(sock, &readfds)) {
	    retval += doRead(sock);
	}
	if (udpSock >= 0 && FD_ISSET(udpSock, &readfds)) {
	    V_UDPDIAG(("Activity on UDP socket\n"));
	    retval += doRead(udpSock);
	}
    }
    return (retval != 0);               /* convert to 1/0 */
}

/* ripped out of above routine */
doRead(asock)
int asock;
{
    struct timeval timeout;
/*    int readfds;*/
    fd_set readfds;
    char buf[BUFSIZ*2];
    char *bufptr;
    int size;
    int count;
    int temp;

    timeout.tv_sec=0;
    timeout.tv_usec=0;
/*    readfds = 1<<asock;*/
    FD_ZERO(&readfds);
    FD_SET(asock, &readfds);
        /* Read info from the xtrek server */
        count=read(asock,buf,BUFSIZ*2);
        if (count<=0) {
#if DEBUG
	    /* (this happens when the client hits 'Q') */
	    fprintf(stderr, "1) read() failed in doRead (%d, error %d)\n",
		count, errno);
	    fprintf(stderr, "asock=%d, sock=%d\n", asock, sock);
#endif
	    if (asock == udpSock) {
		if (errno == ECONNREFUSED) {
		    struct sockaddr_in addr;

		    UDPDIAG(("Hiccup(%d)!  Reconnecting\n", errno));
		    addr.sin_addr.s_addr = remoteaddr;
		    addr.sin_port = htons(udpClientPort);
		    addr.sin_family = AF_INET;
		    if (connect(udpSock, &addr, sizeof(addr)) < 0) {
			perror("hiccup connect");
			UDPDIAG(("Unable to reconnect\n"));
			/* and fall through to disconnect */
		    } else {
			UDPDIAG(("Reconnect successful\n"));
			return (0);
		    }
		}

		UDPDIAG(("*** UDP disconnected (res=%d, err=%d)\n",
			count, errno));
		printUdpInfo();
		closeUdpConn();
		commMode = COMM_TCP;
		return(0);
	    }
	    clientDead=1;
	    return(0);
        }
        bufptr=buf;
        while (bufptr < buf+count) {
	    if (*bufptr < 1 || *bufptr > NUM_PACKETS || handlers[*bufptr].size==0) {
                printf("Unknown packet type: %d, aborting...\n", *bufptr);
                return;
            }
	    size=handlers[*bufptr].size;
	    while (size>count+(buf-bufptr)) {
		/* We wait for up to twenty seconds for rest of packet.
                 * If we don't get it, we assume the client died. 
                 */
		timeout.tv_sec=20;
                timeout.tv_usec=0;
                /*readfds=1<<asock;*/
		FD_ZERO(&readfds);
		FD_SET(asock, &readfds);
                if (select(32,&readfds,0,0,&timeout) == 0) {
		    logmessage("Died while waiting for packet...");
		    fprintf(stderr, "1a) read() failed (%d, error %d)\n",
			count, errno);
                    clientDead=1;
                    return(0);
                }

		temp=read(asock,buf+count,size-(count+(buf-bufptr)));
		count+=temp;
		if (temp<=0) {
		    sprintf(buf, "Died in second read(), return=%d", temp);
		    logmessage(buf);
		    fprintf(stderr, "2) read() failed (%d, error %d)\n",
			count, errno);
		    clientDead=1;
		    return(0);
		}
	    }
	    /* Check to see if the handler is there and the request is legal.
	     * The code is a little ugly, but it isn't too bad to worry about
	     * yet.
	     */
	    packetsReceived[*bufptr]++;
	    if (handlers[*bufptr].handler != NULL) {
		if (((inputMask==-1 || inputMask==*bufptr) && 
		     (me==NULL || !(me->p_flags & (PFWAR|PFREFITTING)))) || 
		     *bufptr==CP_RESETSTATS || *bufptr==CP_UPDATES ||
		     *bufptr==CP_OPTIONS || *bufptr==CP_RESERVED ||
		     *bufptr==CP_SOCKET || *bufptr==CP_BYE) {
		    if (me && me->p_flags & PFSELFDEST) {
			me->p_flags &= ~PFSELFDEST;
			warning("Self Destruct has been canceled");
		    }
		    (*(handlers[*bufptr].handler))(bufptr);
		}
		/* Otherwise we ignore the request */
	    } else {
		printf("Handler for packet %d not installed...\n", *bufptr);
	    }
            bufptr+=size;
            if (bufptr>buf+BUFSIZ) {
                bcopy(buf+BUFSIZ, buf, BUFSIZ);
                if (count==BUFSIZ*2) {
		    /*readfds = 1<<asock;*/
		    FD_ZERO(&readfds);
		    FD_SET(asock, &readfds);
		    if (select(32,&readfds,0,0,&timeout)) {
			temp=read(asock,buf+BUFSIZ,BUFSIZ);
			count=BUFSIZ+temp;
			if (temp<=0) {
			    sprintf(buf, "Died in third read(), return=%d", temp);
			    fprintf(stderr, "3) read() failed (%d, error %d)\n",
				count, errno);
			    logmessage(buf);
			    clientDead=1;
			    return(0);
			}
		    } else {
			count=BUFSIZ;
		    }
                } else {
		    count -=BUFSIZ;
		}
                bufptr-=BUFSIZ;
            }
        }
        return(1);
}

handleTorpReq(packet)
struct torp_cpacket *packet;
{
    ntorp(packet->dir, TMOVE);
}

handlePhasReq(packet)
struct phaser_cpacket *packet;
{
    phaser(packet->dir);
}

handleSpeedReq(packet)
struct speed_cpacket *packet;
{
    set_speed(packet->speed);
}

handleDirReq(packet)
struct dir_cpacket *packet;
{
    me->p_flags &= ~(PFPLOCK | PFPLLOCK);
    set_course(packet->dir);
}

handleShieldReq(packet)
struct shield_cpacket *packet;
{
    if (packet->state) {
	shield_up();
    } else {
	shield_down();
    }
}

handleRepairReq(packet)
struct repair_cpacket *packet;
{
    if (packet->state) {
	repair();
    } else {
	me->p_flags &= ~(PFREPAIR);
    }
}

handleOrbitReq(packet)
struct orbit_cpacket *packet;
{
    if (packet->state) {
	orbit();
    } else {
	me->p_flags &= ~PFORBIT;
	if (me->p_flags & PFDOCK) {
	    players[me->p_docked].p_docked--;
	    players[me->p_docked].p_port[me->p_port[0]] = VACANT;
	    me->p_flags &= ~PFDOCK;
	}
    }
}

handlePractrReq(packet)
struct practr_cpacket *packet;
{
    practice_robo();
}

handleBombReq(packet)
struct bomb_cpacket *packet;
{
    if (packet->state) {
	bomb_planet();
    } else {
	me->p_flags &= ~(PFBOMB);
    }
}

handleBeamReq(packet)
struct beam_cpacket *packet;
{
    if (packet->state==1) {
	beam_up();
    } else if (packet->state) {
	beam_down();
    } else {
	me->p_flags &= ~(PFBEAMUP | PFBEAMDOWN);
    }
}

handleCloakReq(packet)
struct cloak_cpacket *packet;
{
    if (packet->state) {
	cloak_on();
    } else {
	cloak_off();
    }
}

handleDetTReq(packet)
struct det_torps_cpacket *packet;
{
    detothers();
}

handleCopilotReq(packet)
struct copilot_cpacket *packet;
{
/*
 * Unsupported...
    if (packet->state) {
	me->p_flags |= PFCOPILOT;
    } else {
	me->p_flags &= ~PFCOPILOT;
    }
 */
}

handleOutfit(packet)
struct outfit_cpacket *packet;
{
    shipPick=packet->ship;
    teamPick=packet->team;
}

sendPickokPacket(state)
int state;
{
    struct pickok_spacket pickPack;

    pickPack.type=SP_PICKOK;
    pickPack.state=state;
    sendClientPacket(&pickPack);
}

handleLoginReq(packet)
struct login_cpacket *packet;
{
    
    strncpy(namePick, packet->name, 16);
    namePick[15]=0;
    strncpy(passPick, packet->password, 16);
    passPick[15]=0;
    /* Is this a name query or a login? */
    if (packet->query) {
	passPick[15]=1;
    }
    strncpy(login, packet->login, 16);
    login[15]=0;
}

sendClientLogin(stats)
struct stats *stats;
{
    struct login_spacket logPacket;

    logPacket.type=SP_LOGIN;
    if (stats==NULL) {
	logPacket.accept=0;
    } else {
	logPacket.accept=1;
	logPacket.flags=htonl(stats->st_flags);
	bcopy(stats->st_keymap, logPacket.keymap, 96);
    }
    sendClientPacket(&logPacket);
}

handlePlasmaReq(packet)
struct plasma_cpacket *packet;
{
    nplasmatorp(packet->dir, PTMOVE);
}

handleWarReq(packet)
struct war_cpacket *packet;
{
    declare_war(packet->newmask);
}

handlePlanlockReq(packet)
struct planlock_cpacket *packet;
{
    lock_planet(packet->pnum);
}

handlePlaylockReq(packet)
struct playlock_cpacket *packet;
{
    lock_player(packet->pnum);
}

handleDetMReq(packet)
struct det_mytorp_cpacket *packet;
{
    struct torp *atorp;
    int t;

    t=ntohs(packet->tnum);
    if (t<0 || t>=MAXPLAYER*MAXTORP) return;
    atorp= &torps[t];
    
    if (atorp->t_owner != me->p_no) return;
    if (atorp->t_status == TMOVE || atorp->t_status == TSTRAIGHT) {
	atorp->t_status = TOFF;
    }
}

handleTractorReq(packet)
struct tractor_cpacket *packet;
{
    int target;
    struct player *player;

    if (weaponsallowed[WP_TRACTOR]==0) {
	warning("Tractor beams haven't been invented yet.");
	return;
    }
    target=packet->pnum;
    if (packet->state==0) {
	me->p_flags &= ~(PFTRACT | PFPRESS);
	return;
    }
    if (me->p_flags & PFCLOAK) {
	warning("Weapons's Officer:  Cannot tractor while cloaked, sir!");
	return;
    }
    if (target<0 || target>=MAXPLAYER || target==me->p_no) return;
    player= &players[target];
    if (player->p_flags & PFCLOAK) return;
    if (hypot((double) me->p_x-player->p_x,
	    (double) me->p_y-player->p_y) < 
	    ((double) TRACTDIST) * me->p_ship.s_tractrng) {
	if (player->p_flags & PFDOCK) {
	    players[player->p_docked].p_port[player->p_port[0]] = VACANT;
	    players[player->p_docked].p_docked--;
	}
	if (me->p_flags & PFDOCK) {
	    players[me->p_docked].p_docked--;
	    players[me->p_docked].p_port[me->p_port[0]] = VACANT;
	}
	player->p_flags &= ~(PFORBIT | PFDOCK);
	me->p_flags &= ~(PFORBIT | PFDOCK);
	me->p_tractor = target;
	me->p_flags |= PFTRACT;
    } else {
	warning("Weapon's Officer:  Vessel is out of range of our tractor beam.");
    }
}

handleRepressReq(packet)
struct repress_cpacket *packet;
{
    int target;
    struct player *player;

    if (weaponsallowed[WP_TRACTOR]==0) {
	warning("Tractor beams haven't been invented yet.");
	return;
    }
    target=packet->pnum;
    if (packet->state==0) {
        me->p_flags &= ~(PFTRACT | PFPRESS);
        return;
    }
    if (me->p_flags & PFCLOAK) {
	warning("Weapons's Officer:  Cannot tractor while cloaked, sir!");
	return;
    }
    if (target<0 || target>=MAXPLAYER || target==me->p_no) return;
    player= &players[target];
    if (player->p_flags & PFCLOAK) return;
    if (hypot((double) me->p_x-player->p_x,
	    (double) me->p_y-player->p_y) < 
	    ((double) TRACTDIST) * me->p_ship.s_tractrng) {
	if (player->p_flags & PFDOCK) {
	    players[player->p_docked].p_port[player->p_port[0]] = VACANT;
	    players[player->p_docked].p_docked--;
	}
	if (me->p_flags & PFDOCK) {
	    players[me->p_docked].p_docked--;
	    players[me->p_docked].p_port[me->p_port[0]] = VACANT;
	}
	player->p_flags &= ~(PFORBIT | PFDOCK);
	me->p_flags &= ~(PFORBIT | PFDOCK);
	me->p_tractor = target;
	me->p_flags |= (PFTRACT | PFPRESS);
    } else {
	warning("Weapon's Officer:  Vessel is out of range of our tractor beam.");
    }
}

sendMotdLine(line)
char *line;
{
    struct motd_spacket motdPacket;

    motdPacket.type=SP_MOTD;
    strncpy(motdPacket.line, line, 80);
    motdPacket.line[79]='\0';
    sendClientPacket(&motdPacket);
}

handleCoupReq(packet)
struct coup_cpacket *packet;
{
    coup();
}

handleRefitReq(packet)
struct refit_cpacket *packet;
{
    do_refit(packet->ship);
}

handleMessageReq(packet)
struct mesg_cpacket *packet;
{
    char addrbuf[9];
    static long lasttime=0 /*, time()*/ ;
    static int balance=0;	/* make sure he doesn't get carried away */
    long thistime;

    int isCensured();		/* "shut up and play" code 7/21/91 TC */
    int parseIgnore();		/* still more code, 7/24/91 TC */

    /* Some random code to make sure the player doesn't get carried away 
     *  about the number of messages he sends.  After all, he could try to 
     *  jam peoples communications if we let him.
     */
    thistime=time(NULL);
    if (lasttime!=0) {
	balance=balance-(thistime-lasttime);
	if (balance<0) balance=0;
    }
    lasttime=thistime;
    if (balance>=15) {
	warning("Be quiet");
	balance+=3;
	if (balance > time(NULL) + 60) {
	    balance=time(NULL)+60;
	}
	return;
    }
    balance+=3;
    packet->mesg[69]='\0';
    sprintf(addrbuf, " %c%c->", teamlet[me->p_team], shipnos[me->p_no]);
    if (parseIgnore(packet)) return; /* moved this up 4/6/92 TC */
    switch(packet->group) {
    case MALL:
	sprintf(addrbuf+5, "ALL");
	if (isCensured(me->p_login)) {
	    warning("You are censured.  Message was not sent.");
/*	    sprintf(packet->mesg, "%c%c tried to say something but was impeded by sun spot activity.", teamlet[me->p_team], shipnos[me->p_no]);
	    pmessage(packet->mesg, 0, MALL, "GOD->ALL", me->p_no);*/
	    return;
	}
	break;
    case MTEAM:
	if (packet->indiv != FED && packet->indiv != ROM &&
	    packet->indiv != KLI && packet->indiv != ORI) return;
	if (isCensured(me->p_login) && (packet->indiv != me->p_team)) {
	    warning("You are censured.  Message was not sent.");
/*	    sprintf(packet->mesg, "%c%c tried to say something but was impeded by sun spot activity.", teamlet[me->p_team], shipnos[me->p_no]);
	    pmessage(packet->mesg, 0, MALL, "GOD->ALL", me->p_no);*/
	    return;
	}
	sprintf(addrbuf+5, teamshort[packet->indiv]);
	break;
    case MINDIV:
	if (packet->indiv < 0 || packet->indiv >= MAXPLAYER) return;
	if (players[packet->indiv].p_status == PFREE) return;
	if (isCensured(me->p_login) && (players[packet->indiv].p_team != me->p_team)) {
	    warning("You are censured.  Message was not sent.");
/*	    sprintf(packet->mesg, "%c%c tried to say something but was impeded by sun spot activity.", teamlet[me->p_team], shipnos[me->p_no]);
	    pmessage(packet->mesg, 0, MALL, "GOD->ALL", me->p_no);*/
	    return;
	}
	if (ignored[packet->indiv] & MINDIV) {
	    warning("You are ignoring that player.  Message was not sent.");
	    return;
	}
	if ((me->p_team != players[packet->indiv].p_team) &&
	    (isCensured(players[packet->indiv].p_login))) {
	    warning("That player is censured.  Message was not sent.");
	    return;
	}
	sprintf(addrbuf+5, "%c%c ", teamlet[players[packet->indiv].p_team],
	    shipnos[packet->indiv]);
	break;
    default:
	return;
    }
    pmessage2(packet->mesg, packet->indiv, packet->group, addrbuf, me->p_no);
}

handleQuitReq(packet)
struct quit_cpacket *packet;
{
    me->p_flags |= PFSELFDEST;
    selfdest = me->p_updates + 100;
    /* 60 seconds to destruct a starbase */
    if (me->p_ship.s_type==STARBASE) selfdest = me->p_updates + 600;
    warning("Self destruct initiated");
}

sendMaskPacket(mask)
int mask;
{
    struct mask_spacket maskPacket;

    maskPacket.type=SP_MASK;
    maskPacket.mask=mask;
    sendClientPacket(&maskPacket);
}

handleOptionsPacket(packet)
struct options_cpacket *packet;
{
    bcopy(packet->keymap, mystats->st_keymap, 96);
/*    mystats->st_flags = ntohl(packet->flags);*/
    mystats->st_flags = ntohl(packet->flags) |
	(mystats->st_flags & ST_CYBORG); /* hacked fix 8/24/91 TC */
    keeppeace=(mystats->st_flags / ST_KEEPPEACE) & 1;
}

handleSocketReq(packet)
struct socket_cpacket *packet;
{
    nextSocket=ntohl(packet->socket);
    userVersion=packet->version;
    userUdpVersion=packet->udp_version;
}

handleByeReq(packet)
struct bye_cpacket *packet;
{
    noressurect=1;
}

checkVersion()
{
    struct badversion_spacket packet;

    if (userVersion != SOCKVERSION) {
	packet.type=SP_BADVERSION;
	packet.why=0;
	sendClientPacket(&packet);
	flushSockBuf();
	return(0);
    }
    return(1);
}

logEntry()
{
    FILE *logfile;
    int curtime;

    logfile=fopen(LOGFILENAME, "a");
    if (!logfile) return;
    curtime=time(NULL);
    fprintf(logfile, "Joining: %s, (%c) <%s@%s> %s", me->p_name,
	    shipnos[me->p_no],
	    me->p_login, /* debug 2/21/92 TMC */
#ifndef FULL_HOSTNAMES
	    /* revised join log message, 4/13/92 TC */
	    me->p_monitor,
#else
	    me->p_full_hostname,
#endif
	    ctime(&curtime));
/*
    fprintf(logfile, "Joining: %s (%s@%s) (%c), %s", me->p_name,
	me->p_login, me->p_monitor,			/ ATM /
	shipnos[me->p_no], ctime(&curtime));
*/
    fclose(logfile);
}

gwrite(fd, wbuf, bytes)
int fd;
char *wbuf;
register int bytes;
{
    long orig = bytes;
    register long n;
    char tempbuf[80];
    struct timeval to;


    while (bytes) {
	n = write(fd, wbuf, bytes);
	if (n < 0) {
	    if (errno==ENOBUFS) {
		/* The man pages don't mention this as a possibility.
		 * Yet, it happens.  I guess I just wait for 1/10 sec, and
		 *  continue?
		 */
		/* I would use usleep() to do this, but this system ain't
		 *  got it...
		 */
		to.tv_sec=0;
		to.tv_usec=100000;
		select(0,NULL,NULL,NULL,&to);
		continue;
	    }
	    if (fd == udpSock) {
		/* do we want Hiccup code here? */
		UDPDIAG(("Tried to write %d, 0x%x, %d (error %d)\n",
			fd, wbuf, bytes, errno));
		printUdpInfo();
		logmessage("UDP gwrite failed:");
	    }
#ifndef FULL_HOSTNAMES
	    sprintf(tempbuf, "Died in gwrite, n=%d, errno=%d <%s@%s>",
		    n, errno, me->p_login, me->p_monitor);
#else
	    sprintf(tempbuf, "Died in gwrite, n=%d, errno=%d <%s@%s>",
		    n, errno, me->p_login, me->p_full_hostname);
#endif
	    logmessage(tempbuf);
	    return(-1);
	}
	bytes -= n;
	wbuf += n;
    }
    return(orig);
}

handleDockingReq(packet)
struct dockperm_cpacket *packet;
{
    int i;

    if (me->p_ship.s_type == STARBASE) {
	me->p_docked = 0;
	for (i=0; i<NUMPORTS; i++) {
	    if (me->p_port[i] != VACANT)
		players[me->p_port[i]].p_flags &= ~PFDOCK;
	    me->p_port[i] = VACANT;  
	}
	if (packet->state) 
	    me->p_flags |= PFDOCKOK;
	else
	    me->p_flags &= ~PFDOCKOK;
    }
}

handleReset(packet)
struct resetstats_cpacket *packet;
{
    extern int startTkills, startTlosses, startTarms, startTplanets, startTticks;

    if (packet->verify != 'Y') return;

    /* Gee, they seem to want to reset their stats!  Here goes... 
     */
    mystats->st_maxkills=0.0;
    mystats->st_kills=0;
    mystats->st_losses=0;
    mystats->st_armsbomb=0;
    mystats->st_planets=0;
    mystats->st_ticks=0;
    mystats->st_tkills=0;
    mystats->st_tlosses=0;
    mystats->st_tarmsbomb=0;
    mystats->st_tplanets=0;
    mystats->st_tticks=1;
    mystats->st_rank=0;
    mystats->st_sbkills=0;
    mystats->st_sblosses=0;
    mystats->st_sbticks=0;
    mystats->st_sbmaxkills=0.0;

    startTkills = mystats->st_tkills;
    startTlosses = mystats->st_tlosses;
    startTarms = mystats->st_tarmsbomb;
    startTplanets = mystats->st_tplanets;
    startTticks = mystats->st_tticks;
}

handleUpdatesReq(packet)
struct updates_cpacket *packet;
{
    struct itimerval	udt;
    struct timeval timeout;

#define MIN_DELAY 200000	/* load problems 4/3/92 TC */

    timerDelay=ntohl(packet->usecs);
    if (timerDelay < MIN_DELAY) timerDelay=MIN_DELAY;
    if (timerDelay >= 1000000) timerDelay=999999;

    udt.it_interval.tv_sec = 0;
    udt.it_interval.tv_usec = timerDelay;
    udt.it_value.tv_sec = 0;
    udt.it_value.tv_usec = timerDelay;
    setitimer(ITIMER_REAL, &udt, 0);
}

logmessage(string)
char *string;
{
    FILE *fp;

    fp=fopen(LOGFILENAME, "a");
    if (fp) {
	fprintf(fp, "%s\n", string);
	fclose(fp);
    }
}

handleReserved(packet)
struct reserved_cpacket *packet;
{
/*    char temp[20];*/
    struct reserved_cpacket mycp;
    struct reserved_spacket mysp;
    char serverName[64]; /* now get serverName from system 8/2/92 TC */

    if (testtime==1) return;
    if (bcmp(packet->data, testdata, 16) != 0) {
	testtime=1;
	return;
    }
    bcopy(testdata, mysp.data, 16);
    if (gethostname(serverName, 64)) fprintf(stderr, "gethostname() error\n"); /* 8/2/92 TC */
    encryptReservedPacket(&mysp, &mycp, serverName, me->p_no);
    if (bcmp(packet->resp, mycp.resp, 16) != 0) {
	fprintf(stderr, "User verified incorrectly.\n");
	testtime=1;
	return;
    }
    testtime=0;
}

#ifdef ATM_STUFF
    if ((binconfirm == 2) && !strcmp(packet->resp, "Cyborg")) {	/* ATM */
	testtime=0;	/* accept */
	cyborg=1;
	if (me->p_name[0] != '+') {
	    temp[0] = '+';		/* indicate cyborg */
	    strcpy(temp+1, me->p_name);		/* this happens AFTER entry, */
	    temp[15] = '\0';			/* so changing enter() isn't */
	    strcpy(me->p_name, temp);		/* sufficient */
	}
	return;
    }
    if (bcmp(packet->resp, mycp.resp, 16) != 0) {
	fprintf(stderr, "User verified incorrectly.\n");
	testtime=1;
	return;
    }
    testtime=0;
}

#endif

handleScan(packet)		/* ATM */
struct scan_cpacket *packet;
{
/*
    struct scan_spacket response;
    struct player *pp;

    bzero(&response, sizeof(struct scan_spacket));
    response.type = SP_SCAN;
    response.pnum = packet->pnum;
    if (!weaponsallowed[WP_SCANNER]) {
	warning("Scanners haven't been invented yet");
	response.success = 0;
    } else {
	response.success = scan(packet->pnum);

	if (response.success) {
	    / fill in all the goodies /
	    pp = &players[packet->pnum];
	    response.p_fuel = htonl(pp->p_fuel);
	    response.p_armies = htonl(pp->p_armies);
	    response.p_shield = htonl(pp->p_shield);
	    response.p_damage = htonl(pp->p_damage);
	    response.p_etemp = htonl(pp->p_etemp);
	    response.p_wtemp = htonl(pp->p_wtemp);
	}
    }
    sendClientPacket(&response);
*/
}


/*
 * ---------------------------------------------------------------------------
 *	Strictly UDP from here on
 * ---------------------------------------------------------------------------
 */

handleUdpReq(packet)
struct udp_req_cpacket *packet;
{
    struct udp_reply_spacket response;
    int mode;

    response.type = SP_UDP_REPLY;

    if (packet->request == COMM_VERIFY) {
	/* this request should ONLY come through the UDP connection */
	if (commMode == COMM_UDP) {
	    UDPDIAG(("Got second verify from %s; resending server verify\n",
		me->p_name));
	    response.reply = SWITCH_VERIFY;
	    goto send;
	}
	UDPDIAG(("Receieved UDP verify from %s\n", me->p_name));
	UDPDIAG(("--- UDP connection established to %s\n", me->p_name));
#ifdef BROKEN
	warning("WARNING: BROKEN mode is enabled");
#endif

	sequence = 1;		/* reset sequence numbers */
	commMode = COMM_UDP;	/* at last */
	udpMode = MODE_SIMPLE;	/* just send one at a time */
		
	/* note that we don't NEED to send a SWITCH_VERIFY packet; the client */
	/* will change state when it receives ANY packet on the UDP connection*/
	/* (this just makes sure that it gets one) */
	/* (update: recvfrom() currently tosses the first packet it gets...)  */
	response.reply = SWITCH_VERIFY;
	goto send;
/*	return;*/
    }

    if (packet->request == COMM_MODE) {
	/* wants to switch modes; mode is in "conmode" */
	mode = packet->connmode;
	if (mode < MODE_TCP || mode > MODE_DOUBLE) {
	    warning("Server can't do that UDP mode");
	    UDPDIAG(("Got bogus request for UDP mode %d from %s\n",
		mode, me->p_name));
	} else {
	    /* I don't bother with a reply, though it can mess up the opt win */
	    switch (mode) {
	    case MODE_TCP:
		warning("Server will send with TCP only"); break;
	    case MODE_SIMPLE:
		warning("Server will send with simple UDP"); break;
	    case MODE_FAT:
		warning("Server will send with fat UDP; sent full update");
		V_UDPDIAG(("Sending full update to %s\n", me->p_name));
		forceUpdate();
		break;
#ifdef DOUBLE_UDP
	    case MODE_DOUBLE:
		warning("Server will send with double UDP");
		scbufptr=scbuf + sizeof(struct sc_sequence_spacket);
		break;
#else
	    case MODE_DOUBLE:
		warning("Request for double UDP DENIED (set to simple)");
		mode = MODE_SIMPLE;
		break;
#endif /*DOUBLE_UDP*/
	    }

	    udpMode = mode;
	    UDPDIAG(("Switching %s to UDP mode %d\n", me->p_name, mode));
	}
	return;
    }

    if (packet->request == COMM_UPDATE) {
	/* client wants a FULL update */
	V_UDPDIAG(("Sending full update to %s\n", me->p_name));
	forceUpdate();

	return;
    }

    UDPDIAG(("Received request for %s mode from %s\n",
	(packet->request == COMM_TCP) ? "TCP" : "UDP", me->p_name));
    if (packet->request == commMode) {
	/* client asking to switch to current mode */
	if (commMode == COMM_UDP) {
	    /*
	     * client must be confused... whatever the cause, he obviously
	     * isn't connected to us, so we better drop out end and retry.
	     */
	    UDPDIAG(("Rcvd UDP req from %s while in UDP mode; dropping old\n",
		    me->p_name));
	    closeUdpConn();
	    commMode = COMM_TCP;
	    /* ...and fall thru to the UDP request handler */
	} else {
	    /*
	     * Again, client is confused.  This time there's no damage though.
	     * Just tell him that he succeeded.  Could also happen if the
	     * client tried to connect to our UDP socket but failed, and
	     * decided to back off.
	     */
	    UDPDIAG(("Rcvd TCP req from %s while in TCP mode\n", me->p_name));

	    response.reply = SWITCH_TCP_OK;
	    sendClientPacket(&response);

	    if (udpSock >= 0) {
		closeUdpConn();
		UDPDIAG(("Closed UDP socket\n"));
	    }
	    return;
	}
    }

    /* okay, we have a request to change modes */
    if (packet->request == COMM_UDP) {
	udpClientPort = ntohl(packet->port);	/* where to connect to */
	if (!udpAllowed) {
	    UDPDIAG(("Rejected UDP request from %s\n", me->p_name));
	    response.reply = SWITCH_DENIED;
	    response.port = htons(0);
	    goto send;
	} else {
	    if (userUdpVersion != UDPVERSION) {
		char buf[80];
		sprintf(buf, "Server UDP is v%.1f, client is v%.1f",
			(float) UDPVERSION / 10.0,
			(float) userUdpVersion / 10.0);
		warning(buf);
		UDPDIAG(("%s (rejected %s)\n", buf, me->p_name));
		response.reply = SWITCH_DENIED;
		response.port = htons(1);
		goto send;
	    }
	    if (udpSock >= 0) {
		/* we have a socket open, but the client doesn't seem aware */
		/* (probably because our UDP verify got lost down the line) */
		UDPDIAG(("Receieved second request from %s, reconnecting\n",
			me->p_name));
		closeUdpConn(udpSock);
	    }
	    /* (note no openUdpConn(); we go straight to connect) */
	    if (connUdpConn() < 0) {
		response.reply = SWITCH_DENIED;
		response.port = 0;
		goto send;
	    }
	    UDPDIAG(("Connected UDP socket (%d:%d) for %s\n", udpSock,
		    udpLocalPort, me->p_name));

	    /* we are now connected to the client, but he's merely bound */
	    /* don't switch to UDP mode yet; wait until client connects */
	    response.reply = SWITCH_UDP_OK;
	    response.port = htonl(udpLocalPort);

	    UDPDIAG(("packet->connmode = %d\n", packet->connmode));
	    if (packet->connmode == CONNMODE_PORT) {
		/* send him our port # so he can connect to us */
		goto send;
	    } else {
		/* send him a packet; he'll get port from recvfrom() */
		if (gwrite(udpSock, &response, sizeof(response)) !=
							    sizeof(response)) {
		    UDPDIAG(("Attempt to send UDP packet failed; using alt\n"));
		}
		goto send;
	    }
	}
    } else if (packet->request == COMM_TCP) {
	closeUdpConn();
	commMode = COMM_TCP;
	response.reply = SWITCH_TCP_OK;
	response.port = 0;
	UDPDIAG(("Closed UDP socket for %s\n", me->p_name));
	goto send;
    } else {
	fprintf(stderr, "ntserv: got weird UDP request (%d)\n",
		packet->request);
	return;
    }
send:
    sendClientPacket(&response);
}


connUdpConn()
{
    struct sockaddr_in addr;
    int len;

    if (udpSock > 0) {
	fprintf(stderr, "ntserv: tried to open udpSock twice\n");
	return (0);	/* pretend we succeeded (this could be bad) */
    }
    if (udpbufptr != udpbuf) {
	udpbufptr = udpbuf;		/* clear out any old data */
	sequence--;			/* we just killed a sequence packet */
    }

    if ((udpSock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	perror("ntserv: unable to create DGRAM socket");
	return (-1);
    }

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = remoteaddr;	/* addr of our client */
    addr.sin_port = htons(udpClientPort);	/* client's port */

    if (connect(udpSock, &addr, sizeof(addr)) < 0) {
	perror("ntserv: connect to client UDP port");
	UDPDIAG(("Unable to connect() to %s on port %d\n", me->p_name,
		udpClientPort));
	close(udpSock);
	udpSock = -1;
	return (-1);
    }
    UDPDIAG(("connect to %s's port %d on 0x%x succeded\n",
	me->p_name, udpClientPort, remoteaddr));

    /* determine what our port is */
    len = sizeof(addr);
    if (getsockname(udpSock, &addr, &len) < 0) {
	perror("netrek: unable to getsockname(UDP)");
	UDPDIAG(("Can't get our own socket; connection failed\n"));
	close(udpSock);
	udpSock = -1;
	return (-1);
    }
    udpLocalPort = (int) ntohs(addr.sin_port);

    if (udpAllowed > 2)		/* verbose debug mode? */
	printUdpInfo();

    return (0);
}

closeUdpConn()
{
    V_UDPDIAG(("Closing UDP socket\n"));
    if (udpSock < 0) {
	fprintf(stderr, "ntserv: tried to close a closed UDP socket\n");
	return (-1);
    }
    shutdown(udpSock, 2);	/* wham */
    close(udpSock);		/* bam */
    udpSock = -1;		/* (nah) */

    return (0);
}

/* used for debugging */
printUdpInfo()
{
    struct sockaddr_in addr;
    int len;

    len = sizeof(addr);
    if (getsockname(udpSock, &addr, &len) < 0) {
	perror("printUdpInfo: getsockname");
	return;
    }
    UDPDIAG(("LOCAL: addr=0x%x, family=%d, port=%d\n", addr.sin_addr.s_addr,
	addr.sin_family, ntohs(addr.sin_port)));

    if (getpeername(udpSock, &addr, &len) < 0) {
	perror("printUdpInfo: getpeername");
	return;
    }
    UDPDIAG(("PEER : addr=0x%x, family=%d, port=%d\n", addr.sin_addr.s_addr,
	addr.sin_family, ntohs(addr.sin_port)));
}

handleSequence()
{
    /* we don't currently deal with sequence numbers from clients */
}

/*
 * If we're in UDP mode, add a sequence number to the transmission buffer.
 * Returns the #of bytes inserted.
 *
 * This will add a sequence # to transmissions on either channel.  However,
 * the current implementation doesn't put sequences on TCP transmissions
 * because mixed TCP packets and UDP packets rarely arrive in the order
 * in which they were sent.
 */
addSequence(outbuf)
char *outbuf;
{
    struct sequence_spacket *ssp;

    if (commMode != COMM_UDP || udpMode == MODE_TCP)
	return (0);

    ssp = (struct sequence_spacket *) outbuf;
    ssp->type = SP_SEQUENCE;
    ssp->sequence = htons((unsigned short) sequence);
    sequence++;

    return (sizeof(struct sequence_spacket));
}


#ifdef DOUBLE_UDP
/*
 * If we're in double-UDP mode, then we need to send a separate semi-critical
 * transmission over UDP.  We need to give it the same sequence number as the
 * previous transmission, but the sequence packet will have type
 * SP_CP_SEQUENCE instead of SP_SEQUENCE.
 */
sendSC()
{
    struct sequence_spacket *ssp;
    struct sc_sequence_spacket *sc_sp;
    int cc;

    if (commMode != COMM_UDP || udpMode != MODE_DOUBLE) {
	/* mode not active, keep buffer clear */
	scbufptr = scbuf;
	return;
    }
    if (scbufptr - scbuf <= sizeof(struct sc_sequence_spacket)) {
	/* nothing to send */
	return;
    }

    /* copy sequence #, send what we got, then reset buffer */
    sc_sp = (struct sc_sequence_spacket *) scbuf;
    ssp = (struct sequence_spacket *) udpbuf;
    sc_sp->type = SP_SC_SEQUENCE;
    sc_sp->sequence = ssp->sequence;
    if ((cc = gwrite(udpSock, scbuf, scbufptr-scbuf)) != scbufptr-scbuf) {
	fprintf(stderr, "UDP sc gwrite failed (%d, error %d)\n", cc, errno);
	UDPDIAG(("*** UDP diSConnected for %s\n", me->p_name));
	printUdpInfo();
	closeUdpConn();
	commMode = COMM_TCP;
	return;
    }

    scbufptr=scbuf + sizeof(struct sc_sequence_spacket);
}
#endif

/*
 * This is a truncated version of initClientData().  Note that it doesn't
 * explicitly reset all the fat UDP stuff; sendClientData will take care of
 * that by itself eventually.
 *
 * Only semi-critical data is sent, with a few exceptions for non-critical
 * data which would be nice to update (stats, kills, player posn, etc).
 * The critical stuff can't be lost, so there's no point in resending it.
 *
 * (Since the fat data begins in an unqueued state, forceUpdate() should be
 * called immediately after switching to fat mode.  This guarantees that every
 * packet will end up on a queue.  The only real reason for doing this is so
 * that switching to fat mode will clear up your display and keep it cleared;
 * otherwise you could have torps floating around forever because the packet
 * for them isn't on a queue.  Will it reduce the effectiveness of fat UDP?
 * No, because as soon as the player hits the "update all" key it's gonna
 * happen anyway...)
 */
forceUpdate()
{
    static time_t lastone=0;
    time_t now;
    int i;

    now = time(0);
    if (now - lastone < UDP_UPDATE_WAIT) {
	warning("Update request DENIED (chill out!)");
	return;
    }
    lastone = now;

/*    clientDead=0;*/
    for (i=0; i<MAXPLAYER; i++) {
	clientHostile[i].hostile= -1;
	clientStats[i].losses= -1;		/* (non-critical, but nice) */
/*	clientLogin[i].rank= -1;		(critical) */
/*	clientPlayersInfo[i].shiptype= -1;	(critical) */
/*	clientPStatus[i].status= -1;		(critical) */
	clientPlayers[i].x= htonl(-1);		/* (non-critical, but nice) */
	clientPhasers[i].status= -1;
	clientKills[i].kills= htonl(-1);	/* (non-critical, but nice) */
	clientFlags[i].flags= htonl(-1);
	mustUpdate[i]=0;
    }
    for (i=0; i<MAXPLAYER*MAXTORP; i++) {
	clientTorpsInfo[i].status= -1;
/*	clientTorps[i].x= -1;			(non-critical) */
    }
    for (i=0; i<MAXPLAYER*MAXPLASMA; i++) {
	clientPlasmasInfo[i].status= -1;
/*	clientPlasmas[i].x= -1;			(non-critical) */
    }
    for (i=0; i<MAXPLANETS; i++) {
	clientPlanets[i].armies= htonl(-1);
/*	clientPlanetLocs[i].x= htonl(-1);	(critical) */
    }
/*    msgCurrent=(mctl->mc_current+1) % MAXMESSAGE;*/
    clientSelf.pnum= -1;
}

/*
 * ---------------------------------------------------------------------------
 *	Fat City
 * ---------------------------------------------------------------------------
 */

/*
 * This updates the "fat" tables; it's called from sendClientData().
 */
updateFat(packet)
/* Pick a random type for the packet */
struct player_spacket *packet;
{
    FAT_NODE *fatp;
    struct kills_spacket *kp;
    struct torp_info_spacket *tip;
    struct phaser_spacket *php;
    struct plasma_info_spacket *pip;
    /*struct you_spacket *yp;*/
    /*struct status_spacket *sp;*/
    struct planet_spacket *plp;
    struct flags_spacket *fp;
    struct hostile_spacket *hp;
    int idx;

    /* step 1 : find the FAT_NODE for this packet */
    switch (packet->type) {
    case SP_KILLS:
	kp = (struct kills_spacket *) packet;
	idx = (int) kp->pnum;
	fatp = &fat_kills[idx];
	break;
    case SP_TORP_INFO:
	tip = (struct torp_info_spacket *) packet;
	idx = (int) ntohs(tip->tnum);
	fatp = &fat_torp_info[idx];
	break;
    case SP_PHASER:
	php = (struct phaser_spacket *) packet;
	idx = (int) php->pnum;
	fatp = &fat_phaser[idx];
	break;
    case SP_PLASMA_INFO:
	pip = (struct plasma_info_spacket *) packet;
	idx = (int) ntohs(pip->pnum);
	fatp = &fat_plasma_info[idx];
	break;
    case SP_YOU:
	/*yp = (struct you_spacket *) packet;*/
	fatp = &fat_you;
	break;
    case SP_STATUS:
	/*sp = (struct status_spacket *) packet;*/
	fatp = &fat_status;
	break;
    case SP_PLANET:
	plp = (struct planet_spacket *) packet;
	idx = plp->pnum;
	fatp = &fat_planet[idx];
	break;
    case SP_FLAGS:
	fp = (struct flags_spacket *) packet;
	idx = (int) fp->pnum;
	fatp = &fat_flags[idx];
	break;
    case SP_HOSTILE:
	hp = (struct hostile_spacket *) packet;
	idx = (int) hp->pnum;
	fatp = &fat_hostile[idx];
	break;
    default:
	fprintf(stderr,"Fat error: bad semi-critical type in updateFat\n");
	return;
    }

    if (fatp->packet != (PTR) packet) {
	fprintf(stderr, "Fat error: fatp->packet=0x%.8lx, packet=0x%.8lx\n",
		fatp->packet, packet);
	return;
    }
    
    /* step 2 : move this dude to temporary list 0 */
    dequeue(fatp);
    enqueue(fatp, 0);
}

/*
 * This fattens up the transmission, adding up to MAX_FAT_DATA bytes.  The
 * packets which get added will be moved to a higher queue, giving them less
 * priority for next time.  Note that they are added to a parallel temporary
 * list (otherwise they'd could be sent several times in the same transmission
 * as the algorithm steps through the queues), and merged later on.
 *
 * Packets are assigned from head to tail, on a first-fit basis.  If a
 * semi-critical packet is larger than MAX_FAT_DATA, this routine will never
 * send it, but it will skip around it.
 *
 * This routine is called from flushSockBuf, before the transmission is sent.
 *
 * A possible improvement is to have certain packets "expire", never to be
 * seen again.  This way we don't keep resending torp packets for players who
 * are long dead.  This raises the possibility that the dead player's torps
 * will never go away though.
 */
fatten()
{
    int bytesleft;
    FAT_NODE *fatp, *nextfatp;
    int list;

#ifdef FATDIAG
    printf("--- fattening\n");
#endif
    bytesleft = MAX_FAT_DATA;
    for (list = 0; list < MAX_FAT_LIST; list++) {
	fatp = fatlist[list].head;
	while (fatp != NULL) {
	    nextfatp = fatp->prev;	/* move toward tail */
#ifdef FATDIAG
	    if (nextfatp == fatp) {
		printf("Hey!  nextfatp == fatp!\n");
		kill(getpid(), 15);
	    }
#endif
	    if (fatp->pkt_size < bytesleft) {
		/* got one! */
		bcopy(fatp->packet, udpbufptr, fatp->pkt_size);
		udpbufptr += fatp->pkt_size;
		bytesleft -= fatp->pkt_size;

		/* move the packet to a higher queue (if there is one) */
		dequeue(fatp);
		if (list+1 == MAX_FAT_LIST) {
		    /*enqueue(fatp, list);*/	/* keep packets on high queue?*/
		} else {
		    enqueue(fatp, list+1);
		}

		/* done yet? */
		if (bytesleft < MAX_NONFAT)
		    goto done;		/* don't waste time searching anymore */
	    }

	    fatp = nextfatp;
	}
    }

done:
    /* at this point, we either filled with fat or ran out of queued packets */
#ifdef FATDIAG
    printf("--- done\n");
#endif
    V_UDPDIAG(("- Added %d grams of fat\n", MAX_FAT_DATA - bytesleft));
    return (0);		/* some compilers need something after a goto */
}


/*
 * Remove a FAT_NODE from a queue.  If it's at the head or the tail of a
 * list, then we need to figure out which list it's in and update the head
 * or tail pointer.  It's easier to go searching than to maintain a queue
 * number in every FAT_NODE.
 *
 * This routine looks too complex... there must be a simpler way to do this.
 */
dequeue(fatp)
FAT_NODE *fatp;
{
    int i;

#ifdef V_FATDIAG
    for (i = 0; i < MAX_FAT_LIST; i++) {
	printf("fatlist[i].head = 0x%.8lx tail = 0x%.8lx\n",
		fatlist[i].head, fatlist[i].tail);
	if ((fatlist[i].head == NULL && fatlist[i].tail != NULL) ||
	    (fatlist[i].head != NULL && fatlist[i].tail == NULL)) {
	    printf("before!\n");
	    kill(getpid(), 15);
	}
    }
#endif


    if (fatp->next == NULL) {
	/* it's at the head or not in a queue */
	for (i = 0; i < MAX_FAT_LIST; i++) {
	    if (fatlist[i].head == fatp) {
		fatlist[i].head = fatp->prev;		/* move head back */
		if (fatlist[i].head != NULL)
		    (fatlist[i].head)->next = NULL;	/* amputate */
		break;
	    }
	}
    } else {
	/* it's not at the head */
	if (fatp->prev != NULL)
	    fatp->prev->next = fatp->next;
    }

    if (fatp->prev == NULL) {
	/* it's at the tail or not in a queue */
	for (i = 0; i < MAX_FAT_LIST; i++) {
	    if (fatlist[i].tail == fatp) {
		fatlist[i].tail = fatp->next;		/* move head fwd */
		if (fatlist[i].tail != NULL)
		    (fatlist[i].tail)->prev = NULL;	/* amputate */
		break;
	    }
	}
    } else {
	/* it's not at the tail */
	if (fatp->next != NULL)
	    fatp->next->prev = fatp->prev;
    }

#ifdef FATDIAG
    printf("Removed 0x%.8lx...", fatp);		/* FATDIAG */
    for (i = 0; i < MAX_FAT_LIST; i++) {
	if ((fatlist[i].head == NULL && fatlist[i].tail != NULL) ||
	    (fatlist[i].head != NULL && fatlist[i].tail == NULL)) {
	    printf("after: %d %.8lx %.8lx\n", i, fatlist[i].head, fatlist[i].tail);
	    kill(getpid(), 15);
	}
    }
#endif
    fatp->prev = NULL;
    fatp->next = NULL;
}

/*
 * Add a FAT_NODE to the tail of a temporary queue.  The merge() routine
 * merges the temporary queues with the fatlists once the transmission is
 * sent.
 */
enqueue(fatp, list)
FAT_NODE *fatp;
int list;
{
#ifdef FATDIAG
    printf("added to tmplist %d\n", list);		/* FATDIAG */
#endif

    if (tmplist[list].tail == NULL) {
	/* list was empty */
	tmplist[list].tail = tmplist[list].head = fatp;
    } else {
	/* list wasn't empty */
	fatp->next = tmplist[list].tail;
	fatp->next->prev = fatp;
	tmplist[list].tail = fatp;
    }
}

/*
 * This gets called from flushSockBuf after the transmission is sent.  It
 * appends all the packets sitting in temporary queues to the corresponding
 * fat queues, where they will be eligible for fattening next transmission.
 */
fatMerge()
{
    int i;

    for (i = 0; i < MAX_FAT_LIST; i++) {
	if (tmplist[i].head == NULL)
	    continue;		/* temp list is empty, nothing to do */

	if (fatlist[i].head == NULL) {
	    /* fatlist is empty; just copy pointers */
	    fatlist[i].head = tmplist[i].head;
	    fatlist[i].tail = tmplist[i].tail;

	} else {
	    /* stuff in both */
	    (tmplist[i].head)->next = fatlist[i].tail;	/* fwd pointer */
	    (fatlist[i].tail)->prev = tmplist[i].head;	/* back pointer */
	    fatlist[i].tail = tmplist[i].tail;		/* move tail back */
	    tmplist[i].head = tmplist[i].tail = NULL;	/* clear the tmp list */
	}
	tmplist[i].head = tmplist[i].tail = NULL;
#ifdef FATDIAG
	printf("merged list %d: %.8lx %.8lx\n", i, fatlist[i].head,
		fatlist[i].tail);
#endif
    }
}

int isCensured(s)		/* return true if cannot message opponents */
char *s;
{
    return(
	   (strncmp(s, "am4m", 4) == 0) || /* 7/21/91 TC */
	   (strncmp(s, "dm3e", 4) == 0) || /* 7/21/91 TC */
	   (strncmp(s, "gusciora", 8) == 0) || /* 7/25/91 TC */
	   (strncmp(s, "flan", 4) == 0) || /* 4/2/91 TC */
	   (strncmp(s, "kc3b", 4) == 0) || /* 4/4/91 TC */
	   (strncmp(s, "windom", 6) == 0) || /* 7/20/92 TC */
	    0
	   );
}

/* return true if you should eat message */

int parseIgnore(packet)
struct mesg_cpacket *packet;
{
    char c;
    char *s;
    int who;
    int what;
    char buf[80];
    int noneflag;
    extern int ignored[];

#define FALSE 0
#define TRUE 1

/*    if (packet->indiv != me->p_no) return FALSE;*/

    s = packet->mesg;

    who = packet->indiv;
    if ((*s != ':') && (strncmp(s, "     ", 5) != 0)) return FALSE;
    if ((who == me->p_no) || (*s == ' ')) { /* check for borg call 4/6/92 TC */
	if (binconfirm)
	    warning("No cyborgs allowed in the game at this time.");
	else {
	    char buf[80]; char buf2[5];
	    int i; int cybflag = 0;

	    strcpy(buf, "Possible cyborgs: ");
	    for (i = 0; i < MAXPLAYER; i++)
		if ((players[i].p_status != PFREE) &&
		    (players[i].p_stats.st_flags & ST_CYBORG)) {
		    sprintf(buf2, "%s ", players[i].p_mapchars);
		    strcat(buf, buf2);
		    cybflag = 1;
		}
	    if (!cybflag) strcat(buf, "None");
	    warning(buf);
	}
	if (*s != ' ')		/* if not a borg call, eat msg 4/6/92 TC */
	    return TRUE;
	else
	    return FALSE;	/* otherwise, send it 4/6/92 TC */
    }

    if (packet->group != MINDIV) return FALSE; /* below is for indiv only */

    do {
	what = 0;
	switch (*(++s)) {
	  case 'a':
	  case 'A':
	    what = MALL;
	    break;
	  case 't':
	  case 'T':
	what = MTEAM;
	    break;
	  case 'i':
	  case 'I':
	    what = MINDIV;
	    break;
	  case '\0':
	    what = 0;
	    break;
	  default:
	    what = 0;
	    break;
	}
	ignored[who] ^= what;
    } while (what != 0);

    strcpy(buf, "Ignore status for this player: ");
    noneflag = TRUE;
    if (ignored[who] & MALL) {
	strcat(buf, "All "); noneflag = FALSE;
    }
    if (ignored[who] & MTEAM) {
	strcat(buf, "Team "); noneflag = FALSE;
    }
    if (ignored[who] & MINDIV) {
	strcat(buf, "Indiv "); noneflag = FALSE;
    }
    if (noneflag)
	strcat(buf, "None");
    warning(buf);
    return TRUE;
}

/* give session stats if you send yourself a '?' 2/27/92 TC */
/* return true if you should eat message */

int parseQuery(packet)
struct mesg_spacket *packet;	/* was cpacket 4/17/92 TC */
{
    char buf[80];
    float sessionBombing, sessionPlanets, sessionOffense, sessionDefense;
    int deltaArmies, deltaPlanets, deltaKills, deltaLosses, deltaTicks;

    extern int startTkills, startTlosses, startTarms, startTplanets, startTticks;
    
    /* 0-8 for address, 9 is space */
    if (packet->mesg[10] != '?') return FALSE;

    deltaPlanets = me->p_stats.st_tplanets - startTplanets;
    deltaArmies = me->p_stats.st_tarmsbomb - startTarms;
    deltaKills = me->p_stats.st_tkills - startTkills;
    deltaLosses = me->p_stats.st_tlosses - startTlosses;
    deltaTicks = me->p_stats.st_tticks - startTticks;

    if (deltaTicks == 0) return TRUE; /* shouldn't happen */

    sessionPlanets = (float) deltaPlanets * status->timeprod /
	((float) deltaTicks * status->planets);

    sessionBombing = (float) deltaArmies * status->timeprod /
	((float) deltaTicks * status->armsbomb);

    sessionOffense = (float) deltaKills * status->timeprod /
	((float) deltaTicks * status->kills);

    sessionDefense = (float) deltaTicks * status->losses /
	(deltaLosses!=0 ?
	 ((float) deltaLosses * status->timeprod) :
	 (status->timeprod));

    sprintf(buf, "%2s stats: %d planets and %d armies. %d wins/%d losses. %5.2f hours.",
	    me->p_mapchars,
	    deltaPlanets,
	    deltaArmies,
	    deltaKills,
	    deltaLosses,
	    (float) deltaTicks/36000.0);
    bounce(buf, packet->m_from);
    sprintf(buf, "Ratings: Pla: %5.2f  Bom: %5.2f  Off: %5.2f  Def: %5.2f  Ratio: %4.2f",
	    sessionPlanets,
	    sessionBombing,
	    sessionOffense,
	    sessionDefense,
	    (float) deltaKills /
	    (float) ((deltaLosses == 0) ? 1 : deltaLosses));
    bounce(buf, packet->m_from);
    return TRUE;
}

/* new code, sends bouncemsg to bounceto from GOD 4/17/92 TC */
bounce(bouncemsg, bounceto)
char *bouncemsg;
int bounceto;
{
    char buf[10];

    sprintf(buf, "GOD->%s", players[bounceto].p_mapchars);
    pmessage(bouncemsg, bounceto, MINDIV, buf);
}
