/*
 * torp.c
 */
#include "copyright.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include "defs.h"
#include "struct.h"
#include "data.h"

/* Launch torp */

/*
** If a set of given conditions are met, fire a single torp in direction
** course.  Type is used because robots are allowed to shoot straight.
** Torps sent with status TMOVE wobble a bit.  TSTRAIGHT torps (fired
** by robots) move in a straight line.
**
** torp->t_fuse is the life of the torpedo.  It is set here based on
** a random function.  Torps currently live two to five seconds.
*/

static struct timeb lasttime;
static struct timeb curtime;
#define GAP 100
long random();

ntorp(course, type)
unsigned char course;
int type;
{
    register int i;
    register struct torp *k;

    if (me->p_flags & PFWEP) {
	warning("Torpedo launch tubes have exceeded maximum safe temperature!");
	return;
    }

    ftime(&curtime);
    if ((curtime.time == lasttime.time && curtime.millitm < lasttime.millitm+GAP) ||
	(curtime.time == lasttime.time+1 && curtime.millitm+1000-lasttime.millitm < GAP)) {
	return;
    } else {
	lasttime.time = curtime.time;
	lasttime.millitm = curtime.millitm;
    }

    if (me->p_ntorp == MAXTORP) {
	warning("Our computers limit us to having 8 live torpedos at a time captain!");
	return;
    }
    if (me->p_fuel < myship->s_torpcost) {
	warning("We don't have enough fuel to fire photon torpedos!");
	return;
    }
    if (me->p_flags & PFREPAIR) {
	warning("We cannot fire while our vessel is in repair mode.");
	return;
    }
    if ((me->p_cloakphase) && (me->p_ship.s_type != ATT)) {
	warning("We are unable to fire while in cloak, captain!");
	return;
    }

/* change TC 12/9/90 -- my attempt at torp angle stuff */


    if (0 && ((me->p_ship).s_type != STARBASE)) {
	int delta;
	if ((delta = ((int) me->p_dir - (int) course)) < 0)
	    delta = -delta;
	if (delta > topgun) { /* note: 128 = 180 degrees left/right */
	    warning("We only have forward mounted cannons.");
	    return;
	}
    } /* end if topgun */

    me->p_ntorp++;
    me->p_fuel -= myship->s_torpcost;
    me->p_wtemp += (myship->s_torpcost / 10) - 10;	/* Heat weapons */
    for (i = me->p_no * MAXTORP, k = &torps[i];		/* Find a free torp */
	i < me->p_no * MAXTORP + MAXTORP; i++, k++) {
	    if (k->t_status == TFREE)
		break;
    }

    /* Setup data in new torp */
    if (type>TSTRAIGHT || type<TFREE) type=TMOVE;
    k->t_no = i;
    k->t_status = type;
    k->t_owner = me->p_no;
    k->t_team = me->p_team;
    k->t_x = me->p_x;
    k->t_y = me->p_y;
    k->t_dir = course;
    k->t_damage = myship->s_torpdamage;
    k->t_speed = myship->s_torpspeed;
    k->t_war = me->p_hostile | me->p_swar;
    k->t_fuse = myship->s_torpfuse + (random() % 20);
    k->t_turns = myship->s_torpturns;
}

