/*
 * newstartd.c
 *
 *   Please read before use:
 *
 *   There's a really bad hack in newstartd.c that refuses more than 1
 *   connection from any hostname.  It's bad because it uses /usr/ucb/netstat
 *   to figure out who's connected.
 *
 *   At this point, only the hostname of the caller is available.  The
 *   daemon in theory has access to both user name and (partial) hostname,
 *   so that's where multiple logins should be denied.
 *
 *   To bypass this check, create the file defined by NOCOUNT_FILE in
 *   newaccess.c.
 *
 *   Also, you may want to forget newstartd and use listen.c, which is part
 *   of the galaxy server distribution.
 *
 *   4/13/92
 *
 *   Derived from distribution startd.c and xtrekII.sock.c (as of ~1/14/91).
 *
 *   Installation:  update the PROG and LOGFILE variables below, and
 *     set the default port appropriately (presently 2592).  Set debug=1
 *     to run this under dbx.  Link this with access.o and subnet.o.
 *  
 *   Summary of changes:
 *   - subsumes xtrekII.sock into startd (newstartd calls ntserv directly to
 *       reduce OS overhead)
 *   - some debugger support code added
 *   - uses reaper() to keep process table clean (removes <exiting> procs)
 *
 *   8/16/91 Terence Chang
 *
 */

#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <signal.h>
#include <sys/wait.h>

#define PROG 		"/usr/users/terence/bin/ntserv"
#define LOGFILE		"/usr/users/terence/lib/xtrekII/xtreklog"

int port = 2592;
int debug = 0;
int sock = -1;
int newsock = -1;
extern int errno;
int connectionAttemptDetected(), reaper();

/* xtrek.sock global stuff: (TC) */

#define SP_BADVERSION 21
struct badversion_spacket {
    char type;
    char why;
    char pad2;
    char pad3;
};

extern char peerhostname[];	/* Defined in access.c (TC) */
int fd;

main(argc, argv)
int argc;
char **argv;
{
    int tries=0;
    int forked;
    char startprog[255];
    char sockstr[10];
    
    /* xtrekII.sock main() declarations (TC) */
    
    FILE *fp;
    long curtime;
    char logname[BUFSIZ];

    struct badversion_spacket packet;
    int	game;
    char version[81];

    if(argc == 2)
	port = atoi(argv[1]);

    fprintf(stderr,"Accepting xtrek calls to port %d!\n", port);
    signal(SIGCHLD, reaper);
    if (!debug)	detach();

    fd = open(LOGFILE, O_CREAT | O_WRONLY | O_APPEND, 0600);

  keepontrying:
    while(!connectionAttemptDetected()){
	fprintf(stderr, "Whoops.  Bye.\n");
	exit(0);
    }

    if (host_access(NULL)==0) {
	packet.type=SP_BADVERSION;
	packet.why=1;
	write(0, (char *) &packet, sizeof(packet));
	time(&curtime);
	sprintf(logname, "Denied %-32.32s %s",
		peerhostname,
		ctime(&curtime));
	if (fd!=-1) write(fd, logname, strlen(logname));
	sleep(2);
    }
    else if (fork() == 0) { /* we are a clone */
	time(&curtime);
	sprintf(logname, "       %-32.32s %s",
		peerhostname,
		ctime(&curtime));
	if (fd!=-1) write(fd, logname, strlen(logname));
	if (fd!=-1) close(fd);

	execl(PROG, PROG, peerhostname, 0);
	fprintf(stderr,"Error in execl! -- \n");
	reporterror();
    }
    close(0);

    goto keepontrying;
}

int connectionAttemptDetected()
{
    struct sockaddr_in addr;
    struct sockaddr_in naddr;
    struct timeval timeout;
    int readfds;
    int len;

    char *opt;
    int foo = 1;

    if(sock < 0){
	if((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 ){
	    fprintf(stderr,"Hey! I can't make a socket!\n");
	    fprintf(stderr,"I'll try again \n");
	    return(0);
	}
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(port);

/* newstuff*/
	if (setsockopt (sock, SOL_SOCKET, SO_REUSEADDR,
		(char *) &foo, sizeof (int)) < 0)
	{
		perror ("setsockopt REUSEADDR");
		exit (1);
	}
/*end newstuff*/
	
	if(bind(sock, &addr, sizeof(addr)) < 0){
	    if(errno == ENOTSOCK){
		    fprintf(stderr,"bogus! socket is unbindable!\n");
		    fprintf(stderr,"Used Port %d \n", port);
		    reporterror();
		    close(sock);
		    sock = -1;
		    sleep(10);
		    fprintf(stderr,"closing socket!");
		    return(0);
		}
	    if (debug)
		fprintf(stderr, "Taking 10...\n");
	    sleep(10);
	    if (bind(sock, &addr, sizeof(addr)) < 0) {
		if (debug)
		    fprintf(stderr, "Taking 10...\n");
		sleep(10);
		if (bind(sock, &addr, sizeof(addr)) < 0) {
		    fprintf(stderr,"Bogus! I can't bind to the socket!\n");
		    fprintf(stderr,"Error: %d\n", errno);
		    close(sock);
		    return(0);
		}
	    }
	}
	if(listen(sock, 5)<0) {
	    fprintf(stderr,"Listen failed: ");
	    reporterror();
	}
	close(0);
    }
    
    len = sizeof(naddr);
    
    /* hack: to support reaper(), must ignore EINTR errors (TC) */

  nointr:
    newsock = accept(sock, &naddr, &len);
    if ((newsock < 0) && (errno == EINTR))
	goto nointr;

    if(newsock < 0){
	fprintf(stderr,"accept error!");
	reporterror();
	fprintf(stderr,"No one calling!\n");
	shutdown(sock, 2);
	close(sock);
	sock = -1;
	return(0);
    } else {
	if (newsock != 0) {
	    if (dup2(newsock, 0) == -1) {
		fprintf(stderr, "failed dup2\n");
		reporterror();
	    }
	    close(newsock);
	}
	return(1);
    }
}

reporterror()
{
    switch(errno){
      case EBADF:
	fprintf(stderr,"EBADF!\n");
	break;
      case ENOTSOCK:
	fprintf(stderr,"ENOTSOCK!\n");
	break;
      case EOPNOTSUPP:
	fprintf(stderr,"EOPNOTSUPP!\n");
	break;
      case EFAULT:
	fprintf(stderr,"EFAULT!\n");
	break;
      case EWOULDBLOCK:
	fprintf(stderr,"EWOULDBLOCK!\n");
	break;
      case ENOTDIR:
	fprintf(stderr,"ENOTDIR!\n");
	break;
      case EINVAL:
	fprintf(stderr,"EEINVAL!\n");
	break;
      case ENOENT:
	fprintf(stderr,"ENOENT!\n");
	break;
      case ELOOP:
	fprintf(stderr,"ELOOP!\n");
	break;
      case EACCES:
	fprintf(stderr,"EACCES!\n");
	break;
      case ENOEXEC:
	fprintf(stderr,"ENOEXEC!\n");
	break;
      case ENOMEM:
	fprintf(stderr,"ENOMEM!\n");
	break;
      case EADDRNOTAVAIL:
	fprintf(stderr,"EADDRNOTAVAIL!\n");
	break;
      case EADDRINUSE:
	fprintf(stderr,"EADDRINUSE!\n");
	break;
      case EINTR:
	fprintf(stderr,"Interrupted system call.\n");
	break;
      default:
	fprintf(stderr,"reporterror: unknown: %d\n",errno);
    }
}

detach()
{
    int fd;

    fd = open("/dev/tty", O_RDWR, 0);
    if (fd < 0)
	return;
    (void) ioctl(fd, TIOCNOTTY, (char *) NULL);
    (void) close(fd);
}

reaper(sig)
{
    if (debug) fprintf(stderr,"Reaping...\n");
    while (wait3((union wait *) 0, WNOHANG, (struct rusage *) 0) > 0)
	;
}
