#include <X11/Xlib.h>
#include <stdio.h>
/*#include "/usr/att/usr/include/string.h"*/
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/file.h>
#include <math.h>
#include <sys/ipc.h>
/*#include "/usr/att/usr/include/sys/shm.h*/
#include <errno.h>
#include <pwd.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "struct.h"

#define MAX_QUEUE 15 /* if queue gets this big, we're running out of
                        processes and need to exit */

#define HIGH_USAGE       60 /* percentages.  high means report usage at */
#define VERY_HIGH_USAGE  80 /* every interval. 'very' and 'ultra' are */
#define ULTRA_HIGH_USAGE 90 /* just for cute comments. */

#define INTERVAL 300 /* how many seconds between usage samples, report */
                     /* usage every 3 intervals if usage is not high */

#define STAT_CMD "/usr/ucb/vmstat 5 2"

/*

   This program requires that /usr/ucb/vmstat exists, and that the
   output of "/usr/ucb/vmstat 5 2" should look like:

 procs      faults    cpu      memory              page             disk  
 r b w   in  sy  cs us sy id  avm  fre  re at  pi  po  fr  de  sr s0 
 1 0 0  415 457 314 10 19 71  16k 2076   0  0   0   0   0   0   0  0
 1 0 0  232 353 238 11 25 63  16k 2076   0  2   0   0   0   0   0  0
....5....0....5....0....5...

   In particular, the 27th and 28th chars of the 4th line will be parsed
   as a two digit number representing % CPU idle time over the last 5 sec.

*/

struct player *players;
struct player *me;
struct torp *torps;
struct ship *myship;
struct stats *mystats;
struct planet *planets;
struct status *status;
struct phaser *phasers;
struct message *messages;
struct mctl *mctl;
struct team *teams;

openmem()
{
    extern int errno;
    int	shmemKey = PKEY;
    int	shmid;
    struct memory	*sharedMemory;

    errno = 0;
    shmid = shmget(shmemKey, 0, 0);
    if (shmid < 0) {
	if (errno != ENOENT) {
	    perror("shmget");
	    exit(1);
	}
	shmid = shmget(shmemKey, 0, 0);
	if (shmid < 0) {
	    fprintf(stderr, "Daemon not running\n");
	    return 0;
	}
    }
    sharedMemory = (struct memory *) shmat(shmid, 0, 0);
    if (sharedMemory == (struct memory *) -1) {
	perror("shared memory");
	exit (1);
    }
    players = sharedMemory->players;
    torps = sharedMemory->torps;
    planets = sharedMemory->planets;
    status = sharedMemory->status;
    phasers = sharedMemory->phasers;
    mctl = sharedMemory->mctl;
    messages = sharedMemory->messages;
    teams = sharedMemory->teams;
    return 1;
}

int gotmem = 0;

main()
{
    FILE *fp, *popen();
    char buf[100];
    int idle;
    int displayCount = 0;

    while (1) {
	if (!gotmem)		/* try to open shmem */
	    gotmem = openmem();
	if (!gotmem) {		/* if we don't have it, wait */
	    sleep(60);
	    continue;
	}
	if (status->answer > MAX_QUEUE) {
	    /* Um, better not do anything */
	    sleep(60);
	    continue;
	}
	fp=popen(STAT_CMD, "r");
	if (fp==NULL) {
	    mysleep(60);
	    continue;
	}
        mysleep(10);
        if (!gotmem) { /* shmem went away while pipe was running */
            pclose(fp);
            continue;
	}
	fgets(buf, 99, fp);
	fgets(buf, 99, fp);
	fgets(buf, 99, fp);
	fgets(buf, 99, fp); /* only want the 4th line */
	pclose(fp);
	buf[28] = '\0';
	if (sscanf(&buf[26], "%d", &idle) == 1) {
	    sprintf(buf, "GOD->ALL  Current CPU usage is %d%%.", 100-idle);
	    if (idle < 100-ULTRA_HIGH_USAGE)
	        strcat(buf, "  Ack!");
	    else if (idle < 100-VERY_HIGH_USAGE)
	        strcat(buf, "  I can live with it.");
	}

        /* print out usage every 3 intervals or more often if usage is high */

	if (((displayCount % 3) == 0) || (idle < 100-HIGH_USAGE)) {
   	    pmessage(buf, 0, MALL);
	}
        displayCount++;
	mysleep(INTERVAL);	/* wait 5 minutes */
    }
}

mysleep(count)
int count;
{
    while (count-- > 0) {
        /* check to see if daemon's gone... the latter check requires */
        /* a daemon mod, the former doesn't (but requires a queue) */

	if ((status->wait > 0) && (status->count == 0) ||
	    (status->gameup == 0)) {
	    gotmem = 0;		/* mark shmem as unavailable */
	    sleep(10);		/* let the daemon shut down in peace */
	}
	sleep(1);
    }
}

pmessage(str, recip, group)
char *str;
int recip;
int group;
{
    struct message *cur;
    if (++(mctl->mc_current) >= MAXMESSAGE)
	mctl->mc_current = 0;
    cur = &messages[mctl->mc_current];
    cur->m_no = mctl->mc_current;
    cur->m_flags = group;
    cur->m_time = 0;
    cur->m_recpt = recip;
    cur->m_from = 255; /* change 12/11/90 TC */
    (void) sprintf(cur->m_data, "%s", str);
    cur->m_flags |= MVALID;
}
