#include <X11/Xlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/file.h>
#include <math.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <errno.h>
#include <pwd.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "struct.h"

/*
 * What is this?
 * -------------
 * This is the Top Gun tool.  It takes one command line argument, the
 * number of seconds of top gun mode desired.  This mode is basically my
 * version of "dogfighting with weapons from hell", but keeping the ship
 * classes balanced relative to each other:
 *
 *  Torp damage  : 130 pts.  Torp cost  :  100 fuel.
 *  Plasma damage: normal    Plasma cost: 1500 fuel.  Plasmas enhanced.
 *  No kills required for plasmas.  SB rebuild time is zero.
 *
 * Players enjoy it after a stressful game, but generally for not too long;
 * half an hour to an hour is plenty.
 *
 * The daemon, if it understands "TOPGUN=1", will ignore the reconstruction
 * timers on SBs, but limit each team to one SB per team.  Also, players
 * can join any team after they die, as in chaos mode.
 *
 * This tool, once per second, diddles with the stats of live ships as
 * follows:
 *
 * s_torpdamage = 130
 * s_plasmaspeed = 25
 * s_plasmacost = 1500
 * s_torpcost = 100
 *
 * SCOUT
 *   s_plasmadamage = 75
 *   s_plasmafuse = 25
 *   s_plasmaturns = 1
 *
 * ASSAULT
 *   s_plasmadamage = 130
 *   s_plasmafuse = 30
 *   s_plasmaturns = 1
 *
 * CRUISER
 *   s_plasmaturns = 2
 *
 * BATTLESHIP and ASSAULT
 *   s_plasmaturns = 3
 *
 * Installation:
 * -------------
 * Makefile entry looks like any other server tool:
 * 
 * pl3: pl3.c struct.h defs.h
 *	$(CC) -o $@ $(CFLAGS) $(FLAGS) pl3.c $(LIBS)
 *
 * Update the #defines below to reflect your server's paths:
 *
 * "/usr/users/terence/xl/" is the xtrek lib directory
 * ".sysdef" is the name of the system defaults file
 * ".sysdef.topgun" is in the lib directory and looks like this:
 *
 *

TOURN=9
SHIPS=SC,BB,DD,AS,CA,SB
WEAPONS=PLASMA,TRACTOR
PLKILLS=0
SBRANK=0
PLANETS=00,10,20,30
CONFIRM=0
HIDDEN=1
MAXLOAD=100.0
TOPGUN=1
UDP=1

 * Usually I start this up from a shell but the daemon could conceivably
 * exec this tool itself.
 */

#define SYSDEF_SAVE_FILENAME "/usr/users/terence/xl/.sysdef.topgunrestore"

#define SAVE_SYSDEF_CMD "cp /usr/users/terence/xl/.sysdef \
 /usr/users/terence/xl/.sysdef.topgunrestore"

#define TOPGUN_SYSDEF_CMD1 "cp /usr/users/terence/xl/.sysdef.topgun \
 /usr/users/terence/xl/.sysdef.temp"

#define TOPGUN_SYSDEF_CMD2 "mv /usr/users/terence/xl/.sysdef.temp \
 /usr/users/terence/xl/.sysdef"

#define RESTORE_SYSDEF_CMD "mv /usr/users/terence/xl/.sysdef.topgunrestore \
 /usr/users/terence/xl/.sysdef"

struct player *players;
struct player *me;
struct torp *torps;
struct ship *myship;
struct stats *mystats;
struct planet *planets;
struct status *status;
struct phaser *phasers;
struct message *messages;
struct mctl *mctl;
struct team *teams;

int mode=0;

openmem()
{
    extern int errno;
    int	shmemKey = PKEY;
    int	shmid;
    struct memory	*sharedMemory;

    errno = 0;
    shmid = shmget(shmemKey, 0, 0);
    if (shmid < 0) {
	if (errno != ENOENT) {
	    perror("shmget");
	    exit(1);
	}
	shmid = shmget(shmemKey, 0, 0);
	if (shmid < 0) {
	    fprintf(stderr, "Daemon not running\n");
	    exit (1);
	}
    }
    sharedMemory = (struct memory *) shmat(shmid, 0, 0);
    if (sharedMemory == (struct memory *) -1) {
	perror("shared memory");
	exit (1);
    }
    players = sharedMemory->players;
    torps = sharedMemory->torps;
    planets = sharedMemory->planets;
    status = sharedMemory->status;
    phasers = sharedMemory->phasers;
    mctl = sharedMemory->mctl;
    messages = sharedMemory->messages;
    teams = sharedMemory->teams;
}

char buf[100];

main(argc, argv)
int argc;
char **argv;
{
    int i;
    int count;
    int byebye();
    FILE *fp;

    if (argc != 2) {
	printf("Usage: %s <# secs.>\n", argv[0]);
	exit(1);
    }
    openmem();

    if ((fp = fopen(SYSDEF_SAVE_FILENAME, "r")) == NULL) { /* if we haven't saved */
        system(SAVE_SYSDEF_CMD); /* save existing .sysdef for shutdown */
        system(TOPGUN_SYSDEF_CMD1); /* make a copy of .sysdef.topgun */
        system(TOPGUN_SYSDEF_CMD2); /* mv it as .sysdef (to be atomic) */
    }
    else /* don't overwrite saved sysdef (how'd it get there?) */
	fclose(fp);

    for (i = 0; i < NSIG; i++) {
	(void) signal(i, byebye);
    }
    count = atoi(argv[1]);
    pmessage("GOD->ALL", MALL);
    sprintf(buf,"GOD->ALL  Top Gun rules are in effect for %d minutes.",
	    count/60);
    pmessage(buf, 0, MALL);
    rules();

    while (count--) {
	for (i=0; i<MAXPLAYER; i++) {
	    if (players[i].p_status == PFREE)
		continue;
	    players[i].p_ship.s_torpdamage = 130;
	    players[i].p_ship.s_plasmaspeed = 25;
	    players[i].p_ship.s_plasmacost = 1500;
	    players[i].p_ship.s_torpcost = 100;
	    if (players[i].p_ship.s_type == SCOUT) {
		players[i].p_ship.s_plasmadamage = 75;
		players[i].p_ship.s_plasmafuse = 25;
		players[i].p_ship.s_plasmaturns = 1;
	    }
	    else if (players[i].p_ship.s_type == ASSAULT) {
		players[i].p_ship.s_plasmadamage = 130;
		players[i].p_ship.s_plasmafuse = 30;
		players[i].p_ship.s_plasmaturns = 1;
	    }
	    /*		players[i].p_ship.s_phaserdamage=200;*/
	     switch (players[i].p_ship.s_type) { 
	       case CRUISER:
		 players[i].p_ship.s_plasmaturns = 2; /* double turn rate? */
	       case BATTLESHIP:
	       case ASSAULT:
		 players[i].p_ship.s_plasmaturns = 3; /* triple turn rate? */
	     }
	}			/* end for */
	sleep(1);
	if ((count > 0) && ((count % 120) == 0))  {
	    sprintf(buf, ">                -= T O P    G U N =-      (%d min. left) ", (count/60));
	    pmessage(buf , 0, MALL);
	    rules();
	}    
    }
    byebye();
}

rules()
{
    pmessage(">  Torp damage  : 130 pts.  Torp cost  :  100 fuel.", 0, MALL);
    pmessage(">  Plasma damage: normal    Plasma cost: 1500 fuel.  Plasmas enhanced.", 0, MALL);
    pmessage(">  No kills required for plasmas.  SB rebuild time is zero.", 0, MALL);
}

byebye()
{
    int player;

    pmessage("GOD->ALL", 0, MALL);
    sprintf(buf,"GOD->ALL  Top Gun Rules are no longer in effect.");
    pmessage(buf, 0, MALL);
    pmessage("GOD->ALL", 0, MALL);
    system(RESTORE_SYSDEF_CMD);
    sleep(1);

    for (player=0; player<MAXPLAYER; player++) {
	if (players[player].p_status == PFREE)
	    continue;
	players[player].p_explode=5;
	players[player].p_status=3;
	players[player].p_whydead=KPROVIDENCE;
	players[player].p_whodead=0;
    }
    exit(0);
}

pmessage(str, recip, group)
char *str;
int recip;
int group;
{
    struct message *cur;
    if (++(mctl->mc_current) >= MAXMESSAGE)
	mctl->mc_current = 0;
    cur = &messages[mctl->mc_current];
    cur->m_no = mctl->mc_current;
    cur->m_flags = group;
    cur->m_time = 0;
    cur->m_recpt = recip;
    cur->m_from = 255;
    (void) sprintf(cur->m_data, "%s", str);
    cur->m_flags |= MVALID;
}
