#include <X11/Xlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/file.h>
#include <math.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <errno.h>
#include <pwd.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "struct.h"

struct player *players;
struct player *me;
struct torp *torps;
struct ship *myship;
struct stats *mystats;
struct planet *planets;
struct status *status;
struct phaser *phasers;
struct message *messages;
struct mctl *mctl;
struct team *teams;

int mode=0;

openmem()
{
    extern int errno;
    int	shmemKey = PKEY;
    int	shmid;
    struct memory	*sharedMemory;
    
    errno = 0;
    shmid = shmget(shmemKey, 0, 0);
    if (shmid < 0) {
	if (errno != ENOENT) {
	    perror("shmget");
	    exit(1);
	}
	shmid = shmget(shmemKey, 0, 0);
	if (shmid < 0) {
	    fprintf(stderr, "Daemon not running\n");
	    exit (1);
	}
    }
    sharedMemory = (struct memory *) shmat(shmid, 0, 0);
    if (sharedMemory == (struct memory *) -1) {
	perror("shared memory");
	exit (1);
    }
    players = sharedMemory->players;
    torps = sharedMemory->torps;
    planets = sharedMemory->planets;
    status = sharedMemory->status;
    phasers = sharedMemory->phasers;
    mctl = sharedMemory->mctl;
    messages = sharedMemory->messages;
    teams = sharedMemory->teams;
}
char *ships[] = {"SC", "DD", "CA", "BB", "AS", "SB"};

char *statnames[] = {"F", "O", "A", "E", "D"};

char *ranknames[] = { "Ensign", "Lieutenant", "Lt. Cmdr.", "Commander",
			  "Captain", "Flt. Capt.", "Commodore", "Rear Adm.", "Admiral"};

main(argc, argv)
     int argc;
     char **argv;
{
    int i;
    char monitor[17];
    int teams[9];
    int wins[9], losses[9];
    float totalRatings;
    float hours;
    
    if (argc>1) mode= *(argv[1]);
    openmem();
    for (i=0; i<9; i++) teams[i]=0;
    printf("<>=======================================================================<>\n");
    switch(mode) {
      case 'l':
	break;
      case 'd':
	printf("  Pl: Name             Type  Kills Damage Shields Armies   Fuel\n");
	break;
      case 'r':
	printf("  Pl: Name             Status Type Bombing Planets Offense Defense\n");
	break;

      default:
	printf("  Pl: Rank       Name              Login      Display          Ratings   DI\n");
	break;
    }
    printf("<>=======================================================================<>\n");
    for (i=0; i<MAXPLAYER; i++) {
	if (players[i].p_status == PFREE || players[i].p_stats.st_tticks==0)
	    continue;
	teams[players[i].p_team]++;
	totalRatings = bombingRating(players+i)+
	    planetRating(players+i)+
		offenseRating(players+i);
	hours = (float) players[i].p_stats.st_tticks/(10*60*60);
	switch(mode) {
	  case 'd':
	    printf("  %2s: %-16s   %2s %6.2f %6d %7d %6d %6d\n",
		   players[i].p_mapchars, players[i].p_name,
		   ships[players[i].p_ship.s_type],
		   players[i].p_kills,
		   players[i].p_damage,
		   players[i].p_shield,
		   players[i].p_armies,
		   players[i].p_fuel);
	    break;
	  case 'r':
	    printf("  %2s: %-16s %s/%4d  %2s %7.2f %7.2f %7.2f %7.2f\n", 
		   players[i].p_mapchars,
		   players[i].p_name, 
		   statnames[players[i].p_status],
		   players[i].p_ghostbuster,
		   ships[players[i].p_ship.s_type],
		   bombingRating(players+i),
		   planetRating(players+i),
		   offenseRating(players+i),
		   defenseRating(players+i));
	    break;
	  default:
	    strncpy(monitor, players[i].p_monitor, 16);
	    monitor[16]=0;
	    printf(" %c%2s: %-10s %-16s  %-10s %-16s %5.2f %8.2f\n",
		   (players[i].p_stats.st_flags & ST_CYBORG) ? '*' : ' ',
		   players[i].p_mapchars,
		   ranknames[players[i].p_stats.st_rank],
		   players[i].p_name,
		   players[i].p_login,
		   monitor,
		   totalRatings,
		   totalRatings * hours);
	    break;
	}
    }
    printf("<>=======================================================================<>\n");
    printf("  Feds: %d   Roms: %d   Kli: %d   Ori: %d\n", 
	   teams[FED], teams[ROM], teams[KLI], teams[ORI]);
    printf("<>=======================================================================<>\n");
}
