/* wander.c -- make the core planets circle the homeworld
   usage:
   wander -d<num>      update planet locations every <num>/10 seconds
 */

#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/file.h>
#include <math.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <errno.h>
#include <pwd.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "struct.h"
#include "../ntserv/data.h"

struct player *players;
struct player *me;
struct torp *torps;
struct ship *myship;
struct stats *mystats;
struct planet *planets;
struct phaser *phasers;
struct message *messages;
struct mctl *mctl;
struct team *teams;
struct plasmatorp *plasmatorps;
struct status *status;

#define HEIGHT 20
#define WIDTH 80
#define MESSAGES 3
#define THEIGHT HEIGHT+MESSAGES
char *ships = "0123456789abcdefghij";
char *teamnames[] = {"---","FED","ROM","???","KLI","???","???","???","ORI"};


char lc[]="ifrxkxxxo";

int pl_home = 0;
int pl_core[4];
int pl_dist[4];
double increment = 0.01;

main(argc, argv)
int argc;
char **argv;
{
    char buf[80];
    char buf2[80];
    char race = 'F';
    double dx, dy;
    int i;

    pl_home = 0; pl_core[0] = 5; pl_core[1] = 7; pl_core[2] = 8; pl_core[3] = 9;
    openmem();
    planets[5].pl_x = 30000; planets[5].pl_y = 90000;
    planets[7].pl_x = 11000; planets[7].pl_y = 75000;
    planets[8].pl_x =  8000; planets[8].pl_y = 93000;
    planets[9].pl_x = 32000; planets[9].pl_y = 74000;
    for (i=1; i<argc; i++) {
	if (argv[i][0]=='-') {
	    switch(argv[i][1]) {
	      case 'i':
		sscanf(argv[i][2], "%f", &increment);
		if (increment==0.0) increment=0.01;
		break;
	      case 'F': pl_home = 0; pl_core[0] = 5; pl_core[1] = 7; pl_core[2] = 8; pl_core[3] = 9; race = argv[i][1]; break;
	      case 'R': pl_home = 10; pl_core[0] = 12; pl_core[1] = 15; pl_core[2] = 16; pl_core[3] = 19; race = argv[i][1]; break;
	      case 'K': pl_home = 20; pl_core[0] = 24; pl_core[1] = 26; pl_core[2] = 29; pl_core[3] = 25; race = argv[i][1]; break;
	      case 'O': pl_home = 30; pl_core[0] = 34; pl_core[1] = 37; pl_core[2] = 38; pl_core[3] = 35; race = argv[i][1]; break;
	    default:
		usage(argv[0]);
		break;
	    }
	  switch (race) {
	    case 'F':  strcpy(buf2, "Federation"); break;
	    case 'R':  strcpy(buf2, "Romulan"); break;
	    case 'K':  strcpy(buf2, "Klingon"); break;
	    case 'O':  strcpy(buf2, "Orion"); break;
	  }
/*	    sprintf(buf, "GOD->ALL  The %s core planets are circling. Do not panic.", buf2);
	    pmessage(buf, 0, MALL);*/
	}
    }

    for (i = 0; i < 4; i++) {
	dx = (double) (planets[pl_core[i]].pl_x - planets[pl_home].pl_x);
	dy = (double) (planets[pl_home].pl_y - planets[pl_core[i]].pl_y);
	pl_dist[i] = sqrt(dx * dx + dy * dy);
}

    while (1) {
	double dir;
	/* unsigned char dir; */

	for (i = 0; i < 4; i++) {
	    dir =
		(atan2((double) (planets[pl_core[i]].pl_y - planets[pl_home].pl_y),
		       (double) (planets[pl_core[i]].pl_x - planets[pl_home].pl_x))
		 );
	    dir += increment;
	    fprintf(stderr, "Dir is %4.2f.\n", dir);
	    planets[pl_core[i]].pl_x = planets[pl_home].pl_x + pl_dist[i] * cos(dir);
	    planets[pl_core[i]].pl_y = planets[pl_home].pl_y + pl_dist[i] * sin(dir);
	    planets[pl_core[i]].pl_flags |= PLREDRAW;
	}
	sleep(1);
    }
}

char *shiptypes[] = {"SC", "DD", "CA", "BB", "AS", "SB"};

usage(string)
char *string;
{
    printf("Usage: %s [-dnnn]\n", string);
    printf("  -dnnn  delay nnn 1/10 seconds between frames\n");
}

pmessage(str, recip, group)
char *str;
int recip;
int group;
{
    struct message *cur;
    if (++(mctl->mc_current) >= MAXMESSAGE)
	mctl->mc_current = 0;
    cur = &messages[mctl->mc_current];
    cur->m_no = mctl->mc_current;
    cur->m_flags = group;
    cur->m_time = 0;
    cur->m_recpt = recip;
    cur->m_from = 255; /* change 3/30/91 TC */
    (void) sprintf(cur->m_data, "%s", str);
    cur->m_flags |= MVALID;
}

openmem()
{
    extern int errno;
    int	shmemKey = PKEY;
    int	shmid;
    struct memory	*sharedMemory;

    errno = 0;
    shmid = shmget(shmemKey, 0, 0);
    if (shmid < 0) {
	if (errno != ENOENT) {
	    fprintf(stderr, "shmget\n");
	    exit(1);
	}
	shmid = shmget(shmemKey, 0, 0);
	if (shmid < 0) {
	    fprintf(stderr, "Daemon not running\n");
	    exit (1);
	}
    }
    sharedMemory = (struct memory *) shmat(shmid, 0, 0);
    if (sharedMemory == (struct memory *) -1) {
	fprintf(stderr, "shared memory\n");
	exit (1);
    }
    players = sharedMemory->players;
    torps = sharedMemory->torps;
    plasmatorps = sharedMemory->plasmatorps;
    planets = sharedMemory->planets;
    phasers = sharedMemory->phasers;
    mctl = sharedMemory->mctl;
    messages = sharedMemory->messages;
    teams = sharedMemory->teams;
    status = sharedMemory->status;
}

