/* setgalaxy.c -- new tool, supercedes resetplanets

   usage:
   setgalaxy l              restore planet locations
   setgalaxy r              standard reset of galaxy
   setgalaxy t              tourney reset of galaxy - equal agris
   setgalaxy f              flatten all planets to 1 army
   setgalaxy F              top out all planets at 30 armies
   setgalaxy n <num>:<str>  rename planet <num> to <str>
*/

#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/file.h>
#include <math.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <errno.h>
#include <pwd.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "struct.h"
#include "../ntserv/data.h"

struct player *players;
struct player *me;
struct torp *torps;
struct ship *myship;
struct stats *mystats;
struct planet *planets;
struct phaser *phasers;
struct message *messages;
struct mctl *mctl;
struct team *teams;
struct plasmatorp *plasmatorps;
struct status *status;

int agricount[4] = { 0, 0, 0, 0};

struct planet pdata[MAXPLANETS] = {
    {  0, (FED|PLHOME|PLFUEL|PLREPAIR|PLCORE), FED, 20000, 80000, "Earth", 5, 30, FED },
    {  1, FED, FED, 10000, 60000, "Rigel", 5, 30, FED },
    {  2, FED, FED, 25000, 60000, "Canopus", 7, 30, FED },
    {  3, FED, FED, 44000, 81000, "Beta Crucis", 11, 30, FED },
    {  4, FED, FED, 39000, 55000, "Organia", 7, 30, FED },
    {  5, FED|PLCORE, FED, 30000, 90000, "Deneb", 5, 30, FED },
    {  6, FED, FED, 45000, 66000, "Ceti Alpha V", 12, 30, FED },
    {  7, FED|PLCORE, FED, 11000, 75000, "Altair", 6, 30, FED },
    {  8, FED|PLCORE, FED,  8000, 93000, "Vega", 4, 30, FED },
    {  9, FED, FED, 32000, 74000, "Alpha Centauri", 14, 30, FED },
    { 10, (ROM|PLHOME|PLFUEL|PLREPAIR|PLCORE), ROM, 20000, 20000, "Romulus", 7, 30, ROM },
    { 11, ROM, ROM, 45000,  7000, "Eridani", 7, 30, ROM },
    { 12, ROM|PLCORE, ROM,  4000, 12000, "Aldeberan", 9, 30, ROM },
    { 13, ROM, ROM, 42000, 44000, "Regulus", 7, 30, ROM },
    { 14, ROM, ROM, 13000, 45000, "Capella", 7, 30, ROM },
    { 15, ROM|PLCORE, ROM, 28000,  8000, "Tauri", 5, 30, ROM },
    { 16, ROM|PLCORE, ROM, 28000, 23000, "Draconis", 8, 30, ROM },
    { 17, ROM, ROM, 40000, 25000, "Sirius", 6, 30, ROM },
    { 18, ROM, ROM, 25000, 44000, "Indi", 4, 30, ROM },
    { 19, ROM, ROM,  8000, 29000, "Hydrae", 6, 30, ROM },
    { 20, (KLI|PLHOME|PLFUEL|PLREPAIR|PLCORE), KLI, 80000, 20000, "Klingus", 7, 30, KLI },
    { 21, KLI, KLI, 70000, 40000, "Pliedes V", 9, 30, KLI },
    { 22, KLI, KLI, 60000, 10000, "Andromeda", 9, 30, KLI },
    { 23, KLI, KLI, 54000, 40000, "Lalande", 7, 30, KLI },
    { 24, KLI|PLCORE, KLI, 88000,  12000, "Pollux", 6, 30, KLI },
    { 25, KLI, KLI, 90000, 37000, "Lyrae", 5, 30, KLI },
    { 26, KLI|PLCORE, KLI, 69000, 31000, "Scorpii", 7, 30, KLI },
    { 27, KLI, KLI, 83000, 48000, "Mira", 4, 30, KLI },
    { 28, KLI, KLI, 54000, 21000, "Cygni", 5, 30, KLI },
    { 29, KLI|PLCORE, KLI, 73000,  5000, "Castor", 6, 30, KLI },
    { 30, (ORI|PLHOME|PLFUEL|PLREPAIR|PLCORE), ORI, 80000, 80000, "Orion", 5, 30, ORI },
    { 31, ORI, ORI, 92000, 59000, "Cassiopia", 9, 30, ORI },
    { 32, ORI, ORI, 65000, 55000, "El Nath", 7, 30, ORI },
    { 33, ORI, ORI, 52000, 60000, "Spica", 5, 30, ORI },
    { 34, ORI|PLCORE, ORI, 72000, 69000, "Procyon", 7, 30, ORI },
    { 35, ORI, ORI, 64000, 80000, "Polaris", 7, 30, ORI },
    { 36, ORI, ORI, 56000, 89000, "Arcturus", 8, 30, ORI },
    { 37, ORI|PLCORE, ORI, 91000, 94000, "Ursae Majoris", 13, 30, ORI },
    { 38, ORI, ORI, 70000, 93000, "Herculis", 8, 30, ORI },
    { 39, ORI|PLCORE, ORI, 85000, 70000, "Antares", 7, 30, ORI }
};

#define AGRI_LIMIT 3

main(argc, argv)
int argc;
char **argv;
{
    int i;

    srandom(getpid());
    openmem();
    if (argc == 1) usage();

    if (argc == 3) {
	int num;
	char name[16];
	if (sscanf(argv[2], "%d:%[^\n]", &num, name)==2) {
	    if ((num >= 0) && (num <= 39)) {
		printf("Renaming planet #%d to %s.\n", num, name);
		strcpy(planets[num].pl_name, name);
		planets[num].pl_namelen = strlen(name);
		planets[num].pl_flags |= PLREDRAW;
	    }
	    else
		printf("Planet number must be in range (0-39).\n");
	    exit();
	}
	else usage();
    }

    if (*argv[1] == 'l') {
	for (i = 0; i < MAXPLANETS; i++) {
	    planets[i].pl_x = pdata[i].pl_x;
	    planets[i].pl_y = pdata[i].pl_y;
	}
	printf("Restored locations.\n");
	exit();
    }

    if (*argv[1] == 'f') { /* flatten planets */
	for (i = 0; i < 40; i++) {
	    planets[i].pl_armies = 1;
	}
	printf("All planets set to 1 army.\n");
    }
    else if (*argv[1] == 'F') { /* top out planets */
	for (i = 0; i < 40; i++) {
	    planets[i].pl_armies = 30;
	}
	printf("All planets set to 30 armies.\n");
    }
    else if (*argv[1] == 't') { /* tourney reset resources, owners */
	do {
	    bcopy(pdata, planets, sizeof(pdata));
	    doResources();
/*	    printf("Agri counts: %d/%d/%d/%d.\n", agricount[0],
		   agricount[1], agricount[2], agricount[3]);*/
	} while (!((agricount[0] <= AGRI_LIMIT) &&
		 (agricount[0] == agricount[1]) &&
		 (agricount[1] == agricount[2]) &&
		 (agricount[2] == agricount[3])));
    }
    else if (*argv[1] == 'r') { /* reset resources, owners */
	do {
	    bcopy(pdata, planets, sizeof(pdata));
	    doResources();
	} while ((agricount[0] > AGRI_LIMIT) || /* darn bug 1/23/92 TC */
		 (agricount[1] > AGRI_LIMIT) || /* was && */
		 (agricount[2] > AGRI_LIMIT) ||
		 (agricount[3] > AGRI_LIMIT));
	printf("Agri counts: %d/%d/%d/%d.\n", agricount[0],
	       agricount[1], agricount[2], agricount[3]);
    }
    else usage();
}

doResources()
{
    int i;

    agricount[0]=agricount[1]=agricount[2]=agricount[3]=0;
    for (i = 0; i < 40; i++) {
	planets[i].pl_flags = pdata[i].pl_flags;
	if (random() % 4 == 0)
	    planets[i].pl_flags |= PLREPAIR;
	if (random() % 2 == 0)
	    planets[i].pl_flags |= PLFUEL;
	if (random() % 8 == 0) { /* was 6 */
	    planets[i].pl_flags |= PLAGRI;
	    agricount[i/10]++;
	}
    }
}

usage()
{
    printf("   usage:\n");
    printf("   setgalaxy l              restore planet locations\n");
    printf("   setgalaxy r              standard reset of galaxy\n");
    printf("   setgalaxy t              tourney reset of galaxy - equal agris\n");
    printf("   setgalaxy f              flatten all planets to 1 army\n");
    printf("   setgalaxy F              top out all planets at 30 armies\n");
    printf("   setgalaxy n <num>:<str>  rename planet <num> to <str>\n");
    exit();
}

openmem()
{
    extern int errno;
    int	shmemKey = PKEY;
    int	shmid;
    struct memory	*sharedMemory;

    errno = 0;
    shmid = shmget(shmemKey, 0, 0);
    if (shmid < 0) {
	if (errno != ENOENT) {
	    fprintf(stderr, "shmget\n");
	    exit(1);
	}
	shmid = shmget(shmemKey, 0, 0);
	if (shmid < 0) {
	    fprintf(stderr, "Daemon not running\n");
	    exit (1);
	}
    }
    sharedMemory = (struct memory *) shmat(shmid, 0, 0);
    if (sharedMemory == (struct memory *) -1) {
	fprintf(stderr, "shared memory\n");
	exit (1);
    }
    players = sharedMemory->players;
    torps = sharedMemory->torps;
    plasmatorps = sharedMemory->plasmatorps;
    planets = sharedMemory->planets;
    phasers = sharedMemory->phasers;
    mctl = sharedMemory->mctl;
    messages = sharedMemory->messages;
    teams = sharedMemory->teams;
    status = sharedMemory->status;
}

